#!/usr/bin/env bash

# $1 - current program version
# $2 - package name (prname-version)
# $3 - dir where store package
# $4 - directory with AUR files (PKGBUILD. *.install etc.)

PKGBUILD_FILE="${4}/PKGBUILD"

# get package version and release
VER=$(\sed -n 's/^pkgver=\(.*\)$/\1/p' ${PKGBUILD_FILE})
REL=$(\sed -n 's/^pkgrel=\(.*\)$/\1/p' ${PKGBUILD_FILE})
if [[ "$VER" != "$1" ]]; then
  # version in PKGBUILD and current pr version are different
  # - reset package release
  REL=1
else
  # are same
  # - increment package release
  ((REL++))
fi

PACK_NAME=${2}    # pr-ver
PACK_DIR=${3}     # .package
TARBALL=${PACK_NAME}-${REL}.tar.gz
AUR_DIR=${4}      # MakeRecipes/arch
AURBALL=${PACK_NAME}-${REL}.src.tar.gz

FILES_TO_PACK="src* sh_* lib* .doc* doc* tests* vim_* MakeRecipes Makefile ChangeLog INSTALL LICENSE README VERSION"


# create pack-directory tree with files intended to pack
# and tarball pack-directory and create 'current' link
\mkdir -vp ${PACK_DIR}/${PACK_NAME}
\cp -vaf ${FILES_TO_PACK} ${PACK_DIR}/${PACK_NAME}

\cd ${PACK_DIR}
\tar -cvzf ${TARBALL} ${PACK_NAME}
\ln -sf ${TARBALL} current.tar.gz
\cd -

\rm -rf ${PACK_DIR}/${PACK_NAME}
# NOW HAVE: .package / pr-ver.tar.gz, current.tar.gz


# Alter PKGBUID file: checksums, pkg-release
pack_md5sum=$(md5sum ${PACK_DIR}/current.tar.gz)
pack_sha256sum=$(sha256sum ${PACK_DIR}/current.tar.gz)
pack_md5sum="md5sums=('${pack_md5sum%% *}')"
pack_sha256sum="sha256sums=('${pack_sha256sum%% *}')"; \
\sed -i "s/md5sums=(.*)/${pack_md5sum}/" ${PKGBUILD_FILE}; \
\sed -i "s/sha256sums=(.*)/${pack_sha256sum}/" ${PKGBUILD_FILE};

\sed -i "s/^pkgrel=.*$/pkgrel=${REL}/" ${PKGBUILD_FILE}
# NOW HAVE: same + PKGBUILD altered


# make AurBall and create 'current_aur' link
\cp -vaf ${PACK_DIR}/${TARBALL} ${AUR_DIR};     # now PKGBUILD will find tarball in own directory and will not download old version
\cd ${AUR_DIR} && mkaurball -f
\rm ${TARBALL}
\ln -sf ${AURBALL} current_aur.tar.gz
\cd -
\mv -vf ${AUR_DIR}/*.tar.gz ${PACK_DIR}
# NOW HAVE: same + .package / aurball.src.tar.gz , current_aur.tar.gz


fossil rm ${PACK_DIR}/tfman*.tar.gz
fossil add ${PACK_DIR}/${TARBALL}
fossil add ${PACK_DIR}/${AURBALL}


# Note to user
echo "Aurball created, checksums substituted in PKGBUILD"
echo -e "You can now, in this order:\n" \
"\t1.push to repository (to upload new tarball with new sources)\n" \
"\t2.place aurball in AUR (to upload new aurball with new PKGBUILD checksums)"


exit 0
