#ifndef MHCL_BASE_SRCBASE_FAULTS_H
#define MHCL_BASE_SRCBASE_FAULTS_H

/*!F::User's faults - external interface file.
 *
 *  ::Provides function declarations for other modules.
 */




/* current project */
#include "common_mhlbase.h"

/* mh library */
#include "core.h"
#include "errors.h"
#include "debug.h"
#include "strings.h"

/* libc */




/*!:: Fault cathegory.
 */
enum faultcath_en {
   Faultcath_none=0,
   Faultcath_ignore,    //! no influence on program flow, can continue
   Faultcath_continue,  //! change program flow slightly (different output etc.)
   Faultcath_exit,      //! program must cleanup and exit when possible
   Faultcath_abort,     //! program must abort immediately
   faultcath_count };
typedef enum faultcath_en faultcath_en;


typedef struct st_fault st_fault;
/*!:: Fault.
 *    Fault (error) detected during program execution.
 */
struct st_fault
{
   faultcath_en cath;   //! cathegory of fault
   str_t descr;         //! description of fault
};
#define ST_FAULT_Z (st_fault){Faultcath_none,STR_T_Z}

typedef struct st_faults_arr st_faults_arr;
/*!:: Faults array.
 *    Array of pointers to fault structures.
 */
struct st_faults_arr
{
   st_fault** buf;   //! array of pointers
   uint na;          //! number of pointers allocated for use
   uint ns;          //! pointers set
   faultcath_en caths_num[faultcath_count];  //! number of faults from each cathegory
};
#define ST_FAULTS_ARR_Z (st_faults_arr){NULL,0,0,{0}}




extern st_fault* fault_create (faultcath_en cath, const char* fmt, ...)
                  __attr_nonnull__(2)
                  __attr_wures__
                  __attr_format__(__printf__,2,3)
                  ;

extern st_fault* fault_vcreate (faultcath_en cath, const char* fmt, va_list vlist)
                  __attr_nonnull__(2)
                  __attr_wures__
                  __attr_format__(__printf__,2,0)
                  ;

extern void fault_destroy (st_fault** pstFault)
               __attr_nonnull__(1)
               ;

extern void fault_print (FILE* fs, st_fault* stFault,
                     const char* pref, const char* post)
               __attr_nonnull__(1,2)
               ;




extern void faultsarr_add_fault (st_faults_arr* stFaultsArr,
         faultcath_en cath, const char* fmt_descr, ...)
               __attr_nonnull__(1,3)
               __attr_format__(__printf__,3,4)
               ;

extern void faultsarr_vadd_fault (st_faults_arr* stFaultsArr,
         faultcath_en cath, const char* fmt_descr, va_list vlist)
               __attr_nonnull__(1,3)
               __attr_format__(__printf__,3,0)
               ;

extern void faultsarr_finalize (st_faults_arr* stFaultsArr)
               __attr_nonnull__(1)
               ;

/*!1:: Return subsequent faults from array of faults. */
extern st_fault* faultsarr_it (const st_faults_arr* stFaultsArr, uint* idx)
               __attr_nonnull__(1,2)
               ;

extern void faultsarr_print (faultcath_en cath, FILE* fs, const st_faults_arr* stFaultsArr,
   const char* prec_line, const char* pref, const char* post, const char* suc_line)
               __attr_nonnull__(2,3)
               ;




#ifdef DEBUG
bool faults_all_tests_run (void)
      __attr_wures__
      ;
#endif




#endif /*- MHCL_BASE_SRCBASE_FAULTS_H -*/
