#ifndef MHCL_BASE_SRCBASE_LOGS_H
#define MHCL_BASE_SRCBASE_LOGS_H

/*!F::Messages logging - external interface file.
 *
 *  ::Provides function declarations for other modules.
 */




/* current project */
#include "common_mhlbase.h"

/* mh library */
#include "core.h"

/* libc */
#include <stdarg.h>        // va_*
#include <stdio.h>      // printf(), fflush(), FILE*




/*!1:: Print consistently formatted message to specified file stream.
 */
extern void log_string (FILE* filestream,
   const char* prefix, const char* filename, uint line, const char* funcname,
   const char* format, ...)
      __attr_nonnull__(1,2,3,5,6)
      __attr_format__(__printf__,6,7)
      ;




/*!:: Stream where all debug messages go by default. */
#ifndef LOG_DEF_STREAM
#  define LOG_DEF_STREAM stderr
#endif

/* Prefixes for various types of log messages.
 */
#ifndef LOG_PREF_DEBUG
#define LOG_PREF_DEBUG "   DBG "
#endif
#ifndef LOG_PREF_INFO
#define LOG_PREF_INFO "  INFO "
#endif
#ifndef LOG_PREF_WARN
#define LOG_PREF_WARN " -WARN-"
#endif
#ifndef LOG_PREF_ERROR
#define LOG_PREF_ERROR "=ERROR="
#endif
#ifndef LOG_PREF_TEST
#define LOG_PREF_TEST "test__>"
#endif
#ifndef LOG_PREF_ASSERT
#define LOG_PREF_ASSERT "Assert>"
#endif




/*!:: Print message with error prefix to 'LOG_DEF_STREAM'.
 */
#define LogError(...) do { \
   log_string(LOG_DEF_STREAM, LOG_PREF_ERROR, __FILE__, __LINE__, __func__, __VA_ARGS__); \
} while(0)

/*!:: Print message with test prefix to 'LOG_DEF_STREAM'.
 */
#define LogTest(...) do { \
   log_string(LOG_DEF_STREAM, LOG_PREF_TEST, __FILE__, __LINE__, __func__, __VA_ARGS__); \
} while(0)




#ifdef DEBUG
/* Used only by tests in debug mode. */
bool logs_all_tests_run (void)
      __attr_wures__
      ;
#endif




#endif /*- MHCL_BASE_SRCBASE_LOGS_H -*/
