/* This file is part of program distributed under GNU General Public
   License. For more information see README and LICENSE files. */

#ifndef MHCL_BASE_SRCCONTAINERS_ARRPTRS_INTER_H_INC
#define MHCL_BASE_SRCCONTAINERS_ARRPTRS_INTER_H_INC

/*!F::Array of pointers - internal interface file.
 *
 *  ::Provides functions for basic manipulations of this module data structures.
 */

/* employs includes from arrptrs.c */




/* assign all data */
static inline
void
ass_ (
   arr_ptrs_t* arrptrs,
   void** buf,
   uint na,
   uint ns,
   arr_ptrs_el_destr_ft pf_el_destr
){
   arrptrs->buf = buf;
   arrptrs->na = na;
   arrptrs->ns = ns;
   arrptrs->pf_el_destr = pf_el_destr;
   return;
}

/* return specified pointer in array */
static inline
void*
ret_ (
   const arr_ptrs_t* arrptrs,
   size_t idx
){
   return arrptrs->buf[idx];
}

/* alloc buffer for pointers inside array struct */
static inline
void*
prep_ (
   arr_ptrs_t* arrptrs,
   uint ne
){
   size_t nb = ne * sizeof(void*);     // overflow danger
   void* buf_new = malloc(nb);
      if (buf_new == NULL) { return NULL; }
   arrptrs->buf = buf_new;
   arrptrs->na = ne;
   arrptrs->ns = 0;
   return buf_new;
}

/* free strings buffer inside array struct */
static inline
void
clean_ (
   arr_ptrs_t* arrptrs
){
   free(arrptrs->buf);
   arrptrs->buf = NULL; arrptrs->na = 0; arrptrs->ns = 0;
   return;
}

/* free array struct */
static inline
void
destr_ (
   arr_ptrs_t* arrptrs
){
   free(arrptrs);
   return;
}

/* Zero specified number of elements (string structs) in array struct buffer. */
static inline
void
rubout_ (
   arr_ptrs_t* arrptrs,
   size_t idx,
   uint ne
){
   memset(&arrptrs[idx], 0, ne * sizeof(void*));
   return;
}

/* increase buffer inside array struct by `ne_add` pointers. */
static inline
void**
infl_ (
   arr_ptrs_t* arrptrs,
   uint ne_add
){
   size_t ne = arrptrs->na + ne_add;
   size_t nb = ne * sizeof(void*);     // overflow danger
   void** buf_new = calloc(1, nb);
      if (buf_new == NULL) { return NULL; }
   memcpy(buf_new, arrptrs->buf, arrptrs->na*sizeof(void*));
   free(arrptrs->buf);
   arrptrs->buf = buf_new;
   arrptrs->na = ne;
   return buf_new;
}

/* Print debug info about struct. */
static inline
void
dbg_print_ (
   arr_ptrs_t* arrptrs
){
   fprintf(stdout, "buf=%p\n", (void*)arrptrs->buf);
   fprintf(stdout, "ns=%d,na=%d\n\n", arrptrs->ns, arrptrs->na);
   for (size_t i = 0; i < arrptrs->ns; ++i) {
      fprintf(stdout, "idx %zu:%p\n", i, arrptrs->buf[i]);
   }
   /*
   for (size_t i = 0; i < arrptrs->na; ++i) {
      fprintf(stdout, "idx %zu:%zu\n", i, arrptrs->buf[i]);
   }
   */
   return;
}




#endif /*- MHCL_BASE_SRCCONTAINERS_ARRPTRS_INTER_H_INC -*/
