#ifndef MH_LIB_BASE_SRC_BASE_VECTOR_H
#define MH_LIB_BASE_SRC_BASE_VECTOR_H

/*!F::Vector data structure - external interface file.
 *
 *  ::Provides function declarations for other modules.
 */




/* mh library */
#include "core.h"

/* current project */
#include "common_mhlbase.h"
#include "errors.h"

/* libc */
#include <string.h>     // memmove()
//#include <stdlib.h>     // abort()
//#include <stdio.h>      // printf(), fflush()




/*!:: Bookkeeping - common for all vectors */
typedef struct st_vector_bk st_vector_bk;
struct st_vector_bk
{
   size_t ns;        //! elements set - at begining, from el 0
   size_t na;        //! elements allocated - whole space available for els
   uint _ext_els;    //! number of elements to additionaly allocate when extending vector
};
#define ST_VECBK_ZERO (st_vector_bk){0,0,0}


static inline
void
extend_ (
   void** buf,        //! [in/out] buffer to extend
   st_vector_bk* bk, //! [in/out] buffer bookkeeping attributes
   size_t nb_el      //! size of one element in buffer
){
   /*: extend */
   size_t na_new = bk->na + bk->_ext_els;       //: new num of allocated elements
   void* buf_tmp = *buf;                        //: temporary buffer
   buf_tmp = realloc(*buf, na_new * nb_el);
      ErrorTrigIf(buf_tmp == NULL, handle_err);

   /*: set buffer attributes */
   *buf = buf_tmp;
   bk->na = na_new;

   return;

  handle_err:
   return;
}

/*!:: Macro for defining new vector for specified elements type. */
#define Define_vector(NAME, TYPE) \
typedef struct NAME NAME; \
struct NAME \
{ \
   TYPE* buf; \
   size_t (*app) (NAME*, TYPE); \
   size_t (*prep) (NAME*, TYPE); \
   TYPE (*ret_el) (NAME*, size_t idx); \
   st_vector_bk _bk; \
}; \
static inline size_t append_vec_##NAME(NAME* obj, TYPE val) { \
   if (obj->_bk.ns >= obj->_bk.na) { \
      extend_((void**)&obj->buf, &obj->_bk, sizeof(TYPE)); \
         HandleErr(handle_err); \
   } \
   obj->buf[obj->_bk.ns] = val; \
   ++obj->_bk.ns; \
   return obj->_bk.ns - 1; \
  handle_err: \
   return 0; \
} \
static inline size_t prepend_vec_##NAME(NAME* obj, TYPE val) { \
   if (obj->_bk.ns >= obj->_bk.na) { \
      extend_((void**)&obj->buf, &obj->_bk, sizeof(TYPE)); \
         HandleErr(handle_err); \
   } \
   memmove(obj->buf+1, obj->buf, obj->_bk.ns * sizeof(TYPE)); \
   obj->buf[0] = val; \
   ++obj->_bk.ns; \
   return obj->_bk.ns - 1; \
  handle_err: \
   return 0; \
} \
static inline TYPE return_el_vec_##NAME(NAME* obj, size_t idx) { \
   ErrorTrigIf(idx > obj->_bk.ns, handle_err); \
   return obj->buf[idx]; \
  handle_err: \
   return (TYPE){0}; \
} \
static inline NAME* create_vec_##NAME(uint ext_els) { \
   NAME* obj = malloc(sizeof(NAME)); \
      ErrorTrigIf(obj == NULL, handle_err); \
   *obj = (NAME) vec_ini(NAME, ext_els); \
   return obj; \
  handle_err: \
   return NULL; \
}

/*!:: Initialize vector object NAME. */
#define vec_ini(NAME, EL_EXT) {NULL, \
   append_vec_##NAME,prepend_vec_##NAME,return_el_vec_##NAME, \
   {0,0,(EL_EXT)}}

/*!:: Finalize vector object OBJ. */
#define vec_fin(OBJ) \
   free((OBJ)->buf); (OBJ)->buf = NULL; \
   (OBJ)->_bk = ST_VECBK_ZERO;

/*!:: Create vector object NAME. */
#define vec_cre(NAME, EL_EXT) create_vec_##NAME((EL_EXT))

/*!:: Destroy vector object OBJ. */
#define vec_des(OBJ) \
   free((OBJ)->buf); free((OBJ)); (OBJ) = NULL;



#ifdef DEBUG
bool vector_all_tests_run (void)
         __attr_wures__;
#endif




#endif /*- MH_LIB_BASE_SRC_BASE_VECTOR_H -*/
