#ifndef MHCL_BASE_SRCMEMORY_MEMORY_INTER_H_INC
#define MHCL_BASE_SRCMEMORY_MEMORY_INTER_H_INC

/*!F::Memory management - external interface file.
 *
 *  ::Provides function declarations for other modules.
 */




/* current project */
#include "common_mhlbase.h"

/* mh library */
#include "core.h"
#include "errors.h"

/* libc */
#include <stdlib.h>     // malloc() free()




/*!1:: Allocate memory block on heap.
 */
extern void* mem_alloc (size_t size)
                  __attr_wures__
                  __attr_alloc_size__(1)
                  __attr_malloc__
                  ;

/*!1:: Allocate memory block on heap and set all bytes to 0.
 */
extern void* mem_allocz (size_t size)
                  __attr_wures__
                  __attr_alloc_size__(1)
                  __attr_malloc__
                  ;

/*!1:: Change size of memory block pointed to by `ptr`.
 */
extern void* mem_realloc (void* ptr, size_t new_size)
                  __attr_nonnull__(1)
                  __attr_wures__
                  __attr_alloc_size__(2)
                  ;

/*!1:: Change size of memory block pointed to by `ptr` and set all added bytes
 *    to 0.
 */
extern void* mem_reallocz (void* ptr, size_t old_size, size_t new_size)
                  __attr_nonnull__(1)
                  __attr_wures__
                  __attr_alloc_size__(3)
                  ;

/*!:: Free memory block pointed by `ptr` and set `ptr` to NULL.
 *
 * ::Preconditions:
 *  - `ptr` must point to previously allocated memory block or be NULL,
 *  otherwise it points to not valid memory area and behavior is undefined
 *
 * ::Effects:
 *  + `ptr` is not NULL: memory block is freed, `ptr` is set to NULL
 *  + `ptr` is NULL: nothing done
 *
 * ::Errors: NONE.
 */
#define mem_free(ptr) do { if ((ptr)) { free((ptr)); (ptr)=NULL; } } while(0)




#ifdef DEBUG
bool memory_all_tests_run (void)
      __attr_wures__;
#endif




#endif /*- MHCL_BASE_SRCMEMORY_MEMORY_INTER_H_INC -*/
