/* This file is part of program distributed under GNU General Public
   License. For more information see README and LICENSE files. */

#ifndef MHCL_BASE_SRCSTRINGS_CSTRINGS_H
#define MHCL_BASE_SRCSTRINGS_CSTRINGS_H

/*!F::C-like strings manage routines - external interface file.
 *
 *  ::Provides functions for other modules.
 */




/* current project */
#include "common_mhlbase.h"

/* mh library */
#include "core.h"
#include "errors.h"
//#include "memory.h"

/* libc */
#include <stdarg.h>     // va_*, vsnprintf()

/* current module */




/* Prepare and cleanup. */
/* -------------------------------------------------------------------------- */


/*!1:: Create c-string from given c-string. */
extern char* cs_cre (const char* cstr)
               __attr_nonnull__(1)
               __attr_wures__
               ;

/*!1:: Create c-string from given c-string. Supply length. */
extern char* cs_cre2 (const char* cstr, size_t num_chars)
               __attr_nonnull__(1)
               __attr_wures__
               ;

/*!1:: Create c-string from given printf-like format. */
extern char* cs_cref (const char* format, ...)
               __attr_wures__
               __attr_format__(__printf__,1,2)
               ;

/*!1:: ~~cs_cref() */
extern char* cs_vcref (const char* format, va_list vlist)
               __attr_wures__
               __attr_format__(__printf__,1,0)
               ;

/*!1:: Destroy c-string. */
extern void cs_des (char** ptr_cstr)
               __attr_nonnull__(1)
               ;




/* Manipulate. */
/* -------------------------------------------------------------------------- */


/*!1:: Prepend given c-string. */
extern char* cs_prep (char** cstr_dest, const char* cstr_src)
               __attr_nonnull__(1,2)
               ;

/*!1:: Prepend given c-string. Supply sizes. */
extern char* cs_prep2 (char** cstr_dest, size_t num_chars_dest,
                     const char* cstr_src, size_t num_chars_src)
               __attr_nonnull__(1,3)
               ;

/*!1:: Append given c-string. */
extern char* cs_app (char** cstr_dest, const char* cstr_src)
               __attr_nonnull__(1,2)
               ;

/*!1:: Append given c-string. Supply sizes. */
extern char* cs_app2 (char** cstr_dest, size_t num_chars_dest,
                     const char* cstr_src, size_t num_chars_src)
               __attr_nonnull__(1,3)
               ;

/*!1:: Append printf-like formatted c-string.
 */
extern void cs_appf (char** cstr_dest, const char* format, ...)
               __attr_nonnull__(1)
               __attr_format__(__printf__,2,3)
               ;

/*!1:: ~~cs_vappf
 */
extern void cs_vappf (char** cstr_dest, const char* format, va_list vlist)
               __attr_nonnull__(1)
               __attr_format__(__printf__,2,0)
               ;




#ifdef DEBUG
bool cstr_all_tests_run (void)
      __attr_wures__
      ;
#endif




#endif /*- MHCL_BASE_SRCSTRINGS_CSTRINGS_H -*/
