/* This file is part of program distributed under GNU General Public
   License. For more information see README and LICENSE files. */

#ifndef MHCL_BASE_SRCSTRINGS_CSTRINGS_INTER_H_INC
#define MHCL_BASE_SRCSTRINGS_CSTRINGS_INTER_H_INC

/*!F::C-like strings manage routines - internal interface file.
 *
 *  ::Provides functions for basic manipulations of this module data structures.
 */


/* employs includes from cstrings.c */




/* Manipulate data. */
static inline void zero_ (char* cs);

/* Prepare. */
static inline char* constr_ (size_t nb);
static inline void destr_ (char* cs);
static inline char* prep_ (char** cs, size_t nc);
static inline void clean_ (char* cs);

/* Resize. */
static inline char* infl_ (char** cs, uint nb_add);
static inline char* defl_ (char** cs, uint nb_sub);
static inline char* levl_ (char** cs, uint nb_new);

/* Copy. */
static inline void overw_ (char* dest, size_t idx, uint nb, const char* src);
static inline void read_ (char* dest, const char* src, size_t idx, uint nb);
static inline void put_ (char* cs, size_t idx, uint nb, const char* cstr);
static inline void pull_ (char* dest, char* cs, size_t idx, uint nb);




/* set all chars to \0 */
static inline
void
zero_ (
   char* cs
){
   size_t len = strlen(cs);
   memset(cs, '\0', len);
   return;
}

/* alloc `nb` memory, init with \0 char */
static inline
char*
constr_ (
   size_t nb
){
   char* cs_new = malloc(nb);
      if (cs_new == NULL) { return NULL; }
   cs_new[0] = '\0';
   return cs_new;
}

/* free c-string */
static inline
void
destr_ (
   char* cs
){
   free(cs);
   return;
}

/* alloc `nb` memory, assign address to `*cs`, init with \0 char */
static inline
char*
prep_ (
   char** cs,
   size_t nb
){
   char* cs_new = malloc(nb);
      if (cs_new == NULL) { return NULL; }
   cs_new[0] = '\0';
   *cs = cs_new;
   return cs_new;
}

/* free c-string */
static inline
void
clean_ (
   char* cs
){
   free(cs);
   return;
}

/* realloc `*cs` to have `nb` memory more. */
static inline
char*
infl_ (
   char** cs,
   uint nb_add
){
   size_t nb = strlen(*cs) + 1 + nb_add;     // overflow danger
   char* cs_new = realloc(*cs, nb);
      if (cs_new == NULL) { return NULL; }
   *cs = cs_new;
   return cs_new;
}

/* realloc `*cs` to have `nb` memory less. */
static inline
char*
defl_ (
   char** cs,
   uint nb_sub
){
   size_t nb = strlen(*cs) + 1 - nb_sub;     // overflow danger
   char* cs_new = realloc(*cs, nb);
      if (cs_new == NULL) { return NULL; }
   *cs = cs_new;
   return cs_new;
}

/* realloc `*cs` to have `nb` memory. */
static inline
char*
levl_ (
   char** cs,
   uint nc_new
){
   size_t nb = nc_new + 1;
   char* cs_new = realloc(*cs, nb);
      if (cs_new == NULL) { return NULL; }
   cs_new[nb-1] = '\0';
   *cs = cs_new;
   return cs_new;
}

/* copy `nb` from `src` to `dest` at index `idx`. */
static inline
void
overw_ (
   char* dest,
   size_t idx,
   uint nb,
   const char* src
){
   memcpy(&dest[idx], src, nb);
   return;
}

/* copy `nb` from `src` at index `idx` to `dest`. */
static inline
void
read_ (
   char* dest,
   const char* src,
   size_t idx,
   uint nb
){
   memcpy(dest, &src[idx], nb);
   return;
}

/* Insert `nb` from `src` to `dest` at index `idx`, moving following chars
   further. */
static inline
void
put_ (
   char* dest,
   size_t idx,
   uint nb,
   const char* src
){
   size_t nb_dest = strlen(dest) + 1;     // overflow danger
   for (size_t i = nb_dest-nb; i > idx; --i) {
      dest[i+nb-1] = dest[i-1];  //: move chars in dest to make room
   }
   memcpy(&dest[idx], src, nb);  //: copy source over
   return;
}

/* Copy `nb` from `src` at index `idx` to `dest` and pull out copied chars
   from source (bring following chars to begining). */
static inline
void
pull_ (
   char* dest,
   char* src,
   size_t idx,
   uint nb
){
   memcpy(dest, &src[idx], nb);
   size_t nb_src = strlen(src) + 1;    // overflow danger
   for (size_t i = idx; i+nb < nb_src; ++i) {
      src[i] = src[i+nb];
   }
   return;
}




#endif /*- MHCL_BASE_SRCSTRINGS_CSTRINGS_INTER_H_INC -*/
