/* This file is part of program distributed under GNU General Public
   License. For more information see README and LICENSE files. */

#ifndef MHCL_BASE_SRCSTRINGS_STRINGOPS_AUX_H_INC
#define MHCL_BASE_SRCSTRINGS_STRINGOPS_AUX_H_INC

/*!F::String operations - auxillary file.
 *
 *  ::Provides helper functions for main implementation file.
 */




/* employs includes from string_operations.c */




/* Split on outermost {}, leave as is if no matching {}. */
static inline void split_on_braces (strarr_t* arr_dest, const str_t* str)
                     __attr_nonnull__(1,2)
                     ;

/* Split on every comma outside {} pairs. */
static inline void split_on_commas (strarr_t* arr_dest, const str_t* str)
                     __attr_nonnull__(1,2)
                     ;

/* Brace {} expansion. */
static void expand_braces (strarr_t* sarr, const str_t* str, bool fst_call)
               __attr_nonnull__(1,2)
               ;




/* Helper struct for 'split_on_braces()'. */
typedef struct st_incstr_pos st_incstr_pos;
struct st_incstr_pos
{
   char* fst;  //first brace in cstring
   char* lst;  //last
};

/* Will split given string on outermost {} pairs. Nested {} are left as is, as
   parts of split results. String before first { and after last } are added to
   result too. If no braces found whole string is result.
   ENOMEM
   errors not covered properly !!!
 */
static inline
void split_on_braces (
   strarr_t* arr_dest,     //! [out] splitted strings
   const str_t* str        //! string to split
){
   ___Dassert(arr_dest, "'arr_dest' parameter invalid");
   ___Dassert(str, "'str' parameter invalid");

   uint br_open = 0;       //num of braces currently opened
   char* pch = str->buf;   //ptr to current char
   char* pbeg = str->buf;  //ptr to first char in currently parsed string part
   char* pbr_fst = NULL;   //ptr to first opening brace in current pair

   /*: Walk string char by char: */
   while (*pch != '\0')
   {
      /*: opening brace encountered: */
      if (*pch == '{') {
         if (br_open == 0) { pbr_fst = pch; }   //: first opening brace in current pair
         ++br_open;
      }
      /*: closing brace encountered: */
      else if (*pch == '}') {
         /*: is paired with previously opened brace */
         if (br_open > 0) {
            --br_open;
            //: outermost brace pair closed
            if (br_open == 0)
            {
               char old = 0;

               /*: Append part of string before first opening brace: */
               old = *pbr_fst; *pbr_fst = '\0'; //: save old char, place temp \0
               if (*pbeg != '\0') {
                  sarr_app(arr_dest, pbeg);
                     HandleErr(handle_err);
               }
               *pbr_fst = old;                  //: restore old char

               /*: Append part of string between braces pair: */
               old = *pch; *pch = '\0';         //: save old char, place temp \0
               if (*(pbr_fst+1) != '\0') {
                  sarr_app(arr_dest, pbr_fst+1);
                     HandleErr(handle_err);
               }
               *pch = old;                      //: restore old char

               pbeg = pch+1;
            }
         }
         /*: fault - closing brace without matchin opening brace */
         else
         {
            goto handle_err;
         }
      }
      ++pch;
   }

   /*: Append last part (after last closing brace): */
   if (pbeg != NULL && pbeg != pch && pbeg != str->buf && *pbeg != '\0') {
      sarr_app(arr_dest, pbeg);
         HandleErr(handle_err);
   }

   /*: fault - too many opened braces */
   if (br_open > 0) {
      goto handle_err;
   }

   /*: Finish: */
   return;

  handle_err:
   sarr_fin(arr_dest);
   return;
}

/* Will split given string on each comma that is doscovered outside of all
   braces pairs. Commas inside {} are ignored.
   ENOMEM
   errors not covered properly !!!
 */
static inline
void split_on_commas (
   strarr_t* arr_dest,     //! [in/out] splitted strings
   const str_t* str        //! string to split
){
   ___Dassert(arr_dest, "'arr_dest' parameter invalid");
   ___Dassert(str, "'str' parameter invalid");

   uint br_open = 0;       //num of braces currently opened
   char* pch = str->buf;   //ptr to current char
   char* pbeg = str->buf;  //ptr to first char in currently parsed string part
   char* pcom = NULL;      //ptr to prev discovered comma

   /*: Walk string char by char: */
   while (*pch != '\0')
   {
      /*: opening brace encountered: */
      if (*pch == '{')
      {
         ++br_open;
      }
      /*: closing brace encountered: */
      else if (*pch == '}')
      {
         if (br_open > 0) { --br_open; }
      }
      /*: comma encountered: */
      else if (*pch == ',')
      {
         //: split only if we are outside of all braces
         if (br_open == 0)
         {
            pcom = pch;
            char old = *pcom; *pcom = '\0';  //: save old char, place temp \0
            sarr_app(arr_dest, pbeg);
               HandleErr(handle_err);
            *pcom = old;                     //: restore old char
            pbeg = pch + 1;
         }
      }
      ++pch;
   }

   /*: Append last part (after last comma): */
   if (pbeg != NULL && pbeg != pch && pbeg != str->buf && *pbeg != '\0') {
      sarr_app(arr_dest, pbeg);
         HandleErr(handle_err);
   }

   /*: fault - too many opened braces */
   if (br_open > 0) {
      goto handle_err;
   }

   /*: Finish: */
   return;

  handle_err:
   sarr_fin(arr_dest);
   return;
}

static
void
expand_braces (
   strarr_t* sarr,      //! [out] resulting strings
   const str_t* str,    //! string to expand
   bool fst_call        //! on first call we do differently
){
   ___Dassert(sarr, "'sarr' parameter invalid");
   ___Dassert(str, "'str' parameter invalid");

   /* ....................................................................... */
   /*: First call: */
   if (fst_call == true)
   {
      strarr_t sarr_split = {0};    //result of splitting
      /*: Split on braces: */
      split_on_braces(&sarr_split, str);
         HandleErr(handle_err);

      /*: Split successful (pair of outermost braces found): */
      if (sarr_split.ns > 0)
      {
         const str_t* str_split = NULL;
         /*: For each splitting result: */
         for (uint i = 0; (str_split = sarr_its(&sarr_split, &i)) != NULL;) {
            strarr_t sarr_tmp = {0};
            //: Expand braces again - recursive call:
            expand_braces(&sarr_tmp, str_split, false);
               HandleErr(handle_err);
            //: Multiple expansion with uptill now results:
            if (sarr->ns == 0) { sarr_ini(sarr, ""); }//: no more expansion: issue empty array
            sarr_multipleA(sarr, &sarr_tmp);       //: multiple expansion and result array
               HandleErr(handle_err);
            sarr_fin(&sarr_tmp);                      //: finish expansion result
         }
         sarr_fin(&sarr_split);
      }
      /*: No braces pairs and any single commas found: */
      else
      {
         sarr_ini(sarr, str->buf);  //: set input string as result
      }
   }
   /* ....................................................................... */
   /*: Subsequent calls: */
   else
   {
      strarr_t sarr_split = {0};    //result of splitting
      /*: Split on commas: */
      split_on_commas(&sarr_split, str);
         HandleErr(handle_err);

      /*: Split successful (commas outside {} found): */
      if (sarr_split.ns > 0)
      {
         const str_t* str_split = NULL;
         /*: For each splitting result: */
         for (uint i = 0; (str_split = sarr_its(&sarr_split, &i)) != NULL;) {
            strarr_t sarr_tmp = {0};
            //: Expand braces again - recursive call:
            expand_braces(&sarr_tmp, str_split, false);
               HandleErr(handle_err);
            //: Add expansion to uptill now results:
            sarr_appA(sarr, &sarr_tmp);   //: append expansion to results
               HandleErr(handle_err);
            sarr_fin(&sarr_tmp);                      //: finish expansion result
         }
         sarr_fin(&sarr_split);
      }
      /*: No braces pairs and any single commas found: */
      else
      {
         expand_braces(sarr, str, true);
            HandleErr(handle_err);
      }
   }

  handle_err:
   return;
}




#endif /*- MHCL_BASE_SRCSTRINGS_STRINGOPS_AUX_H_INC -*/
