#ifndef MHCL_BASE_SRCSTRINGS_STRINGS_H
#define MHCL_BASE_SRCSTRINGS_STRINGS_H
/*!F::String structure management - external interface file.
 *
 *  ::Provides functions for other program modules.
 */




/* current project */
#include "common_mhlbase.h"

/* mh library */
#include "core.h"

/* libc */
#include <stdarg.h>     // va_*, vsnprintf()

/* current module */




/* Data. */
/* -------------------------------------------------------------------------- */

typedef struct str_t str_t;
/*!:: String, wrapper for c-string.
 * ::Note:  'nb' may be more then 'nc+1'
 */
struct str_t
{
   char* buf;     //! buffer for cstring
   size_t nb;     //! bytes allocated (may be more then nc+1)
   size_t nc;     //! chars upto and excluding \0
   size_t mnc;    //! multibyte chars
   int arb;       //! for arbitrary use by library user, library doesn't touch this at all
};
#define STR_T_Z (str_t){NULL,0,0,0,0}




/* Prepare and cleanup. */
/* -------------------------------------------------------------------------- */


/*!1:: Create string from given c-string. */
extern str_t* str_cre (const char* cstr)
               __attr_nonnull__(1)
               __attr_wures__
               ;

/*!1:: Create string from given string. */
extern str_t* str_cres (const str_t* str)
               __attr_nonnull__(1)
               __attr_wures__
               ;

/*!1:: Create string from given c-string. Supply length. */
extern str_t* str_cre2 (const char* cstr, size_t num_chars)
               __attr_nonnull__(1)
               __attr_wures__
               ;

/*!1:: Destroy string. */
extern void str_des (str_t** str)
               __attr_nonnull__(1)
               ;

/*!1:: Initialize string with given c-string. */
extern char* str_ini (str_t* str, const char* cstr)
               __attr_nonnull__(1,2)
               ;

/*!1:: Initialize string with given c-string. Supply length. */
extern char* str_ini2 (str_t* str, const char* cstr, size_t num_chars)
               __attr_nonnull__(1,2)
               ;

/*!1:: Initialize string from given printf-like format. */
extern char* str_inif (str_t* str, const char* fmt, ...)
               __attr_format__(__printf__,2,3)
               ;

/*!1:: Initialize string with given string. */
extern char* str_inis (str_t* str_dest, const str_t* str_src)
               __attr_nonnull__(1,2)
               ;

/*!1:: Finalize string. */
extern void str_fin (str_t* str)
               __attr_nonnull__(1)
               ;

/*!1:: Fill string with given c-string. */
extern void str_fill (str_t* str_dest, char* cstr_src)
               __attr_nonnull__(1,2)
               ;

/*!1:: Fill string with given c-string. Supply properties. */
extern void str_fill2 (str_t* str_dest, char* cstr_src, size_t nbytes,
      size_t multichars)
               __attr_nonnull__(1,2)
               ;

/*!1:: Fill string with given string. */
extern void str_fills (str_t* str_dest, str_t* str_src)
               __attr_nonnull__(1,2)
               ;

/*!1:: Empty string. */
extern void str_empty (str_t* str)
               __attr_nonnull__(1)
               ;




/* Manipulate. */
/* -------------------------------------------------------------------------- */


/*!1:: Replace string with given c-strign. */
extern char* str_repl (str_t* str_dest, const char* cstr)
               __attr_nonnull__(1,2)
               ;

/*!1:: Replace string with given c-strign. Supply properties. */
extern char* str_repl2 (str_t* str_dest, const char* cstr, size_t nb, size_t mnc)
               __attr_nonnull__(1,2)
               ;

/*!1:: Replace string with given string. */
extern char* str_repls (str_t* str_dest, const str_t* str_src)
               __attr_nonnull__(1,2)
               ;

/*!1:: Prepend given c-string to string. */
extern char* str_prep (str_t* str_dest, const char* cstr_src)
               __attr_nonnull__(1,2)
               ;

/*!1:: Append given c-string to string. */
extern char* str_app (str_t* str_dest, const char* cstr_src)
               __attr_nonnull__(1,2)
               ;

/*!1:: Append given string to string. */
extern char* str_apps (str_t* str_dest, const str_t* str_src)
               __attr_nonnull__(1,2)
               ;




/* Examine. */
/* -------------------------------------------------------------------------- */


/*!1:: Return pointer to internal c-string buffer. */
extern char* str_ret_buf (str_t* str)
               __attr_nonnull__(1)
               ;

/*!1:: Return true if string is empty.  */
extern bool str_is_empty (str_t* string)
               __attr_nonnull__(1)
               __attr_wures__
               __attr_const__
               ;

/*!1:: Return true if all properties of string are zero.  */
extern bool str_is_zero (str_t* string)
               __attr_nonnull__(1)
               __attr_wures__
               __attr_const__
               ;




/* Compare. */
/* -------------------------------------------------------------------------- */

/*!1:: Compare strings.  */
extern int str_cmp (const str_t* str1, const str_t* str2)
               __attr_nonnull__(1,2)
               __attr_wures__
               __attr_const__
               ;




/* Print. */
/* -------------------------------------------------------------------------- */


/*!1:: Print content of string to specified file stream. */
extern void str_print (FILE* filestream, str_t* str)
               __attr_nonnull__(1,2)
               ;




#ifdef DEBUG
/* Used only by tests in debug mode. */
bool strings_all_tests_run (void)
         __attr_wures__;
#endif




#endif /*- MHCL_BASE_SRCSTRINGS_STRINGS_H -*/
