/* This file is part of program distributed under GNU General Public
   License. For more information see README and LICENSE files. */

#ifndef MH_LIB_BASE_STRINGS_STRINGS_ARR_INTER_H_INC
#define MH_LIB_BASE_STRINGS_STRINGS_ARR_INTER_H_INC


/*!F::Arrays of strings management - internal interface file.
 *
 *  ::Provides functions for basic manipulations of module data structures.
 */


/* employs includes from strings.c */




/* Those functions poses NO overflow checks! */
/* Manipulate data. */
static inline void zero_ (strarr_t* sarr);
static inline void ass_ (strarr_t* sarr, str_t* buf, size_t ne, size_t ns);
static inline str_t* ret_ (const strarr_t* sarr, size_t idx);

/* Prepare. */
static inline strarr_t* constr_ (void);
static inline void destr_ (strarr_t* sarr);
static inline str_t* prep_ (strarr_t* sarr, uint ne);
static inline void clean_ (strarr_t* sarr);
static inline void rubout_ (str_t str_arr[], size_t idx, uint ne);

/* Resize. */
static inline str_t* infl_ (strarr_t* sarr, uint ne_add);
static inline str_t* defl_ (strarr_t* sarr, uint ne_sub);
static inline str_t* levl_ (strarr_t* sarr, uint ne_new);

/* Copy. */
static inline char* subst_ (strarr_t* sarr, size_t idx, const str_t* str);
static inline char* dupl_ (str_t* dest, const strarr_t* sarr, size_t idx);
static inline char* stick_ (strarr_t* sarr, size_t idx, const str_t* str);
static inline char* bring_ (str_t* dest, strarr_t* sarr, size_t idx);




/* set all struct data to \0 */
static inline
void
zero_ (
   strarr_t* sarr
){
   sarr->buf = NULL;
   sarr->na = 0;
   sarr->ns = 0;
   return;
}

/* assign all data */
static inline
void
ass_ (
   strarr_t* sarr,
   str_t* buf,
   size_t ne,
   size_t ns
){
   sarr->buf = buf;
   sarr->na = ne;
   sarr->ns = ns;
   return;
}

/* return pointer to specified string in array */
static inline
str_t*
ret_ (
   const strarr_t* sarr,
   size_t idx
){
   return &sarr->buf[idx];
}

/* alloc array struct and zero all data */
static inline
strarr_t*
constr_ (void)
{
   strarr_t* sarr_new = malloc(sizeof(strarr_t));
      if (sarr_new == NULL) { return NULL; }
   sarr_new->buf = NULL; sarr_new->na = 0; sarr_new->ns = 0;
   return sarr_new;
}

/* free array struct */
static inline
void
destr_ (
   strarr_t* sarr
){
   free(sarr);
   return;
}

/* alloc buffer for strings inside array struct */
static inline
str_t*
prep_ (
   strarr_t* sarr,
   uint ne
){
   size_t nb = ne * sizeof(str_t);     // overflow danger
   str_t* buf_new = malloc(nb);
      if (buf_new == NULL) { return NULL; }
   sarr->buf = buf_new;
   sarr->na = ne;
   sarr->ns = 0;
   return buf_new;
}

/* free strings buffer inside array struct */
static inline
void
clean_ (
   strarr_t* sarr
){
   free(sarr->buf);
   sarr->buf = NULL; sarr->na = 0; sarr->ns = 0;
   return;
}

/* Zero specified number of elements (string structs) in array struct buffer. */
static inline
void
rubout_ (
   str_t str_arr[],
   size_t idx,
   uint ne
){
   memset(&str_arr[idx], 0, ne * sizeof(str_t));
   return;
}

/* increase buffer inside array struct by `ne_add` strings */
static inline
str_t*
infl_ (
   strarr_t* sarr,
   uint ne_add
){
   size_t ne = sarr->na + ne_add;
   size_t nb = ne * sizeof(str_t);     // overflow danger
   str_t* buf_new = calloc(1, nb);
      if (buf_new == NULL) { return NULL; }
   memcpy(buf_new, sarr->buf, sarr->na*sizeof(str_t));
   free(sarr->buf);
   sarr->buf = buf_new;
   sarr->na = ne;
   return buf_new;
}

/* increase buffer inside array struct by `ne_prep` strings at begining */
static inline
str_t*
infl2_ (
   strarr_t* sarr,
   uint ne_prep
){
   size_t ne = sarr->na + ne_prep;
   size_t nb = ne * sizeof(str_t);     // overflow danger
   str_t* buf_new = calloc(1, nb);
      if (buf_new == NULL) { return NULL; }
   memcpy(&buf_new[1], sarr->buf, sarr->na*sizeof(str_t));
   free(sarr->buf);
   sarr->buf = buf_new;
   sarr->na = ne;
   return buf_new;
}

/* lessen buffer inside array struct by `ne_sub` strings */
static inline
str_t*
defl_ (
   strarr_t* sarr,
   uint ne_sub
){
   size_t ne = sarr->na - ne_sub;
   size_t nb = ne * sizeof(str_t);     // overflow danger
   str_t* buf_new = malloc(nb);
      if (buf_new == NULL) { return NULL; }
   memcpy(buf_new, sarr->buf, nb);
   free(sarr->buf);
   sarr->buf = buf_new;
   sarr->na = ne;
   return buf_new;
}

/* resize buffer inside array struct to be `ne_new` string elements large */
static inline
str_t*
levl_ (
   strarr_t* sarr,
   uint ne_new
){
   size_t nb = ne_new * sizeof(str_t);    // overflow danger
   str_t* buf_new = realloc(sarr->buf, nb);
      if (buf_new == NULL) { return NULL; }
   sarr->buf = buf_new;
   sarr->na = ne_new;
   return buf_new;
}

/* replace string in array at specified index with new string */
static inline
char*
subst_ (
   strarr_t* sarr,
   size_t idx,
   const str_t* str
){
   return str_repls(&sarr->buf[idx], str);
}

/* replace given string with string in array at specified index */
static inline
char*
dupl_ (
   str_t* dest,
   const strarr_t* sarr,
   size_t idx
){
   return str_repls(dest, &sarr->buf[idx]);
}

/* put string at specified index, move existing strings 1 element forward */
static inline
char*
stick_ (
   strarr_t* sarr,
   size_t idx,
   const str_t* str
){
   free(sarr->buf[sarr->na-1].buf);             //: free last string buffer
   for (size_t i = sarr->na-1; i > idx; --i) {
      //if (i >= sarr->na) {
         //if (sarr->buf[i].buf != NULL) free(sarr->buf[i].buf);
      //}
      str_fills(&sarr->buf[i], &sarr->buf[i-1]);//: move all strings by one el towards end
   }
   sarr->buf[idx].buf = NULL;                   //: assign NULL to avoid freeing this buffer
   return str_repls(&sarr->buf[idx], str);      //: place new string at freed index
}

/* pull string from specified index, move remaining strings 1 element backwards */
static inline
char*
bring_ (
   str_t* dest,
   strarr_t* sarr,
   size_t idx
){
   char* res = str_repls(dest, &sarr->buf[idx]);
   free(sarr->buf[idx].buf);
   for (size_t i = idx; i < sarr->na-1; ++i) {
      //sarr->buf[i] = sarr->buf[i+1];
      str_fills(&sarr->buf[i], &sarr->buf[i+1]);
   }
   str_empty(&sarr->buf[sarr->na-1]);
   return res;
}




#endif /*- MH_LIB_BASE_STRINGS_STRINGS_ARR_INTER_H_INC -*/
