#ifndef MHCL_BASE_SRCSTRINGS_STRINGS_INTER_H_INC
#define MHCL_BASE_SRCSTRINGS_STRINGS_INTER_H_INC


/*!F::String structure management - internal interface file.
 *
 *  ::Provides functions for basic manipulations of this module data structures.
 */


/* employs includes from strings.c */




/* Manipulate data. */
static inline void zero_ (str_t* str);
static inline void ass_ (str_t* str, char* buf, size_t nb, size_t nc, size_t mnc);
static inline void ass_ncs_ (str_t* str, size_t nc, size_t mnc);
static inline char* ret_ (str_t* str, size_t idx);

/* Prepare. */
static inline str_t* constr_ (void);
static inline void destr_ (str_t* str);
static inline char* prep_ (str_t* str, size_t nb);
static inline void clean_ (str_t* str);

/* Resize. */
static inline char* infl_ (str_t* str, uint nb_add);
static inline char* defl_ (str_t* str, uint nb_sub);
static inline char* levl_ (str_t* str, uint nb_new);

/* Copy. */
static inline void overw_ (str_t* str, size_t idx, uint nb, const char* cstr);
static inline void read_ (char* dest, const str_t* str, size_t idx, uint nb);
static inline void put_ (str_t* str, size_t idx, uint nb, const char* cstr);
static inline void pull_ (char* dest, str_t* str, size_t idx, uint nb);




/* set all struct data to \0 */
static inline
void
zero_ (
   str_t* str
){
   str->buf = NULL;
   str->nb = 0;
   str->nc = 0;
   str->mnc = 0;
   return;
}

/* assign all data */
static inline
void
ass_ (
   str_t* str,
   char* buf,
   size_t nb,
   size_t nc,
   size_t mnc
){
   str->buf = buf;
   str->nb = nb;
   str->nc = nc;
   str->mnc = mnc;
   return;
}

/* assign specific data */
static inline
void
ass_ncs_ (
   str_t* str,
   size_t nc,
   size_t mnc
){
   str->nc = nc;
   str->mnc = mnc;
}

/* return pointer to specified byte in buffer */
static inline
char*
ret_ (
   str_t* str,
   size_t idx
){
   return &str->buf[idx];
}

/* alloc str_t struct and zero all data */
static inline
str_t*
constr_ (void)
{
   str_t* str_new = malloc(sizeof(str_t));
      if (str_new == NULL) { return NULL; }
   str_new->buf = NULL; str_new->nb = 0; str_new->nc = 0; str_new->mnc = 0;
   return str_new;
}

/* free str_t struct */
static inline
void
destr_ (
   str_t* str
){
   free(str);
   return;
}

/* alloc buffer inside str_t */
static inline
char*
prep_ (
   str_t* str,
   size_t nb
){
   char* buf_new = malloc(nb);
      if (buf_new == NULL) { return NULL; }
   str->buf = buf_new;
   str->nb = nb;
   str->nc = 0;
   str->mnc = 0;
   return buf_new;
}

/* free buffer inside str_t */
static inline
void
clean_ (
   str_t* str
){
   free(str->buf);
   str->buf = NULL; str->nb = 0; str->nc = 0; str->mnc = 0;
   return;
}

/* increase buffer inside str_t by `nb_add` bytes */
static inline
char*
infl_ (
   str_t* str,
   uint nb_add
){
   size_t nb = str->nb + nb_add;             // overflow danger
   char* buf_new = realloc(str->buf, nb);
      if (buf_new == NULL) { return NULL; }
   str->buf = buf_new;
   str->nb = nb;
   return buf_new;
}

/* lessen buffer inside str_t by `nb_add` bytes */
static inline
char*
defl_ (
   str_t* str,
   uint nb_sub
){
   size_t nb = str->nb - nb_sub;
   char* buf_new = realloc(str->buf, nb);
      if (buf_new == NULL) { return NULL; }
   str->buf = buf_new;
   str->nb = nb;
   return buf_new;
}

/* resize buffer inside str_t to be `nb_add` bytes large */
static inline
char*
levl_ (
   str_t* str,
   uint nb_new
){
   char* buf_new = realloc(str->buf, nb_new);
      if (buf_new == NULL) { return NULL; }
   str->buf = buf_new;
   str->nb = nb_new;
   return buf_new;
}

/* copy `nb` from `cs` to `str` at index `idx`. */
static inline
void
overw_ (
   str_t* str,
   size_t idx,
   uint nb,
   const char* cs
){
   memcpy(&str->buf[idx], cs, nb);
   return;
}

/* copy `nb` from `str` at index `idx` to `cs`. */
static inline
void
read_ (
   char* cs,
   const str_t* str,
   size_t idx,
   uint nb
){
   memcpy(cs, &str->buf[idx], nb);
   return;
}

/* Insert `nb` from `cs` to `str` at index `idx`, moving following chars
   further. */
static inline
void
put_ (
   str_t* str,
   size_t idx,
   uint nb,
   const char* cs
){
   for (size_t i = str->nb-nb; i > idx; --i) {
      str->buf[i+nb-1] = str->buf[i-1];
   }
   memcpy(&str->buf[idx], cs, nb);
   return;
}

/* Copy `nb` from `str` at index `idx` to `cs` and pull out copied chars
   from source (bring following chars to begining). */
static inline
void
pull_ (
   char* cs,
   str_t* str,
   size_t idx,
   uint len
){
   memcpy(cs, &str->buf[idx], len);
   for (size_t i = idx; i+len < str->nb; ++i) {
      str->buf[i] = str->buf[i+len];
   }
   return;
}




#endif /*- MHCL_BASE_SRCSTRINGS_STRINGS_INTER_H_INC -*/
