#ifndef MHCL_FS_SRCDIRS_DS_INFO_H
#define MHCL_FS_SRCDIRS_DS_INFO_H


/*!F::Directories information retrieve - external interface file.
 *
 *  ::Provides functions for other program modules.
 */




/* current project */
#include "common_mhlfs.h"
#include "fs_storage.h"

/* mh library */
#include "core.h"
#include "errors.h"
#include "debug.h"
#include "strings.h"
#include "strings_arr.h"

/* libc */
#include <dirent.h>     // opendir(), DIR, struct dirent
#include <sys/stat.h>   // stat(), struct stat




/* Entry. */
/* -------------------------------------------------------------------------- */


/*!1:: Create entry structure. */
extern st_ent* ds_ent_alloc (const DIR* dirstream)
                        __attr_nonnull__(1)
                        __attr_wures__
                        __attr_malloc__
                        ;

/*!1:: Destroy entry structure. */
extern void ds_ent_free (st_ent** ptr_dir_entry)
            __attr_nonnull__(1)
            ;

/*!1:: Read next entry from directory stream. */
extern bool ds_read_entry
      (st_ent* dir_entry, DIR* dirstream, const char* dirname)
            __attr_nonnull__(1,2,3)
            __attr_wures__
            ;




/* Directory. */
/* -------------------------------------------------------------------------- */


/*!1:: Create directory structure. */
extern st_dir* ds_dir_create (const char* dirname)
                        __attr_nonnull__(1)
                        __attr_wures__
                        ;

/*!1:: Destroy directory structure. */
extern void ds_dir_destroy (st_dir** stDirectory)
               __attr_nonnull__(1)
               ;

/*!1:: Read entries for given directory into directory struct and read each
 *       entry's stats if requested. */
extern void ds_read_directory (st_dir* stDirectory, bool read_stats)
               __attr_nonnull__(1)
               ;

/*!1:: Read entries for given directory into directory struct. */
extern void ds_read_directory_stats (st_dir* stDirectory)
               __attr_nonnull__(1)
               ;

/*!1:: Return number of entries in specified directory. */
extern size_t ds_count_dir_entries (const char* dirname)
               __attr_nonnull__(1)
               __attr_wures__
               ;




/* Miscelaneous. */
/* -------------------------------------------------------------------------- */


/*!1:: !!! not finished not tested !!! */
extern void ds_wordexpand_name (strarr_t* sarr_mat, const char* pattern)
               ;

/*!1:: Perform globbing on given pattern. */
extern void ds_glob_name (strarr_t* strarr_matches, strarr_t* strarr_errors,
         const str_t* pattern)
                  __attr_nonnull__(1,2,3)
                  ;




#ifdef DEBUG
/* Used only by tests in debug mode. */
extern bool dsinf_all_tests_run (void)
      __attr_wures__;
#endif




#endif /*- MHCL_FS_SRCDIRS_DS_INFO_H -*/
