#ifndef MHCL_FS_SRCDIRS_DS_OPERATE_H
#define MHCL_FS_SRCDIRS_DS_OPERATE_H


/*!F::Operations on directories - external interface file.
 *
 *  ::Provides functions for other program modules.
 */




/* current project */
#include "common_mhlfs.h"
#include "fs_storage.h"

/* mh library */
#include "core.h"
#include "errors.h"
#include "debug.h"

/* libc */
#include <sys/stat.h>   // struct stat




/* Make. */
/* -------------------------------------------------------------------------- */


/*!1:: Make file and return fd. */
extern FD ds_mkfile (const char* filename, mode_t mode)
               __attr_nonnull__(1)
               ;

/*!1:: Make file, fd is closed. */
extern void ds_mkfile2 (const char* filename, mode_t mode)
               __attr_nonnull__(1)
               ;

/*!1:: Make directory. */
extern void ds_mkdir (const char* dirname, mode_t mode)
               __attr_nonnull__(1)
               ;

/*!1:: Make hard link. */
extern void ds_mkhlink (const char* link_name, const char* target_name)
               __attr_nonnull__(1,2)
               ;

/*!1:: Make symbolic link. */
extern void ds_mkslink (const char* link_name, const char* target_name)
               __attr_nonnull__(1,2)
               ;




/* Remove. */
/* -------------------------------------------------------------------------- */


/*!1:: Remove directory. */
extern void ds_rmdir (const char* dirname)
               __attr_nonnull__(1)
               ;

/*!1:: Remove directory, supply attributes. */
extern void ds_rmdir2 (const char* dirname, const struct stat* dir_attrs_orig)
               __attr_nonnull__(1)
               ;

/*!1:: Remove file. */
extern void ds_rmfile (const char* filename)
               __attr_nonnull__(1)
               ;

/*!1:: Remove file, supply attributes. */
extern void ds_rmfile2 (const char* filename, const struct stat* file_attrs_orig)
               __attr_nonnull__(1)
               ;

/*!1:: Remove entry. */
extern void ds_rmentry (const char* entry_name)
               __attr_nonnull__(1)
               ;

/*!1:: Remove entry, supply attributes. */
extern void ds_rmentry2 (const char* entry_name, const struct stat* entry_attrs_orig)
               __attr_nonnull__(1)
               ;




/* Copy/Remove. */
/* -------------------------------------------------------------------------- */


/*!1:: Copy specified file to other file,. */
extern void ds_copy (const char* fname_src, const char* fname_dest, int flags)
               __attr_nonnull__(1,2)
               ;

/*!1:: Copy specified file to other file, supply attributes. */
extern void ds_copy2 (const char* fname_src, const char* fname_dest,
         struct stat* attrs_src, const struct stat* attrs_dest, int flags)
               __attr_nonnull__(1,2)
               ;

/*!1:: Move specified file to other name or to specified directory. */
extern void ds_move (const char* fname_old, const char* fname_new, int flags)
               __attr_nonnull__(1,2)
               ;

/*!1:: Move specified file to other name or to specified directory, supply attributes. */
extern void ds_move2 (const char* fname_old, const char* fname_new,
         struct stat* attrs_old, const struct stat* attrs_new, int flags)
               __attr_nonnull__(1,2)
               ;




#ifdef DEBUG
/* Used only by tests in debug mode. */
extern bool dsoperate_all_tests_run (void)
      __attr_wures__;
#endif




#endif /*- MHCL_FS_SRCDIRS_DS_OPERATE_H -*/
