/*!F::File stream basic functions - main implementation file.
 *
 *  ::Purpose:
 *    Basic file stream management - open, close etc.
 *
 *  ::Goals:
 *    - thread-safe, reentrant
 *    - buffer/integer overflow safe
 *    - propagate or handle errors with routines from "errors.h"
 *
 *  ::Functionality:
 *
 *  ::Limitations:
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    File:             This file is part of program distributed under GNU General Public License.
 *                      For terms of distribution, warranty and licensing see README file.
 *    Created:          2014-06-02
 *    (c) Copyright:    Maciej Helminiak 2014
 *    License:          GNU General Public License, version 3
 *                      Copy of the license is distributed along with this program
 *                      or can be seen at <http://www.gnu.org/licenses/>
 */




/* current project */
#include "common_mhlfs.h"

/* mh library */
#include "core.h"
#include "errors.h"
#include "debug.h"
//#include "memory.h"

/* libc */

/* current module */
#include "fstream_basic.h"




/* Mode - see man fopen */
void
fs_open (
   FILE** pfs,
   const char* fname,
   const char* mode
){
   ___Dassert(pfs, "'pfs' parameter invalid");
   ___Dassert(fname, "'fname' parameter invalid");
   ___Dassert(mode, "'mode' parameter invalid");

   FILE* fs = fopen(fname, mode);
      ErrorTrigIf(errno != 0, HANDLE_ERR);

   *pfs = fs;

   return;

  HANDLE_ERR:
   return;
}

void
fs_close (
   FILE** pfs
){
   ___Dassert(pfs, "'pfs' parameter invalid");
   ___Dassert(*pfs, "'*pfs' parameter invalid");

   fclose(*pfs);
      ErrorTrigIf(errno != 0, HANDLE_ERR);
   *pfs = NULL;

   return;

  HANDLE_ERR:
   return;
}
