/* This file is part of program distributed under GNU General Public
   License. For more information see README and LICENSE files. */

#ifndef MHCL_FS_SRCSTATS_STAT_CHANGE_AUX_H_INC
#define MHCL_FS_SRCSTATS_STAT_CHANGE_AUX_H_INC


/*!F::File statistics change - auxillary file.
 *
 *  ::Provides helper functions for main implementation file.
 */




/* employs includes from stat_change.c */




/* :: Create file mode of type 'mode_t' from given c-string and return it.
 *    C-string should be of format "rwxrwxrwx" and char after last x must
 *    be '\0'.
 * ::Preconditions:
 * ::Errors:
 *    0 returned.
 *    EINVAL         -- one of chars in c-string has wrong value
*/
static inline
mode_t                  //! mode
compute_mode_ (
   const char* cs_mode  //! c-string with mode characters
){
   ___Dassert(cs_mode, "'cs_mode' parameter invalid");

   if (cs_mode[9] != '\0') { ErrorSetIf(true, EINVAL, handle_err); };

   mode_t mode = 0;

   //USER
   if (cs_mode[0] == 'r') { mode |= S_IRUSR; }
   else if (cs_mode[0] != '-') { ErrorSetIf(true, EINVAL, handle_err); }

   if (cs_mode[1] == 'w') { mode |= S_IWUSR; }
   else if (cs_mode[1] != '-') { ErrorSetIf(true, EINVAL, handle_err); }

   if (cs_mode[2] == 'x') { mode |= S_IXUSR; }
   else if (cs_mode[2] == 's') { mode |= S_IXUSR; mode |= S_ISUID; }
   else if (cs_mode[2] == 'S') { mode |= S_ISUID; }
   else if (cs_mode[2] != '-') { ErrorSetIf(true, EINVAL, handle_err); }

   //GROUP
   if (cs_mode[3] == 'r') { mode |= S_IRGRP; }
   else if (cs_mode[3] != '-') { ErrorSetIf(true, EINVAL, handle_err); }

   if (cs_mode[4] == 'w') { mode |= S_IWGRP; }
   else if (cs_mode[4] != '-') { ErrorSetIf(true, EINVAL, handle_err); }

   if (cs_mode[5] == 'x') { mode |= S_IXGRP; }
   else if (cs_mode[5] == 's') { mode |= S_IXGRP; mode |= S_ISGID; }
   else if (cs_mode[5] == 'S') { mode |= S_ISGID; }
   else if (cs_mode[5] != '-') { ErrorSetIf(true, EINVAL, handle_err); }

   //OTHERS
   if (cs_mode[6] == 'r') { mode |= S_IROTH; }
   else if (cs_mode[6] != '-') { ErrorSetIf(true, EINVAL, handle_err); }

   if (cs_mode[7] == 'w') { mode |= S_IWOTH; }
   else if (cs_mode[7] != '-') { ErrorSetIf(true, EINVAL, handle_err); }

#if defined(BSD)
   if (cs_mode[8] == 'x') { mode |= S_IXOTH; }
   else if (cs_mode[8] == 't') { mode |= S_IXGRP; mode |= S_ISVTX; }
   else if (cs_mode[8] == 'T') { mode |= S_ISVTX; }
   else if (cs_mode[8] != '-') { ErrorSetIf(true, EINVAL, handle_err); }
#else
   if (cs_mode[8] == 'x') { mode |= S_IXOTH; }
   else if (cs_mode[8] != '-') { ErrorSetIf(true, EINVAL, handle_err); }
#endif

   return mode;

  handle_err:
   return 0;
}



#endif /*- MHCL_FS_SRCSTATS_STAT_CHANGE_AUX_H_INC -*/
