#ifndef MHCL_FS_SRCSTATS_STAT_INFO_H
#define MHCL_FS_SRCSTATS_STAT_INFO_H


/*!F::File statistics retrieve - external interface file.
 *
 *  ::Provides functions for other program modules.
 */




#include "core.h"
#include "errors.h"
#include "debug.h"

#include "common_mhlfs.h"
#include "dstream_info.h"

#include <sys/types.h>  // mode_t




/* Test file. */
/* -------------------------------------------------------------------------- */


/*!1:: Check if specified file exist. */
extern bool stat_file_exists (const char* file)
               __attr_nonnull__(1)
               __attr_wures__
               ;

/*!1:: Check if specified file exist. Resolve symlinks. */
extern bool stat_file_exists2 (const char* file)
               __attr_nonnull__(1)
               __attr_wures__
               ;

/*!1:: Check if specified entry is directory. */
extern bool stat_file_is_dir (const char* file)
               __attr_nonnull__(1)
               __attr_wures__
               ;

/*!1:: Check if specified entry is directory. */
extern bool stat_fs_is_dir (FILE* fs)
               __attr_nonnull__(1)
               __attr_wures__
               ;

/*!1:: Check if specified files are same. */
extern bool stat_files_are_same (const char* file1, const char* file2)
               __attr_nonnull__(1)
               __attr_wures__
               ;




/* File name modification. */
/* -------------------------------------------------------------------------- */

/*!1:: Remove . and .. from given path. */
extern void stat_replacedots (char** new_path, const char* path)
               __attr_nonnull__(1,2)
               ;

/*!1:: Remove all trailing '/'. */
extern void stat_remove_endslashess (str_t* name)
               __attr_nonnull__(1)
               ;




/* Symlinks. */
/* -------------------------------------------------------------------------- */


/*!1:: Read target name of symbolic link. */
extern void stat_readlink (char** ptr_target, uint tar_presumable_len,
                              const char* slink_name)
               __attr_nonnull__(1,3)
               ;

/*!1:: Make given path absolute. */
extern void stat_absolutepath (char** absol_path, const char* file)
               __attr_nonnull__(1,2)
               ;




/* Attributes. */
/* -------------------------------------------------------------------------- */


/*!1:: Read file attributes of given entry. */
extern void stat_read_attrs (st_ent* stEntry)
               __attr_nonnull__(1)
               ;

/*!1:: Return type of specified entry. */
extern char stat_file_type_is (const char* fname, bool deref_sl)
               __attr_nonnull__(1)
               __attr_wures__
               ;

/*!1:: Extract file type from file mode. */
extern char stat_extract_ftype (char* ptr_char, mode_t file_mode)
               __attr_nonnull__(1)
               ;

/*!1:: Extract file permissions from file mode. */
extern char* stat_extract_perms (char (*permissions)[10], mode_t file_mode)
               __attr_nonnull__(1)
               ;




/* File owners. */
/* -------------------------------------------------------------------------- */


/*!1:: Find user name for given uid. */
extern char* stat_look_user (char** ptr_cstring_user, uid_t user_uid)
               __attr_nonnull__(1)
               ;

/*!1:: Find group name for given gid. */
extern char* stat_look_group (char** ptr_cstring_group, gid_t group_gid)
               __attr_nonnull__(1)
               ;

/*!1:: Return uid for given user name. */
extern uid_t stat_look_uid (const char* cs_user)
               __attr_nonnull__(1)
               __attr_wures__
               ;

/*!1:: Return gid for given group name. */
extern gid_t stat_look_gid (const char* cs_group)
               __attr_nonnull__(1)
               __attr_wures__
               ;




#ifdef DEBUG
/* Used only by tests in debug mode. */
extern bool statinfo_all_tests_run (void)
      __attr_wures__;
#endif




#endif /*- MHCL_FS_SRCSTATS_STAT_INFO_H -*/
