/* This file is part of program distributed under GNU General Public
   License. For more information see README and LICENSE files. */

#ifndef MHCL_PRCONF_SRCPARS_PRCONFPARS_AUX_H_INC
#define MHCL_PRCONF_SRCPARS_PRCONFPARS_AUX_H_INC


/*!F::Command-line arguments parser - auxillary file.
 *
 *  ::Provides helper functions for main implementation file.
 */




/* employs includes from cl_args_parser.c */




static void parse_argument (st_specs_arr* stArgSpecs, const char* arg, st_faults_arr* Faults);
static void deduce_arg_charact (st_arg_ch* arg_ch, const char* arg);
static void cleanup_arg_charact (st_arg_ch* arg_ch);
static void set_arg_in_spec (st_specs_arr* stArgSpecs, st_arg_ch* arg_ch, st_faults_arr* Faults);
static st_opt_spec* find_option_in_specs (const str_t* name, arr_ptrs_t* stOperSpecs, size_t bfrombeg);
static void validate_and_append_value_to_operand (st_oper_spec* stOperSpec, str_t* val, st_faults_arr* Faults);
static void validate_and_overw_values_in_option (st_opt_spec* stOptSpec, strarr_t* vals, st_faults_arr* Faults);
static void validate_and_append_values_to_option (st_opt_spec* stOptSpec, strarr_t* vals, st_faults_arr* Faults);
static void option_on (st_opt_spec* stOptSpec);
static void option_toggle (st_opt_spec* stOptSpec);
static void option_lvl_up (st_opt_spec* stOptSpec);




/* Create arg charact. based on given argument and then find and set option or
   operand related to this argument, set Faults in option not found or other
   violations.
 */
static
void
parse_argument (
   st_specs_arr* stArgSpecs,  //! [out] arguments specification
   const char* arg,           //! argument to parse
   st_faults_arr* Faults      //! [out] faults detected during parsing
){
   ___DParam(stArgSpecs, != NULL);
   ___DParam(arg, != NULL);
   ___DParam(Faults, != NULL);

   st_arg_ch arg_ch = ST_ARG_CH_Z;

   /*: Arg charact. */
   deduce_arg_charact(&arg_ch, arg);
      HandleErr(HANDLE_ERR);

   if (arg_ch.kind == Argkind_err) {
      faultsarr_add_fault(Faults, Faultcath_exit, "invalid option format: '%s'", arg);
         HandleErr(HANDLE_ERR);
      goto HANDLE_SUCC;
   }

   /*: Set arg in spec. */
   set_arg_in_spec(stArgSpecs, &arg_ch, Faults);
      HandleErr(HANDLE_ERR);

   /*: Finish: */
  HANDLE_SUCC:
   cleanup_arg_charact(&arg_ch);
   return;

  HANDLE_ERR:
   cleanup_arg_charact(&arg_ch);
   return;
}

static
void
deduce_arg_charact (
   st_arg_ch* arg_ch,   //! [out] argument characteristics
   const char* arg      //! argument
){
   ___DParam(arg_ch, != NULL);
   ___DParam(arg, != NULL);

   arg_ch->clarg = arg;
   const char* pa = arg;   //pointer to current char inside arg

   /*: Deduce argument kind */
   if (pa[0] != '-') {
      arg_ch->kind = Argkind_oper;
   } else if (pa[1] != '-') {
      // -n*+     *-any but not \0 or =
      // -*+      +-only \0 or =
      if (pa[1] == 'n' && pa[2] != '\0' && pa[2] != '=' && (pa[3] == '=' || pa[3] == '\0')) {
         arg_ch->kind = Argkind_short_off;
         pa = &pa[2];
      } else if (pa[1] != '\0' && pa[1] != '=' && (pa[2] == '=' || pa[2] == '\0')) {
         arg_ch->kind = Argkind_short;
         pa = &pa[1];
      } else {
         arg_ch->kind = Argkind_err;
         goto HANDLE_SUCC;
      }
   } else if (pa[2] != '-') {
      // --no-*   *-any but not \0 or =
      // --*
      if (pa[2] == 'n' && pa[3] == 'o' && pa[4] == '-') {
         arg_ch->kind = Argkind_long_off;
         pa = &pa[5];
      } else {
         arg_ch->kind = Argkind_long;
         pa = &pa[2];
      }
   } else {
      arg_ch->kind = Argkind_err;
      goto HANDLE_SUCC;
   }

   /*: Deduce argument name and values */
   if (arg_ch->kind == Argkind_oper) {
      //: operand
      str_ini(&arg_ch->name, pa);
         HandleErr(HANDLE_ERR);
   } else {
      //: option
      char* ac = strchr(pa, '=');
      if (ac == NULL) {
         //: no values in arg
         str_ini(&arg_ch->name, pa);
            HandleErr(HANDLE_ERR);
      } else {
         //: values found
         *ac = '\0';
         str_ini(&arg_ch->name, pa);
         *ac = '=';
            HandleErr(HANDLE_ERR);
         pa = ac + 1;   //move behind '='
         sarr_split_line (&arg_ch->vals, pa, ",");
            HandleErr(HANDLE_ERR);
      }
   }

   /*: Finish: */
  HANDLE_SUCC:
   return;

  HANDLE_ERR:
   str_fin(&arg_ch->name);
   sarr_fin(&arg_ch->vals);
   return;
}

static
void
cleanup_arg_charact (
   st_arg_ch* arg_ch    //! [out] argument characteristics
){
   ___DParam(arg_ch, != NULL);
   arg_ch->kind = Argkind_none;
   str_fin(&arg_ch->name);
   sarr_fin(&arg_ch->vals);
   return;
}

static
void
set_arg_in_spec (
   st_specs_arr* stArgSpecs,  //! [out] arguments specification
   st_arg_ch* arg_ch,         //! argument characteristics
   st_faults_arr* Faults      //! [out] faults detected during parsing
){
   ___DParam(stArgSpecs, != NULL);
   ___DParam(arg_ch, != NULL);
   ___DParam(Faults, != NULL);

   // offset of name in option spec, short for short opts, long for long
   size_t bfrombeg = offsetof(st_opt_spec, sname);    //default is short name

   /*: Take actions dependant upon kind of argument: */
   switch (arg_ch->kind)
   {
   case Argkind_oper:
   {
      //: Get first operand specification still having room for one value:
      bool appended = false;
      for (uint i = 0; i < stArgSpecs->oper_specs.ns; ++i)
      {
         st_oper_spec* el_oper = arrptrs_el_ret(&stArgSpecs->oper_specs, i);
         if (el_oper == NULL) { continue; }
         if (el_oper->vn_max < 0 || (size_t)el_oper->vn_max > el_oper->vals.ns) {
            validate_and_append_value_to_operand(el_oper, &arg_ch->name, Faults);
               HandleErr(HANDLE_ERR);
            appended = true;
            break;
         }
      }
      //: No room for more vlaues in specs
      if (appended == false) {
         faultsarr_add_fault(Faults, Faultcath_exit, "too many operands");
            HandleErr(HANDLE_ERR);
      }
      break;
   }
   case Argkind_long:
      bfrombeg = offsetof(st_opt_spec, lname);
   case Argkind_short:
   {
      //: Find option in specifications:
      st_opt_spec* opt_spec = find_option_in_specs(&arg_ch->name, &stArgSpecs->opt_specs, bfrombeg);
      if (opt_spec == NULL) {
         faultsarr_add_fault(Faults, Faultcath_exit, "option '%s' not recognized", arg_ch->clarg);
            HandleErr(HANDLE_ERR);
         break;
      }

      //: Set specification
      if (arg_ch->vals.ns > 0)
      {
         switch (opt_spec->beh) {
         case Behopt_on:
            validate_and_overw_values_in_option(opt_spec, &arg_ch->vals, Faults);
            break;
         case Behopt_toggle:
            validate_and_overw_values_in_option(opt_spec, &arg_ch->vals, Faults);
            break;
         case Behopt_level:
            validate_and_append_values_to_option(opt_spec, &arg_ch->vals, Faults);
            break;
         case Behopt_none:
         case Behopt_count:
         default:
            ___Sentinel("wrong option behavior passed to %s", __func__); abort();
         }
               HandleErr(HANDLE_ERR);
         option_on(opt_spec);
      }
      else
      {
         switch (opt_spec->beh) {
         case Behopt_on:
            option_on(opt_spec);
            break;
         case Behopt_toggle:
            option_toggle(opt_spec);
            break;
         case Behopt_level:
            option_lvl_up(opt_spec);
            break;
         case Behopt_none:
         case Behopt_count:
         default:
            ___Sentinel("wrong option behavior passed to %s", __func__); abort();
         }
      }

      break;
   }
   case Argkind_long_off:
      bfrombeg = offsetof(st_opt_spec, lname);
   case Argkind_short_off:
   {
      //: Values are forbidden when switching option off
      if (arg_ch->vals.ns > 0) {
         faultsarr_add_fault(Faults, Faultcath_exit, "disable option (%s) does not take values", arg_ch->name.buf);
            HandleErr(HANDLE_ERR);
         break;
      }
      //: Find option in specifications:
      st_opt_spec* opt_spec = find_option_in_specs(&arg_ch->name, &stArgSpecs->opt_specs, bfrombeg);
      if (opt_spec == NULL) {
         faultsarr_add_fault(Faults, Faultcath_exit, "option '%s' not recognized", arg_ch->clarg);
            HandleErr(HANDLE_ERR);
         break;
      }
      //: Switch option off
      opt_spec->lvl = 0;
      sarr_fin(&opt_spec->vals);

      break;
   }
   case Argkind_none:
   case Argkind_err:
   case argkind_count:
   default:
      ___Sentinel("wrong argument kind passed to %s", __func__); abort();
   }

   return;

  HANDLE_ERR:
   return;
}

static
st_opt_spec*
find_option_in_specs (
   const str_t* name,
   arr_ptrs_t* stOptSpecs,
   size_t bfrombeg
){
   ___DParam(name, != NULL);
   ___DParam(stOptSpecs, != NULL);
   ___DParam(bfrombeg, > 0);

   for (uint i = 0; i < stOptSpecs->ns; ++i)
   {
      st_opt_spec* el_opt = arrptrs_el_ret(stOptSpecs, i);
      if (el_opt == NULL) { continue; }
      if (str_cmp(name, (str_t*)((char*)el_opt + bfrombeg)) == 0) {
         return el_opt;
      }
   }
   return NULL;
}

static
void
validate_and_append_value_to_operand (
   st_oper_spec* stOperSpec,
   str_t* val,
   st_faults_arr* Faults
){
   ___DParam(stOperSpec, != NULL);
   ___DParam(val, != NULL);
   ___DParam(Faults, != NULL);

   //: validate
   if (stOperSpec->p_vv != NULL) {
      ssize_t res =  sarr_cmps(stOperSpec->p_vv, val);
      if (res == -1) {
         faultsarr_add_fault(Faults, Faultcath_exit, "not valid operand value: '%s'", val->buf);
            HandleErr(HANDLE_ERR);
      }
   }
   //: append
   sarr_apps(&stOperSpec->vals, val);
      HandleErr(HANDLE_ERR);

   return;

  HANDLE_ERR:
   return;
}

static
void
validate_and_overw_values_in_option (
   st_opt_spec* stOptSpec,
   strarr_t* vals,
   st_faults_arr* Faults
){
   ___DParam(stOptSpec, != NULL);
   ___DParam(vals, != NULL);
   ___DParam(Faults, != NULL);

   sarr_fin(&stOptSpec->vals);

   const str_t* el_val = NULL; uint ii = 0;
   while ((el_val = sarr_its(vals, &ii)))
   {
      //: validate
      if (stOptSpec->p_vv != NULL) {
         ssize_t res =  sarr_cmps(stOptSpec->p_vv, el_val);
         if (res == -1) {
            faultsarr_add_fault(Faults, Faultcath_exit, "not valid option value: '%s'", el_val->buf);
               HandleErr(HANDLE_ERR);
         }
      }
      //: append
      sarr_apps(&stOptSpec->vals, el_val);
         HandleErr(HANDLE_ERR);
   }

   return;

  HANDLE_ERR:
   return;
}

static
void
validate_and_append_values_to_option (
   st_opt_spec* stOptSpec,
   strarr_t* vals,
   st_faults_arr* Faults
){
   ___DParam(stOptSpec, != NULL);
   ___DParam(vals, != NULL);
   ___DParam(Faults, != NULL);

   const str_t* el_val = NULL; uint ii = 0;
   while ((el_val = sarr_its(vals, &ii)))
   {
      //: validate
      if (stOptSpec->p_vv != NULL) {
         ssize_t res =  sarr_cmps(stOptSpec->p_vv, el_val);
         if (res == -1) {
            faultsarr_add_fault(Faults, Faultcath_exit, "not valid option value: '%s'", el_val->buf);
               HandleErr(HANDLE_ERR);
         }
      }
      //: append
      sarr_apps(&stOptSpec->vals, el_val);
         HandleErr(HANDLE_ERR);
   }

   return;

  HANDLE_ERR:
   return;
}

static
void
option_on (
   st_opt_spec* stOptSpec
){
   ___DParam(stOptSpec, != NULL);
   stOptSpec->lvl = 1;
}

static
void
option_toggle (
   st_opt_spec* stOptSpec
){
   ___DParam(stOptSpec, != NULL);
   stOptSpec->lvl = !stOptSpec->lvl;
}

static
void
option_lvl_up (
   st_opt_spec* stOptSpec
){
   ___DParam(stOptSpec, != NULL);
   ++stOptSpec->lvl;
}

static
bool
mandatory_and_not_set (
   st_opt_spec* stOptSpec,
   st_oper_spec* stOperSpec
){
   if (stOptSpec != NULL) {
      if (stOptSpec->mand == true && stOptSpec->lvl == 0) {
         return true;
      }
   }
   if (stOperSpec != NULL) {
      if (stOperSpec->mand == true && stOperSpec->vals.ns == 0) {
         return true;
      }
   }
   return false;
}

//too many: +1, too few -1, ok 0
static
int
num_of_vals_check (
   st_opt_spec* stOptSpec,
   st_oper_spec* stOperSpec
){
   if (stOptSpec != NULL && stOptSpec->lvl > 0) {
      if ((size_t)stOptSpec->vn_min > stOptSpec->vals.ns) {
         return -1;
      }
      if (stOptSpec->vn_max >= 0 && (size_t)stOptSpec->vn_max < stOptSpec->vals.ns) {
         return 1;
      }
   }
   if (stOperSpec != NULL && stOperSpec->vals.ns > 0) {
      if ((size_t)stOperSpec->vn_min > stOperSpec->vals.ns) {
         return -1;
      }
      if (stOperSpec->vn_max >= 0 && (size_t)stOperSpec->vn_max < stOperSpec->vals.ns) {
         return 1;
      }
   }
   return 0;
}







#endif /*- MHCL_PRCONF_SRCPARS_PRCONFPARS_AUX_H_INC -*/
