/* This file is part of program distributed under GNU General Public
   License. For more information see README and LICENSE files. */

#ifndef MHCL_CLAP_SRCSPEC_CLARGSTOR_H_INC
#define MHCL_CLAP_SRCSPEC_CLARGSTOR_H_INC


/*!F::Command-line arguments data - shared include file.
 *
 *  ::Purpose:
 *    Data structures holding specification of command-line arguments.
 *
 *  ::Goals:
 *    - thread-safe, reentrant, portable (POSIX systems)
 *    - buffer/integer overflow safe
 *    - propagate or handle errors with routines from "errors.h"
 *
 *  ::Functionality:
 *    - storage for cl-arg specification and array of it
 *
 *  ::Limitations:
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    File:             This file is part of program distributed under GNU General Public License.
 *                      For terms of distribution, warranty and licensing see README file.
 *    Created:          2014-01-12
 *    (c) Copyright:    Maciej Helminiak 2014
 *    License:          GNU General Public License, version 3
 *                      Copy of the license is distributed along with this program
 *                      or can be seen at <http://www.gnu.org/licenses/>
 */




/* current project */
#include "common_prconf.h"

/* mh library */
#include "core.h"
#include "strings.h"
#include "strings_arr.h"
#include "arrptrs.h"




// +++++++++++++++++++++++++++++++++++
// TOREM:
enum clarg_type_en {
   Clatype_none=0,      //! No type
   Clatype_opt_toggle,  //! option that is toggled every time it is specified
   Clatype_opt_combine, //! option that increases/decreases its value level every time is specified
                        //! or if option take values they are added to values list
   Clatype_opt_replace, //! option that increases/decreases its value level every time is specified
                        //! or if option take values they are replaced
   Clatype_oper,        //! operand
   clatype_count };
typedef enum clarg_type_en  clarg_type_en;




typedef struct st_clarg_spec st_clarg_spec;
/*!:: Command line argument specification.
*     Structure containig option or operand characteristics.
*/
struct st_clarg_spec
{
   /*- Characteristics set during specification: -*/
   clarg_type_en type;  //! cl-argument type
   uint id;             //! order in command-line - used by operands, 0 for options
   str_t sname;         //! short name - used by options
   str_t lname;         //! long name - used by options
   sint vn_min;         //! min number of values accepted
   sint vn_max;         //! max number of values accepted
   strarr_t* p_valid_vals; //! valid values
   bool mandatory;      //! option mandatory?
   /*- Values set during parsing: -*/
   int val_lvl;         //! some opts increase level each time they are specified on cl
                        //! 0-not set, n+/-n - value level
   strarr_t* p_cl_vals; //! after parsing: values set on cl, NULL if none
   size_t found;        //! after parsing: option found on cl? order of appearance on cl 0-NOT found
};

#define ST_CLARG_SPEC_Z (st_clarg_spec){Clatype_none,0,STR_T_Z,STR_T_Z,0,0,NULL,false,0,NULL,0}




typedef struct st_clarg_specs_arr st_clarg_specs_arr;
/*!:: Array of command line arguments specifications.
*     Structure containig array of option and operand specifications.
*/
struct st_clarg_specs_arr
{
   st_clarg_spec* OptEntries;   //! array of specification structs
   uint ns;          //! structures that are set
   uint na;          //! structures that are allocated alltogether
   /*- admin -*/
   size_t anum;      //! arbitrary number, freely used by functions
   int argc;         //! num of cl-args given on cl
};

#define ST_CLARG_SPECS_ARR_Z {NULL,0,0,0,0}










#endif /*- MHCL_CLAP_SRCSPEC_CLARGSTOR_H_INC -*/
