#ifndef MHCL_PRCONF_SRCSPEC_PRCONFSPEC_H
#define MHCL_PRCONF_SRCSPEC_PRCONFSPEC_H


/*!F::Command-line arguments specificator - external interface file.
 *
 *  ::Provides functions for other program modules.
 */




/* mh library */
#include "core.h"
#include "errors.h"

/* current project */
#include "common_prconf.h"
#include "clarg_storage.h.inc"     // structures to hold arguments data




enum behopt_en {
   Behopt_none = 0,
   Behopt_on,        //! switches on / overwrites values
   Behopt_toggle,    //! each appearance toggles option on/off / overwrites values
   Behopt_level,     //! increases set level / appends values
   Behopt_count };
typedef enum behopt_en behopt_en;


typedef struct st_opt_spec st_opt_spec;
struct st_opt_spec
{
   uint id;
   behopt_en beh;       //! bahavior
   str_t sname;         //! short name - used by options
   str_t lname;         //! long name - used by options
   sint vn_min;         //! min number of values accepted
   sint vn_max;         //! max number of values accepted
   strarr_t* p_vv;      //! valid values
   bool mand;           //! option mandatory?
   // parsing result
   int lvl;          //! some opts increase level each time they are specified on cl //! 0-not set, n+/-n - value level
   strarr_t vals;    //! after parsing: values set on cl, NULL if none
};
#define ST_OPT_SPEC_Z {STR_T_Z,STR_T_Z,0,0,NULL,false,0,STRARR_T_Z}


typedef struct st_oper_spec st_oper_spec;
struct st_oper_spec
{
   uint id;
   uint vn_min;         //! min number of values accepted
   sint vn_max;         //! max number of values accepted
   strarr_t* p_vv;      //! valid values
   bool mand;           //! operand mandatory?
   // parsing result
   strarr_t vals;    //! after parsing: values set on cl, NULL if none
};
#define ST_OPER_SPEC_Z {0,0,NULL,false,STRARR_T_Z}


typedef struct st_specs_arr st_specs_arr;
struct st_specs_arr
{
   arr_ptrs_t opt_specs;
   arr_ptrs_t oper_specs;
};

#define ST_SPECS_ARR_Z {ARR_PTRS_T_Z,ARR_PTRS_T_Z}




enum argkind_en {
   Argkind_none = 0,
   Argkind_short,
   Argkind_long,
   Argkind_short_off,
   Argkind_long_off,
   Argkind_oper,
   Argkind_err,
   argkind_count };
typedef enum argkind_en argkind_en;


typedef struct st_arg_ch st_arg_ch;
struct st_arg_ch
{
   argkind_en kind;
   str_t name;
   strarr_t vals;
   const char* clarg;      //! reference to original cl argument
};
#define ST_ARG_CH_Z {0,STR_T_Z,STRARR_T_Z,NULL}

/*!1:: Initialize argument specifications. */
extern void cla_specs_initialize (st_specs_arr* stArgSpecs, uint nopts, uint nopers)
               __attr_nonnull__(1)
               ;

/*!1:: Establish new option specification. */
extern void cla_opt_establish (st_specs_arr* stArgSpecs, uint id, behopt_en beh,
      const char* sname, const char* lname, sint vals_min, sint vals_max,
      strarr_t* p_vv, bool mand)
               __attr_nonnull__(1,4,5)
               ;

/*!1:: Establish new operand specification. */
extern void cla_oper_establish (st_specs_arr* stArgSpecs, uint id,
      sint vals_min, sint vals_max, strarr_t* p_vv, bool mand)
               __attr_nonnull__(1)
               ;

/*!1:: Cleanup argument specifications. */
extern void cla_specs_cleanup (st_specs_arr* stArgSpecs)
               __attr_nonnull__(1)
               ;


#ifdef DEBUG
/* Used only by tests in debug mode. */
extern bool specificator_all_tests_run (void)
      __attr_wures__;
#endif




#endif /*- MHCL_PRCONF_SRCSPEC_PRCONFSPEC_H -*/
