/*!F::Tfman file system actions - main implementation file.
 *
 *  ::Purpose:
 *    Parse text file containig text representation of directory or whole
 *    directories tree. Deduct set of file system actions to perform on entires.
 *
 *  ::Goals:
 *    - thread-safe, reentrant
 *    - buffer/integer overflow safe
 *    - propagate or handle errors with routines from "errors.h"
 *
 *  ::Functionality:
 *
 *  ::Limitations:
 *
 *  ::Properties:
 *    compiler options:   -std=c99
 *    thread safe
 *
 *  ::Other:
 *    File:             This file is part of 'Tfman' program.
 *                      For terms of distribution, warranty and licensing see README file.
 *    Created:          2014-05-08
 *    (c) Copyright:    Maciej Helminiak 2014
 *    License:          GNU General Public License, version 3
 *                      Copy of the license is distributed along with this program
 *                      or can be seen at <http://www.gnu.org/licenses/>
 */




/* mh library */
#include "core.h"
#include "errors.h"
#include "debug.h"
#include "cstrings.h"
#include "strings.h"
#include "dstream_operate.h"
#include "stat_change.h"
#include "stat_info.h"

/* current project */
#include "common_tfman.h"
#include "parse_acts.h"

/* libc */
#include <string.h>     // memset(), strcmp()
#include <stdio.h>      // printf(), putc()
#include <unistd.h>     // stat()
#include <sys/stat.h>   // mkdir(), open()
#include <sys/types.h>  // stat()
#include <libgen.h>     // basename() - POSIX version (there is GNU ver)

/* current module */
#include "perform_actions.h"
#include "perform_actions_aux.h.inc"




/*!:: Present actions holded in actions bundle in readable form.
 *    Prepare names for each action.
 * ::Errors:
 *    !!! printf() errors - add handling!
 */
size_t                           //! number of actions presented
acts_sift_and_present_to_user (
   const st_actions_table* stTable,    //! actions bundle
   const st_parser_main* stParsMain    //! main parser structure
){
  ___Dassert(stTable, "'stTable' parameter invalid");
  ___Dassert(stTable, "'stParsMain' parameter invalid");

   size_t actions_sum = 0;

   printf("\nFollowing actions will be performed in presented order:\n");

   /*: Priority actions: */
   { /*--*/

      const arr_ptrs_t* parrActs = &stTable->arrActKindsPrior;   //helper var

      /*: For each priority action: */
      uint ii = 0;
      st_action* el_act = NULL;
      while ((el_act = arrptrs_it(parrActs, &ii)) != NULL)
      {
         present_action_to_user_and_dottize_names(el_act, stParsMain);
            HandleErr(handle_err);
         ++actions_sum;
      }
   } /*--*/

   /*: Normal actions: */
   for (size_t i = Acttype_none+1; i < acttype_count; ++i) //: For each kind of action:
   {
      const arr_ptrs_t* parrActs = &stTable->arrActKinds[i];   //helper var
      //bool act_present = false;

      /*: For each action (in array of same kind actions): */
      uint ii = 0;
      st_action* el_act = NULL;
      while ((el_act = arrptrs_it(parrActs, &ii)) != NULL)
      {
         present_action_to_user_and_dottize_names(el_act, stParsMain);
            HandleErr(handle_err);
         ++actions_sum;
      }
   }

   return actions_sum;

  handle_err:
   return 0;
}




/*!:: Perform actions holded in actions bundle.
 * ::Errors:
 *    perform_action_()       -- this function errors (in practice all errors of perform_*() functions)
 */
void
acts_perform (
   st_actions_table* stTable    //! actions bundle
){
   ___Dassert(stTable, "'stTable' parameter invalid");

   /*: Priority actions: */
   { /*--*/

      arr_ptrs_t* parrActs = &stTable->arrActKindsPrior;   //helper var

      /*: For each priority action: */
      uint ii = 0;
      st_action* el_act = NULL;
      while ((el_act = arrptrs_it(parrActs, &ii)) != NULL)
      {
         perform_action_on_filesystem(el_act);
            HandleErr(handle_err);
      }
   } /*--*/

   /*: Normal actions: */
   for (size_t i = Acttype_none+1; i < acttype_count; ++i) //: For each kind of action
   {
      arr_ptrs_t* parrActs = &stTable->arrActKinds[i];   //helper var
      //bool act_present = false;

      /*: For each action (in array of same kind actions): */
      uint ii = 0;
      st_action* el_act = NULL;
      while ((el_act = arrptrs_it(parrActs, &ii)) != NULL)
      {
         perform_action_on_filesystem(el_act);
            HandleErr(handle_err);
      }
   }

   /*: Finish: */
   return;

  handle_err:
   //no easy way to restore to original state
   return;
}
