/* This file is part of Tfman, program distributed under GNU General Public
   License. For more information see README and LICENSE files. */

#ifndef TFMAN_SRCACTS_PARS_ACTIONS_AUX_H_INC
#define TFMAN_SRCACTS_PARS_ACTIONS_AUX_H_INC


/*!F::Tfman file system actions - auxillary file.
 *
 *  ::Provides helper functions for main implementation file.
 */




/* employs includes from actionsimage.c */




/* perform_*()       -- perform specified file system action
 *
 * Note:
 *  . names are absolute or relative to current working directory of process
 *    running 'tfman' program
 */




/* Present action to user */
static inline void present_action_to_user_and_dottize_names (st_action* stAct,
                           const st_parser_main* stParsMain)
                     __attr_nonnull__(1,2)
                     ;

/* Perform action on filesystem */
static inline void perform_action_on_filesystem (const st_action* stAct)
                     __attr_nonnull__(1)
                     ;

static inline void perform_act_for_name (acttype_en type, str_t* fname, const str_t* oper)
                     __attr_nonnull__(2,3)
                     ;




static inline
void
present_action_to_user_and_dottize_names (
   st_action* stAct,
   const st_parser_main* stParsMain    //! main parser structure
){
   ___Dassert(stAct, "'stAct' parameter invalid");
   ___Dassert(stParsMain, "'stParsMain' parameter invalid");


   //: remove dots from operand
   char* oper_dott = NULL;
   stat_replacedots(&oper_dott, stAct->oper.buf);
   str_repl(&stAct->oper, oper_dott);
      HandleErr(handle_err);
   cs_des(&oper_dott);

   /*: For each name in action: */
   uint jj = 0;
   str_t* el_nam = NULL;
   while ((el_nam = sarr_its(&stAct->fnames, &jj)) != NULL)
   {
      //: remove dots from name
      char* name_dott = NULL;
      stat_replacedots(&name_dott, el_nam->buf);
      str_repl(el_nam, name_dott);
         HandleErr(handle_err);
      cs_des(&name_dott);

      const st_act_def* p_act_def = stAct->p_act_def;
      //: print action:
      int space = 30 - el_nam->nc;
      printf("%s:%*s [%s] %s %s\n",
         el_nam->buf, space, " ", p_act_def->name, p_act_def->tag, stAct->oper.buf);

      //: TODO: workout different way of marking unsuccessful globbing
      if (strstr(el_nam->buf, TFM_GLOBFAILED_MSG) != NULL) {
         //: erase name to deactivate it
         str_repl(el_nam, "");
            HandleErr(handle_err);
      }
   }

   return;

  handle_err:
   return;
}

static inline
void
perform_action_on_filesystem (
   const st_action* stAct 
){
   ___Dassert(stAct, "'stAct' parameter invalid");

   /*: For each name in action: */
   uint jj = 0;
   str_t* el_nam = NULL;
   while ((el_nam = sarr_its(&stAct->fnames, &jj)) != NULL)
   {
      if (el_nam->nc < 1) { continue; }   //: file name empty - ignore action
      perform_act_for_name(stAct->p_act_def->type, el_nam, &stAct->oper);
      if (stAct->ign_err == true && IsErr()) {
         fprintf(stdout, "ACTION FAILED: %s (%s)\n", ErrMessage(), ErrDescr());
         ClrErr();
      }
         HandleErr(handle_err);
   }

   return;

  handle_err:
   return;
}




#define PERFORMING1(NAME) printf("performing: %s(): %s\n", __func__, (NAME))
#define PERFORMING2(NAME,OP) printf("performing: %s(): %s\t%s\n", __func__, (NAME), (OP))

/* Create entry in file system. If name ends with '/' created entry is empty
   directory, otherwise it is empty file.
   Errors:
   . ds_mkdir()
   . ds_mkfile2()
 */
static inline
void
perform_create_ (
   const str_t* str_name      //! name of entry
){
   ___Dassert(str_name, "'str_name' parameter invalid");

   const char* cs_name = str_name->buf;
   size_t nc = str_name->nc;

   PERFORMING1(cs_name);

   if (cs_name[nc-1] == '/') {
      ds_mkdir(cs_name, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
   } else {
      ds_mkfile2(cs_name, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
   }
      HandleErrm(handle_err, "couldn't create '%s'", cs_name);

   /*: Finish: */
   return;

  handle_err:
   return;
}

/* Change permissions of entry.
   Errors:
   . stat_chmod()
 */
static inline
void
perform_chperms_ (
   const str_t* str_name,     //! name of entry
   const str_t* str_op        //! mode string in format "rwxrwxrwx"
){
   ___Dassert(str_name, "'str_name' parameter invalid");
   ___Dassert(str_op, "'str_op' parameter invalid");

   PERFORMING2(str_name->buf, str_op->buf);

   stat_chmod(str_name->buf, 0, str_op->buf);
      HandleErrm(handle_err, "couldn't change permissions of '%s'", str_name->buf);

   /*: Finish: */
   return;

  handle_err:
   return;
}

/* Change user owning entry.
   Errors:
   . stat_chusr()
 */
static inline
void
perform_chuser_(
   const str_t* str_name,     //! name of entry
   const str_t* str_op        //! user name
){
   ___Dassert(str_name, "'str_name' parameter invalid");
   ___Dassert(str_op, "'str_op' parameter invalid");

   PERFORMING2(str_name->buf, str_op->buf);

   stat_chusr(str_name->buf, -1, str_op->buf);
      HandleErrm(handle_err, "couldn't change owning user of '%s'", str_name->buf);

   /*: Finish: */
   return;

  handle_err:
   return;
}

/* Change group owning entry.
   Errors:
   . stat_chusr()
 */
static inline
void
perform_chgroup_(
   const str_t* str_name,     //! name of entry
   const str_t* str_op        //! group name
){
   ___Dassert(str_name, "'str_name' parameter invalid");
   ___Dassert(str_op, "'str_op' parameter invalid");
   PERFORMING2(str_name->buf, str_op->buf);

   stat_chgrp(str_name->buf, -1, str_op->buf);
      HandleErrm(handle_err, "couldn't change owning group of '%s'", str_name->buf);

   /*: Finish: */
   return;

  handle_err:
   return;
}

/* Create symbolic link.
   Errors:
   . ds_mkslink()
 */
static inline
void
perform_slink_(
  str_t* str_name,     //! name of entry
  const str_t* str_op        //! name of target
){
  ___Dassert(str_name, "'str_name' parameter invalid");
  ___Dassert(str_op, "'str_op' parameter invalid");

  stat_remove_endslashess(str_name);

  PERFORMING2(str_name->buf, str_op->buf);

  ds_mkslink (str_name->buf, str_op->buf);
    HandleErrm(handle_err, "couldn't create symlink '%s->%s'", str_name->buf, str_op->buf);

  /*: Finish: */
  return;

 handle_err:
  return;
}

/* Create symbolic link. Overwrite if target exist.
   Errors:
   . ds_mkslink()
 */
static inline
void
perform_slink_o_(
   str_t* str_name,     //! name of entry
   const str_t* str_op        //! name of target
){
   ___Dassert(str_name, "'str_name' parameter invalid");
   ___Dassert(str_op, "'str_op' parameter invalid");

  stat_remove_endslashess(str_name);

  PERFORMING2(str_name->buf, str_op->buf);

  ds_rmentry2(str_name->buf, NULL);
  if (GetErr() == ENOENT) {
    ClrErr();
  }
    HandleErrm(handle_err, "couldn't create symlink '%s->%s'", str_name->buf, str_op->buf);

  ds_mkslink (str_name->buf, str_op->buf);
    HandleErrm(handle_err, "couldn't create symlink '%s->%s'", str_name->buf, str_op->buf);

  /*: Finish: */
  return;

 handle_err:
  return;
}

/* Create hard link.
   Errors:
   . ds_mkhlink()
 */
static inline
void
perform_hlink_(
   str_t* str_name,     //! name of entry
   const str_t* str_op        //! name of target
){
   ___Dassert(str_name, "'str_name' parameter invalid");
   ___Dassert(str_op, "'str_op' parameter invalid");

  stat_remove_endslashess(str_name);

  PERFORMING2(str_name->buf, str_op->buf);

  ds_mkhlink (str_name->buf, str_op->buf);
    HandleErrm(handle_err, "couldn't create hardlink '%s->%s'", str_name->buf, str_op->buf);

  /*: Finish: */
  return;

 handle_err:
  return;
}

/* Create hard link. Overwrite if target exist.
   Errors:
   . ds_mkhlink()
 */
static inline
void
perform_hlink_o_(
   str_t* str_name,     //! name of entry
   const str_t* str_op        //! name of target
){
   ___Dassert(str_name, "'str_name' parameter invalid");
   ___Dassert(str_op, "'str_op' parameter invalid");

  stat_remove_endslashess(str_name);

  PERFORMING2(str_name->buf, str_op->buf);

  ds_rmentry2(str_name->buf, NULL);
  if (GetErr() == ENOENT) {
    ClrErr();
  }
    HandleErrm(handle_err, "couldn't create hardlink '%s->%s'", str_name->buf, str_op->buf);

  ds_mkhlink (str_name->buf, str_op->buf);
    HandleErrm(handle_err, "couldn't create hardlink '%s->%s'", str_name->buf, str_op->buf);

  /*: Finish: */
  return;

 handle_err:
  return;
}

/* Change file type.
   Errors:
   . ENOMEM          -- no memory available
   . stat_file_type_is()   -- this function errors
   . ds_move2()       -- this function errors
   . ds_mkslink()    -- this function errors
   . ds_mkfile2()     -- this function errors
   . ds_mkdir()      -- this function errors
   . ds_rmentry2()    -- this function errors
 */
static inline
void
perform_chtype_(
   const str_t* str_name,     //! name of entry
   const str_t* str_op        //! new type - one letter: "l","-","d"
){
   ___Dassert(str_name, "'str_name' parameter invalid");
   ___Dassert(str_op, "'str_op' parameter invalid");

   PERFORMING2(str_name->buf, str_op->buf);
   printf("Not implemented yet\n");
   return;

   /* ....................................................................... */
   /*: Prepare: */
   //: Get current file type
   char type = stat_file_type_is(str_name->buf, false);
      HandleErr(handle_err);

   //: cover case when new and old file types are same
   if (str_op->buf[0] == type && str_op->buf[1] == '\0') {
      goto handle_succ;
   }

   char* name_new = NULL;

   //: create unique name
   name_new = cs_cref("%s~", str_name->buf);
      HandleErr(handle_err);
   while (stat_file_exists(name_new)) {
      cs_app(&name_new, "~");
         HandleErr(handle_err);
   }

   //: move file there
   ds_move2(str_name->buf, name_new, NULL, NULL, 0);
      HandleErr(handle_err);


   /* ....................................................................... */
   /*: Create file of desired type: */
   if (type == '-')
   {
      /*: reg -> l */
      if (strcmp(str_op->buf, "l") == 0)
      {
         //: create symlink
         ds_mkslink (str_name->buf, basename(name_new));
         if (GetErr() != 0) {
            ds_move2(name_new, str_name->buf, NULL, NULL, 0);
               HandleErr(handle_err);
         }
            HandleErr(handle_err);
      }
      /*: reg -> d */
      else if (strcmp(str_op->buf, "d") == 0)
      {
         //: create directory
         ds_mkdir(str_name->buf, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
         if (GetErr() != 0) {
            ds_move2(name_new, str_name->buf, NULL, NULL, 0);
               HandleErr(handle_err);
         }
            HandleErr(handle_err);
      }
      /*: reg -> "anything else" */
      else
      {
         printf("ERROR1\n");
      }
   }
   else if (type == 'd')                                          //DIR
   {
      /*: d -> reg */
      if (strcmp(str_op->buf, "-") == 0)
      {
         ds_mkfile2(str_name->buf, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
         if (GetErr() != 0) {
            ds_move2(name_new, str_name->buf, NULL, NULL, 0);
               HandleErr(handle_err);
         }
            HandleErr(handle_err);
      }
      /*: d -> l */
      else if (strcmp(str_op->buf, "l") == 0)
      {
         ds_mkslink (str_name->buf, basename(name_new));
         if (GetErr() != 0) {
            ds_move2(name_new, str_name->buf, NULL, NULL, 0);
               HandleErr(handle_err);
         }
            HandleErr(handle_err);
      }
   }
   else if (type == 'l')                                          //SYMLINK
   {
      /*: l -> reg */
      if (strcmp(str_op->buf, "-") == 0)
      {
         ds_mkfile2(str_name->buf, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
         if (GetErr() != 0) {
            ds_move2(name_new, str_name->buf, NULL, NULL, 0);
               HandleErr(handle_err);
         }
            HandleErr(handle_err);
      }
      /*: l -> d */
      else if (strcmp(str_op->buf, "d") == 0)
      {
         ds_mkdir(str_name->buf, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
         if (GetErr() != 0) {
            ds_move2(name_new, str_name->buf, NULL, NULL, 0);
               HandleErr(handle_err);
         }
            HandleErr(handle_err);
      }
   }
   else                                                              //OTHER
   {
      printf("ERROR2\n");
   }

   /* ....................................................................... */
   /*: cleanup temp file */
   ds_rmentry2(name_new, NULL);
      HandleErr(handle_err);

   cs_des(&name_new);

   /* ....................................................................... */
   /*: Finish: */
  handle_succ:
   return;

  handle_err:
   cs_des(&name_new);
   return;
}

/* Copy entry.
   Errors:
   . ds_copy2()
 */
static inline
void
perform_copy_ (
   const str_t* str_name,     //! name of entry
   const str_t* str_op        //! copy target
){
   ___Dassert(str_name, "'str_name' parameter invalid");
   ___Dassert(str_op, "'str_op' parameter invalid");

   PERFORMING2(str_name->buf, str_op->buf);

   ds_copy2(str_name->buf, str_op->buf, NULL, NULL, FSF_INTO_DIR);
      HandleErrm(handle_err, "couldn't copy '%s->%s'", str_name->buf, str_op->buf);

   /*: Finish: */
   return;

  handle_err:
   return;
}

/* Copy entry, overwrite if target exist.
   Errors:
   . ds_copy2()
 */
static inline
void
perform_copy_o_ (
   const str_t* str_name,     //! name of entry
   const str_t* str_op        //! copy target
){
   ___Dassert(str_name, "'str_name' parameter invalid");
   ___Dassert(str_op, "'str_op' parameter invalid");

   PERFORMING2(str_name->buf, str_op->buf);

   // copy to dest~
   // rm dest
   // mv dest~ dest
   char* tmp_dest = cs_cref("%s~", str_op->buf);
      HandleErrm(handle_err, "couldn't copy '%s->%s'", str_name->buf, str_op->buf);
   while (stat_file_exists(tmp_dest)) {
      char* new_dest = cs_cref("%s~", tmp_dest);
         HandleErrm(handle_err, "couldn't copy '%s->%s'", str_name->buf, str_op->buf);
      cs_des(&tmp_dest);
      tmp_dest = new_dest;
   }

   ds_copy2(str_name->buf, tmp_dest, NULL, NULL, FSF_NONE);
      HandleErrm(handle_err, "couldn't copy '%s->%s'", str_name->buf, str_op->buf);

   ds_rmentry2(str_op->buf, NULL);
   if (GetErr() == ENOENT) {
      ClrErr();
   }
      HandleErrm(handle_err, "couldn't copy '%s->%s'", str_name->buf, str_op->buf);

   ds_move2(tmp_dest, str_op->buf, NULL, NULL, FSF_NONE);
      HandleErrm(handle_err, "couldn't copy '%s->%s'", str_name->buf, str_op->buf);

   /*: Finish: */
   cs_des(&tmp_dest);
   return;

  handle_err:
   if (tmp_dest != NULL) { cs_des(&tmp_dest); }
   return;
}

/* Move entry.
   Errors:
   . ds_move2()
 */
static inline
void
perform_move_ (
   const str_t* str_name,     //! name of entry
   const str_t* str_op
){
   ___Dassert(str_name, "'str_name' parameter invalid");
   ___Dassert(str_op, "'str_op' parameter invalid");

   PERFORMING2(str_name->buf, str_op->buf);

   ds_move2(str_name->buf, str_op->buf, NULL, NULL, FSF_INTO_DIR);
      HandleErrm(handle_err, "couldn't move '%s->%s'", str_name->buf, str_op->buf);

   /*: Finish: */
   return;

  handle_err:
   return;
}

/* Move entry, overwrite if target exist.
   Errors:
   . ds_move2()
 */
static inline
void
perform_move_o_ (
   const str_t* str_name,     //! name of entry
   const str_t* str_op
){
   ___Dassert(str_name, "'str_name' parameter invalid");
   ___Dassert(str_op, "'str_op' parameter invalid");

   PERFORMING2(str_name->buf, str_op->buf);

   // move to dest~
   // rm dest
   // mv dest~ dest
   char* tmp_dest = cs_cref("%s~", str_op->buf);
      HandleErrm(handle_err, "couldn't copy '%s->%s'", str_name->buf, str_op->buf);
   while (stat_file_exists(tmp_dest)) {
      char* new_dest = cs_cref("%s~", tmp_dest);
         HandleErrm(handle_err, "couldn't copy '%s->%s'", str_name->buf, str_op->buf);
      cs_des(&tmp_dest);
      tmp_dest = new_dest;
   }

   ds_move2(str_name->buf, tmp_dest, NULL, NULL, FSF_NONE);
      HandleErrm(handle_err, "couldn't copy '%s->%s'", str_name->buf, str_op->buf);

   ds_rmentry2(str_op->buf, NULL);
   if (GetErr() == ENOENT) {
      ClrErr();
   }
      HandleErrm(handle_err, "couldn't move '%s->%s'", str_name->buf, str_op->buf);

   ds_move2(tmp_dest, str_op->buf, NULL, NULL, FSF_NONE);
      HandleErrm(handle_err, "couldn't move '%s->%s'", str_name->buf, str_op->buf);

   /*: Finish: */
   cs_des(&tmp_dest);
   return;

  handle_err:
   if (tmp_dest != NULL) { cs_des(&tmp_dest); }
   return;
}

/* Move entry to wastebin. - NOT IMPLEMENTED
   Errors:
 */
static inline
void
perform_trash_ (
   const str_t* str_name,     //! name of entry
   const str_t* str_wb        //! waste bin directory
){
   ___DParam(str_name, != NULL);
   ___DParam(str_wb, != NULL);

   PERFORMING1(str_name->buf);
   printf("Not implemented yet\n");

   //DUMMY
   str_wb = str_name;
   str_name = str_wb;

   /*: Finish: */
   return;
}

/* Remove entry irreversible.
   Errors:
   . ds_rmentry2()
 */
static inline
void
perform_remove_ (
   const str_t* str_name      //! name of entry
){
   ___Dassert(str_name, "'str_name' parameter invalid");

   PERFORMING1(str_name->buf);

   ds_rmentry2(str_name->buf, NULL);
      HandleErrm(handle_err, "couldn't remove '%s'", str_name->buf);

   /*: Finish: */
   return;

  handle_err:
   return;
}

/* Swap names. - NOT IMPLEMENTED
   Errors:
 */
static inline
void perform_swnames_(
   const str_t* str_name,     //! name of entry
   const str_t* str_op
){
   ___Dassert(str_name, "'str_name' parameter invalid");
   ___Dassert(str_op, "'str_op' parameter invalid");

   PERFORMING2(str_name->buf, str_op->buf);
   printf("Not implemented yet\n");

   /*: Finish: */
   return;
}




/* Perform given action.
   Errors:
      all errors defined for perform_*() functions
 */
static inline
void
perform_act_for_name (
   acttype_en type,   //! type of action
   str_t* fname,     //! entry name
   const str_t* oper       //! operand name
){
   ___Dassert(fname, "'fname' parameter invalid");
   ___Dassert(oper, "'oper' parameter invalid");

   str_t* tmp = NULL;      //needed temp for not implemented fun

   /*: Action: */
   switch (type)
   {
      case Acttype_create:
         perform_create_(fname);
         break;

      case Acttype_chperms:
         perform_chperms_(fname, oper);
         break;

      case Acttype_chuser:
         perform_chuser_(fname, oper);
         break;

      case Acttype_chgroup:
         perform_chgroup_(fname, oper);
         break;

      case Acttype_slink:
         perform_slink_(fname, oper);
         break;

      case Acttype_slink_o:
         perform_slink_o_(fname, oper);
         break;

      case Acttype_hlink:
         perform_hlink_(fname, oper);
         break;

      case Acttype_hlink_o:
         perform_hlink_o_(fname, oper);
         break;

      case Acttype_copy:
         perform_copy_(fname, oper);
         break;

      case Acttype_copy_o:
         perform_copy_o_(fname, oper);
         break;

      case Acttype_move:
         perform_move_(fname, oper);
         break;

      case Acttype_move_o:
         perform_move_o_(fname, oper);
         break;

      case Acttype_trash:
         tmp = str_cre("wastebin_dir");
         perform_trash_(fname, tmp);
         str_des(&tmp);
         break;

      case Acttype_remove:
         perform_remove_(fname);
         break;

      case Acttype_swnames:
         perform_swnames_(fname, oper);
         break;

      case Acttype_none:    /* \/ Through */
      case acttype_count:
      case Acttype_prop:
      case opertype_count:
      case Actspec_name_beg:
      case Actspec_name_end:
      case Actspec_da_cont:
      case Actspec_da_both:
      case Actspec_da_ignor:
      case Actspec_ign_errors:
      case Actspec_acts_target:
      case spectype_count:
      default:
         ___Sentinel("wrong action type passed to %s", __func__); abort();
   }
      HandleErr(handle_err);

   /*: Finish: */
   return;

  handle_err:
   //no easy way to restore to original state
   return;
}



#endif /*- TFMAN_SRCACTS_PARS_ACTIONS_AUX_H_INC -*/
