#ifndef TFMAN_SRC_COMMON_TFMAN_H
#define TFMAN_SRC_COMMON_TFMAN_H




#define _ISOC99_SOURCE
#define _POSIX_C_SOURCE 200809L
//#define _XOPEN_SOURCE 700






#define TFMAN_PR_PACK "mhsys"
#define TFMAN_PR_NAME "tfman"
#define TFMAN_PR_LONGNAME "text file manager"

#include "VERSION"

#ifndef VER_MAJOR
#define VER_MAJOR 0
#endif
#ifndef VER_MINOR
#define VER_MINOR 0
#endif
#ifndef VER_PATCH
#define VER_PATCH 0
#endif

#define TFMAN_PR_VER VER_MAJOR.VER_MINOR.VER_PATCH
#define TFMAN_PR_FULLNAME TFMAN_PR_NAME " ("TFMAN_PR_PACK") " _STR_(TFMAN_PR_VER)




#ifdef DEBUG
# define TFM_ERR_PRINT EPR_DEBUG_INFO
#else
# define TFM_ERR_PRINT EPR_NONE
#endif


#define TFM_GLOBFAILED_MSG "    -- GLOBBING FAILED - ACTION IGNORED"




// options must start from 0
#define O_HELP 0
#define O_VERBOSE 1
#define O_QUIET 2
#define O_MODE 3
#define O_RECURS 4
#define O_LAYOUT 5
#define O_SORT 6
#define O_FORMAT 7
#define O_CLONE_NAMES 8
#define O_NAMES_ONLY 9
#define O_INDENT 10
#define O_FILE 11
#define O_APPEND 12
#define O_SHOW_ONLY 13
#define O_NO_ASK 14
#define O_LINES 15
// operands must start from 0
#define OPER_A 0




/* === PROGRAM DEFAULT SETTINGS === */


/*: --- General: --- */
/*:   Verbosity level, recursion level
 *    Text representation layout, indenting, mode when not specified on cl
 */
#define TFM_VERBOSITY_LVL 1


/*: --- Scan: --- */
#define TFM_RECURSION_LVL 1
#define TFM_DEF_LAYOUT Lay_block
#define TFM_DEF_INDENT 2
#define TFM_DEF_CLONE_NAMES false

/*: Fields used to compose entry when creating text representation */
#define TFM_ENTRY_FIELDS {\
   Format_name, Format_user, Format_group, Format_perms,\
   Format_lntarget, Format_entrs, Format_size, Format_size2,\
   Format_mtime, Format_props, Format_type, Format_none\
}
/*: Sort order of entries when creating text representation */
#define TFM_SORT_ORDER {\
   Sort_dirfst, Sort_nameidot, Sort_size, Sort_none\
}

/*: Postfixes for size fields */
#define TFM_SIZE_POSTFIXES { " B", "kB", "mB", "gB", "tB" }

/*: Prefix strings used in text representation in scan mode
 *    dirname block mode, dirname tree mode, entry, active entry
 */
#define TFM_SCAN_PREFIXES "#", "|"

/*: Marker strings used in text representation in scan mode
 *    name begin, name end, link target, separator, space (between fields),
 *    access time, modification time, change time
 */
#define TFM_SCAN_MARKERS "[", "]", "_>", "!!>", "    ", " ", "a.", "m.", "c."


/*: --- Parse: --- */
#define TFM_ASK_USER true

/* FORBIDDEN TAGS: . / - _ */

/*: Strings triggering action deducting for line in parse mode
 */
//: dir, entry, priority entry, group, close, group strong, close, end parsing,
//: indentation
#define TFM_PARSE_EFFTAGS "#", "|", "$", "{", "}", "{{", "}}", "!!", "END", 2
//: exclude entry from: dir, group, both; disable whole group/dir
#define TFM_PARSE_EFFTAGS_MOD "=", "-", "E", "x"

/*: Strings marking actions for on-going entry line
 */
//: slink, slink_overw, hlink, hlink_overw, copy, copy_overw, move, move_overw
#define TFM_PARSE_ACT_TAGS1 "->", "->*", "=>", "=>*", ">", ">*", ">>", ">>*"
//: properties (user,group,perms), create, trash, remove, swap names
#define TFM_PARSE_ACT_TAGS2 ":", "~", "+", "*", "<>"
//: name, end, content actions, dir and content actions, ignore following actions,
//: ignore errors for following actions, following actions perfrom on target
#define TFM_PARSE_ACT_TAGS_SPEC "[", "]", "=", "==", "||", "!", "++"




#endif /*- TFMAN_SRC_COMMON_TFMAN_H -*/
