#ifndef TFMAN_SRCPARS_DATATOACTS_H
#define TFMAN_SRCPARS_DATATOACTS_H


/*!F::Transformator data-actions - external interface file.
 *
 *  ::Provides functions for other program modules.
 */




/* current project */
#include "common_tfman.h"

/* mh library */
#include "core.h"
#include "strings.h"
#include "strings_arr.h"
#include "arrptrs.h"

/* libc */

/* current module */
#include "parse_img.h"




/* Data. */
/* -------------------------------------------------------------------------- */


typedef struct st_action st_action;
/*!:: Action.
 *    File system action.
 */
struct st_action
{
   const st_act_def* p_act_def;      //! type of action
   strarr_t fnames;  //! canonical filenames (prepared, after brace expansion and globbing)
   str_t oper;       //! operand
   bool ign_err;      //! do not terminate on error
};
#define ST_ACTION_Z (st_action){Acttype_none,STRARR_T_Z,STR_T_Z}


typedef struct st_actions_table st_actions_table;
/*!:: Table of actions.
 *    All actions retrieved from parsed file, groupped by kind. Order in groups
 *    as in parsed file.
 */
struct st_actions_table
{
   arr_ptrs_t arrActKindsPrior;           //! array of pointers to priority actions - perfomed before anything else in file order
   arr_ptrs_t arrActKinds[acttype_count];  //! array of pointers to actions
                     //! separate for each action type
};
#define ST_ACTIONS_TABLE_Z (st_actions_table){(arr_ptrs_t){0},{(arr_ptrs_t){0}}}




extern void pars_txtrepdata_to_actions (st_actions_table* stTable,
                              st_txt_img* stTxtRep, st_parser_main* stParsMain,
                              uint start_line, uint end_line)
               __attr_nonnull__(1,2,3)
               ;

extern void pars_cleanup_actstable (st_actions_table* stTable)
               __attr_nonnull__(1)
               ;

extern void pars_txtrepdata_to_names (strarr_t* Names,
                              st_txt_img* stTxtRep, st_parser_main* stParsMain,
                              uint start_line, uint end_line)
               __attr_nonnull__(1,2,3)
               ;




#ifdef DEBUG
/* Used only by tests in debug mode. */
extern bool parsacts_all_tests_run (void)
      __attr_wures__;
#endif




#endif /*- TFMAN_SRCPARS_DATATOACTS_H -*/
