/* This file is part of Tfman, program distributed under GNU General Public
   License. For more information see README and LICENSE files. */

#ifndef TFMAN_SRCPARS_DATATOACTS_AUX_H_INC
#define TFMAN_SRCPARS_DATATOACTS_AUX_H_INC


/*!F::Transformator data-actions - auxillary file.
 *
 *  ::Provides helper functions for main implementation file.
 */




/* employs includes from data_toactions.c */




static void retrieve_tags_line (arr_ptrs_t* parrTags, const st_line_ch* stLine)
               __attr_nonnull__(1,2)
               ;

static void retrieve_tags_group (arr_ptrs_t* parrTags, const st_line_ch* stLine)
               __attr_nonnull__(1,2)
               ;

static void retrieve_tags_dir (arr_ptrs_t* parrTags, const st_line_ch* stLine)
               __attr_nonnull__(1,2)
               ;

static uint special_actions_serving (bool* p_accept_acts, bool* p_err, bool* p_target,
                  acttype_en act_type, uint act_cath)
               __attr_nonnull__(1,2,3)
               ;

/* Prepend directory name to name if it comply to certain requirements. */
static inline str_t* prepare_name (const str_t* name, str_t* dir)
                     __attr_nonnull__(1,2)
                     ;

static inline void perform_globbing (strarr_t* names_glob, str_t* name)
                     __attr_nonnull__(1,2)
                     ;

static inline void prepare_operand (str_t* oper, str_t* dir, acttype_en kind)
                     __attr_nonnull__(1,2)
                     ;

static void prepend_parent_dirs_to_names (strarr_t* names_prep, const strarr_t* entries_be,
                              const st_line_ch* stLine, str_t* pardir_name)
               __attr_nonnull__(1,2,3,4)
               ;

static void glob_all_names (strarr_t* names_globbed, strarr_t* names_prep)
               __attr_nonnull__(1,2)
               ;

static void create_actions_for_all_words (st_actions_table* stTable, const st_line_ch* stLine,
      strarr_t* names_glob, arr_ptrs_t* parrTags, str_t* pardir_name)
               __attr_nonnull__(1,2,3,4)
               ;

static void assemble_action_to_table (st_actions_table* stTable,
              acttype_en type, const st_word_ch* prev_file_word,
              const strarr_t* fullnames, const str_t* oper, bool ign_err)
               __attr_nonnull__(1,3,4,5)
               ;

static void assemble_action_to_table_prior (st_actions_table* stTable,
              acttype_en type, const st_word_ch* prev_file_word,
              const strarr_t* fullnames, const str_t* oper, bool ign_err)
               __attr_nonnull__(1,3,4,5)
               ;


   /* Aux function for retrieve_tags_dir() */
   static inline void change_last_word_if_is_file (st_word_ch** p_last_file_word,
                                                   st_word_ch* word)
   {
      switch (word->p_act_def->type) {
      case Acttype_slink:      /*- \/ THROUGH -*/
      case Acttype_slink_o:
      case Acttype_hlink:
      case Acttype_hlink_o:
      case Acttype_copy:
      case Acttype_copy_o:
      case Acttype_move:
      case Acttype_move_o:
      case Acttype_swnames:
         *p_last_file_word = word;
      default:
      case Acttype_none:
      case Acttype_create:
      case Acttype_chperms:
      case Acttype_chuser:
      case Acttype_chgroup:
      case Acttype_prop:
      case Acttype_trash:
      case Acttype_remove:
      case opertype_count:
      case Actspec_name_beg:
      case Actspec_name_end:
      case Actspec_da_cont:
      case Actspec_da_both:
      case Actspec_da_ignor:
      case Actspec_ign_errors:
      case Actspec_acts_target:
      case spectype_count:
      case acttype_count:
         break;
      }
   }

static
void
retrieve_tags_line (
   arr_ptrs_t* parrTags,      //! [out] array of pointers to words with tags
   const st_line_ch* stLine   //! line characteristics
){
   st_word_ch* last_file_word = NULL;  //: last previous word that takes file as operand

   bool accept_acts = true;   //: accept action tags
   bool ignor_err = false;    //: ignore errors for subsequent actions in line
   bool on_target = false;    //: subsequent actions perform on last target
   uint ii = 0; st_word_ch* el_wch = NULL;
   while ((el_wch = arrptrs_it(&stLine->arrWordsCh, &ii)) != NULL)
   {
      //: Special actions:
      uint operation = special_actions_serving(&accept_acts, &ignor_err, &on_target,
             el_wch->p_act_def->type, 1);
      if (operation == 1) { continue; }
      //if (operation == 2) { return; }

      if (ignor_err == true) { el_wch->ign_err = true; }
      if (on_target == true) { el_wch->prev_file_word = last_file_word; }
      arrptrs_cpy(parrTags, el_wch->p_act_def->type, el_wch);
      change_last_word_if_is_file(&last_file_word, el_wch);
      continue;
   }
}

static
void
retrieve_tags_group (
   arr_ptrs_t* parrTags,    //! [out] array of pointers to words with tags
   const st_line_ch* stLine   //! line characteristics
){
   bool accept_acts = true;   //: accept action tags
   bool ignor_err = false;    //: ignore errors for subsequent actions in line
   bool on_target = false;    //: subsequent actions perform on last target
   bool process_grp = true;     //: consider current entry part of this group

   st_group_ch* cur_grp = stLine->p_grp;
   while (cur_grp != NULL)
   {
      st_word_ch* last_file_word = NULL;  //: last previous word that takes file as operand
      // { #%
      if (cur_grp->p_line_def->type == Linetype_grp_beg
          && stLine->p_line_def->type != Linetype_entry
          && stLine->p_line_def->type != Linetype_prior)
      {
         process_grp = false;
      }
      // {{ #% - leave as is
      // | $ - leave as is

      uint ii = 0; st_word_ch* el_wch = NULL;

      if (process_grp == true)
      while ((el_wch = arrptrs_it(&cur_grp->arrWordsCh, &ii)) != NULL)
      {
         //: Special actions:
         uint operation = special_actions_serving(&accept_acts, &ignor_err, &on_target,
               el_wch->p_act_def->type, 2);
         if (operation == 1) { continue; }
         //if (operation == 2) { return; }

         if (arrptrs_el_ret(parrTags, el_wch->p_act_def->type) == NULL) {
            if (ignor_err == true) { el_wch->ign_err = true; }
            if (on_target == true) { el_wch->prev_file_word = last_file_word; }
            arrptrs_cpy(parrTags, el_wch->p_act_def->type, el_wch);
         }
         change_last_word_if_is_file(&last_file_word, el_wch);

         continue;
      }
      cur_grp = cur_grp->p_grp_parent;
   }
}

static
void
retrieve_tags_dir (
   arr_ptrs_t* parrTags,      //! [out] array of pointers to words with tags
   const st_line_ch* stLine   //! line characteristics
){
   bool accept_acts = false;  //: accept action tags
   bool ignor_err = false;    //: ignore errors for subsequent actions in line
   bool on_target = false;    //: subsequent actions perform on last target

   st_line_ch* cur_dir = stLine->p_dir;
   while (cur_dir != NULL)
   {
      st_word_ch* last_file_word = NULL;  //: last previous word that takes file as operand

      uint ii = 0; st_word_ch* el_wch = NULL;
      while ((el_wch = arrptrs_it(&cur_dir->arrWordsCh, &ii)) != NULL)
      {
         //: Special actions:
         uint operation = special_actions_serving(&accept_acts, &ignor_err, &on_target,
                el_wch->p_act_def->type, 3);
         if (operation == 1) { continue; }
         //if (operation == 2) { break; }

         if (arrptrs_el_ret(parrTags, el_wch->p_act_def->type) == NULL) {
            if (ignor_err == true) { el_wch->ign_err = true; }
            if (on_target == true) { el_wch->prev_file_word = last_file_word; }
            arrptrs_cpy(parrTags, el_wch->p_act_def->type, el_wch);
         }
         el_wch->prev_file_word = last_file_word;
         change_last_word_if_is_file(&last_file_word, el_wch);
      }
      cur_dir = cur_dir->p_dir;
   }
}

static
uint                    //! 0-none 1-continue 2-break
special_actions_serving (
   bool* p_accept_acts, //! [out] accept following actions?
   bool* p_err,         //! [out] stop on errors for subsequent actions in line
   bool* p_target,      //! [out] stop on errors for subsequent actions in line
   acttype_en act_type,  //! kind of action served
   uint act_cath        //! 0-none, 1-line, 2-group 3-dir
){
   ___DParam(p_accept_acts, != NULL);
   ___DParam(p_err, != NULL);
   ___DParam(p_target, != NULL);

   if (act_cath == 1)      //line
   {
      if (act_type == Acttype_none)
         { return 1; }     //: no action tag in word
      else if (act_type == Actspec_da_both)
         { *p_accept_acts = true; return 1; }
      else if (act_type == Actspec_da_ignor || act_type == Actspec_da_cont)
         { *p_accept_acts = false; return 1; }

      if (*p_accept_acts == false) { return 1; }

      if (act_type == Actspec_ign_errors) { *p_err = !(*p_err); return 1; }
      if (act_type == Actspec_acts_target) { *p_target = !(*p_target); return 1; }
   }
   else if (act_cath == 2) //group
   {
      if (act_type == Acttype_none)
         { return 1; }     //: no action tag in word
      else if (act_type == Actspec_da_cont || act_type == Actspec_da_both)
         { *p_accept_acts = true; return 1; }
      else if (act_type == Actspec_da_ignor)
         { *p_accept_acts = false; return 1; }

      if (*p_accept_acts == false) { return 1; }

      if (act_type == Actspec_ign_errors) { *p_err = !(*p_err); return 1; }
      if (act_type == Actspec_acts_target) { *p_target = !(*p_target); return 1; }
   }
   else if (act_cath == 3) //dir
   {
      if (act_type == Acttype_none)
         { return 1; }     //: no action tag in word
      else if (act_type == Actspec_da_cont || act_type == Actspec_da_both)
         { *p_accept_acts = true; return 1; }
      else if (act_type == Actspec_da_ignor)
         { *p_accept_acts = false; return 1; }

      if (*p_accept_acts == false) { return 1; }

      if (act_type == Actspec_ign_errors) { *p_err = !(*p_err); return 1; }
      if (act_type == Actspec_acts_target) { *p_target = !(*p_target); return 1; }
   }

   return 0;
}

/* Append '/' to dir if not there and then prepare file name in following way:
      starts with '/'      :leave as is      like: /   /name   /dir/name
      starts with './'     :leave as is      like: ./  ./name  ./dir/name
      "."                  :prepend path
      ".NAME", "NAME"      :prepend path
      difference between . and ./:
         .     - containig directory - directory specified by #
         ./    - current working directory - where pr is executed
   ENOMEM
   errors not covered properly !!!
 */
static inline
str_t*
prepare_name (
   const str_t* name,
   str_t* pardir
){
   ___Dassert(name, "'name' parameter invalid");
   ___Dassert(pardir, "'pardir' parameter invalid");

   str_t* prep_name = NULL;

   /*: prepare dir */
   if (pardir->buf[0] != '\0' && pardir->buf[pardir->nc-1] != '/') {
      str_app(pardir, "/");
         HandleErr(handle_err);
   }

   /*: prepare name */
   if (name != NULL) {
      prep_name = str_cres(name);
         HandleErr(handle_err);
      const char* cs_fname = prep_name->buf;
      if (cs_fname[0] == '/'                       // "/" "/anything"
      || (cs_fname[0] == '.' && cs_fname[1] == '/')// "./anything"
      ) {                                          // LEAVE AS IS
         ;  /*- nothing -*/
      } else if (cs_fname[0] == '.' && cs_fname[1] == '\0') {  // "."
         str_repls(prep_name, pardir);                                 // PUT DIR NAME
            HandleErr(handle_err);
      } else {                                     // ".anything" "anything"
         str_prep(prep_name, pardir->buf);                 // PREPEND DIR
            HandleErr(handle_err);
      }
   }

   return prep_name;

  handle_err:
   if (prep_name != NULL) { str_des(&prep_name); }
   return NULL;
}

/* True if any of * [ ] ? is present unescaped in `str_word`. */
static inline bool contains_glob_chars (const str_t* str_word)
                     __attr_nonnull__(1)
                     __attr_wures__
                     ;

/* Return true if given word contains any of following characters: '* [ ] ?'.
   Characters can not be preceeded with escape char '\'.
 */
static inline
bool                    //! true if word contains glob chars
contains_glob_chars (
   const str_t* str_word      //! word
){
   ___Dassert(str_word, "'str_word' parameter invalid");
   if (  str_word->buf[0] == '*'
      || str_word->buf[0] == '?'
      || str_word->buf[0] == '['
      || str_word->buf[0] == ']' )
   {
      return true;
   }

   for (size_t i = 1; i < str_word->nc; ++i) {
      if (
            (     str_word->buf[i] == '*'
               || str_word->buf[i] == '?'
               || str_word->buf[i] == '['
               || str_word->buf[i] == ']'
            )
            &&
            (
               str_word->buf[i-1] != '\\'
            )
         )

      {
         return true;
      }
   }

   return false;
}

static inline
void
perform_globbing (
   strarr_t* names_glob,   //! [out] names resulted from globbing
   str_t* name             //! name to glob
){
   ___Dassert(names_glob, "'names_glob' parameter invalid");
   ___Dassert(name, "'name' parameter invalid");

   if (!contains_glob_chars(name)) {
      sarr_apps(names_glob, name);    //app pattern unmodified
         HandleErr(HANDLE_ERR);
      goto HANDLE_SUCC;
   }

   /* Perform globbing *?[] - match existing files */
   strarr_t* sarr_m = sarr_cre2(0);  //glob matches - defined here for purpose
   strarr_t* sarr_e = sarr_cre2(0);  //glob scan errors
      HandleErr(HANDLE_ERR);

   ds_glob_name(sarr_m, sarr_e, name);
      HandleErrm(HANDLE_ERR, "error globbing '%s'", name->buf);

   //: matches found
   if (sarr_el_ret(sarr_m, 0)[0] != '\0') {
      sarr_appA(names_glob, sarr_m);
         HandleErr(HANDLE_ERR);
   //: no matches
   } else {
      sarr_apps(names_glob, name);    //app pattern unmodified
      str_app(sarr_el_rets(names_glob, names_glob->ns-1), TFM_GLOBFAILED_MSG);
         HandleErr(HANDLE_ERR);
   }

   sarr_des(&sarr_m);
   sarr_des(&sarr_e);      //globbing errors IGNORED

  HANDLE_SUCC:
   return;

  HANDLE_ERR:
   if (sarr_e != NULL) { sarr_des(&sarr_e); }
   if (sarr_m != NULL) { sarr_des(&sarr_m); }
   return;
}

static inline
void
prepare_operand (
   str_t* oper,   //! [out]
   str_t* pardir,
   acttype_en type
){
   ___DParam(oper, != NULL);
   ___DParam(pardir, != NULL);

   //str_t* prep_oper = NULL;

   switch (type)
   {
         // actions do not take operands
      case Acttype_trash:      /*- \/ THROUGH -*/
      case Acttype_remove:
      case Acttype_create:
         //prep_oper = str_cre("");
         str_repl(oper, "");
            HandleErr(handle_err);
         break;
         // actions take non-file operands
      case Acttype_chperms:
      case Acttype_chuser:
      case Acttype_chgroup:
      case Acttype_prop:
         //prep_oper = str_cres(oper);
            //HandleErr(handle_err);
         /* NO CHANGE */
         break;
         // actions take FILE operands
      case Acttype_slink:      /*- \/ THROUGH -*/
      case Acttype_slink_o:
      case Acttype_hlink:
      case Acttype_hlink_o:
      case Acttype_copy:
      case Acttype_copy_o:
      case Acttype_move:
      case Acttype_move_o:
      case Acttype_swnames:
      {
         str_t* prep_oper = prepare_name(oper, pardir);
         str_repls(oper, prep_oper);
            HandleErr(handle_err);
         str_des(&prep_oper);
         break;
         // special actions - never should get here
      }
      case Acttype_none:       /*- \/ THROUGH -*/
      case opertype_count:
      case Actspec_name_beg:
      case Actspec_name_end:
      case Actspec_da_cont:
      case Actspec_da_both:
      case Actspec_da_ignor:
      case Actspec_ign_errors:
      case Actspec_acts_target:
      case spectype_count:
      case acttype_count:
      default:
         ___Sentinel("wrong action type (%d) passed to %s", type, __func__); abort();
   }

   return;

  handle_err:
   return;
}

static
void
prepend_parent_dirs_to_names (
   strarr_t* names_prep,      //! [out]
   const strarr_t* entries_be,      //! entry names
   const st_line_ch* stLine,  //! line characteristics
   str_t* pardir_name
){
   str_t* name = NULL;

   if (stLine->ind > 0 || stLine->p_line_def->type != Linetype_dir)
   {
      uint ii = 0;
      str_t* el_nam = NULL;
      while ((el_nam = sarr_its(entries_be, &ii)) != NULL)
      {
         name = prepare_name(el_nam, pardir_name);
            HandleErr(handle_err);
         sarr_att2(names_prep, name->buf, name->nc, 0);
            HandleErr(handle_err);
         free(name);
      }
   }
   else
   {
      sarr_appA(names_prep, entries_be);
         HandleErr(handle_err);
   }

   return;

  handle_err:
   if (name != NULL) { str_des(&name); }
   return;
}

static
void
glob_all_names (
   strarr_t* names_globbed,
   strarr_t* names_prep
){
   uint ii = 0;
   str_t* el_nam = NULL;
   while ((el_nam = sarr_its(names_prep, &ii)) != NULL)
   {
      perform_globbing(names_globbed, el_nam);
         HandleErr(handle_err);
   }

   return;

  handle_err:
   return;
}


  // Aux fun fo create_actions_for_all_words()
  bool in_perms_format (const str_t* oper)
  {
    if (oper->nc != 9) { return false; }
    for (size_t i = 0; i < 9; ++i) {
      if (strchr("rwx-", oper->buf[i]) == NULL) { return false; }
    }
    return true;
  }

static
void
create_actions_for_all_words (
  st_actions_table* stTable, //! [out] file system actions
  const st_line_ch* stLine,  //! line characteristics
  strarr_t* names_glob,      //! names after globbing
  arr_ptrs_t* parrTags,      //! array of pointers to words with tags
  str_t* pardir_name         //! parent dir name
){
  /*: For each word: */
  for (size_t jj = 0; jj < parrTags->ns; ++jj)
  {
      st_word_ch* el_word = arrptrs_el_ret(parrTags, jj);
      if (el_word == NULL) { continue; }

      acttype_en act_type[3] = { el_word->p_act_def->type, Acttype_none, Acttype_none };
      st_word_ch* prev_file_word = el_word->prev_file_word;

      /*: prepare operand */
      if (! el_word->oper_prepared) {
        for (size_t i = 0; i < STWORDCH_OPER_NUM; ++i) {
          if (el_word->oper[i].buf != NULL) {
             prepare_operand(&el_word->oper[i], pardir_name, act_type[0]); //: only if kind of acctions requires it
                HandleErr(handle_err);
          }
        }
        el_word->oper_prepared = true;
      }
      /*: prepare operand of prev file action if any */
      if (prev_file_word != NULL && !prev_file_word->oper_prepared) {
        for (size_t i = 0; i < STWORDCH_OPER_NUM; ++i) {
          if (prev_file_word->oper[i].buf != NULL) {
             prepare_operand(&prev_file_word->oper[i], pardir_name, act_type[0]); //: only if kind of acctions requires it
                HandleErr(handle_err);
          }
        }
        prev_file_word->oper_prepared = true;
      }

      /*: Assemble actions to proper table: */

      if (stLine->p_line_def->type == Linetype_dir) {
        //: for dir append '/' where not present
        uint ii = 0;
        str_t* el_nam = NULL;
        while ((el_nam = sarr_its(names_glob, &ii)) != NULL) {
          if (el_nam->buf[el_nam->nc-1] != '/') {
            str_app(el_nam, "/");
              HandleErr(handle_err);
          }
        }
      }

      if (act_type[0] == Acttype_prop) {
        //: properties action split to few actions (user,group,perms)

        if (el_word->oper[2].buf != NULL) {
          //: all 3 set
          act_type[0] = Acttype_chuser;
          act_type[1] = Acttype_chgroup;
          act_type[2] = Acttype_chperms;
        } else {
          //: 2 set:
          act_type[0] = Acttype_none;
          bool got_perms = false;
          if (el_word->oper[0].buf != NULL) {
            if (in_perms_format(&el_word->oper[0])) {
              act_type[0] = Acttype_chperms;
              got_perms = true;
            } else {
              act_type[0] = Acttype_chuser;
            }
          }
          if (el_word->oper[1].buf != NULL) {
            if (got_perms == true) {
              act_type[1] = Acttype_chgroup;
            } else {
              if (in_perms_format(&el_word->oper[1])) {
                act_type[1] = Acttype_chperms;
              } else {
                act_type[1] = Acttype_chgroup;
              }
            }
          }
        }

      }

      if (stLine->p_line_def->type == Linetype_prior) {
        for (size_t i = 0; i < 3; ++i) {
          if (act_type[i] == Acttype_none) { continue; }
          assemble_action_to_table_prior(stTable, act_type[i], prev_file_word, names_glob, &el_word->oper[i], el_word->ign_err);
            HandleErr(handle_err);
        }
      } else {
        for (size_t i = 0; i < 3; ++i) {
          if (act_type[i] == Acttype_none) { continue; }
          assemble_action_to_table(stTable, act_type[i], prev_file_word, names_glob, &el_word->oper[i], el_word->ign_err);
            HandleErr(handle_err);
        }
      }
   }

   return;

  handle_err:
   return;
}

static
void
assemble_action_to_table (
  st_actions_table* stTable,  //! [out] file system actions
  acttype_en type,
  const st_word_ch* prev_file_word,
  const strarr_t* fullnames,
  const str_t* oper,
  bool ign_err
){
  ___DParam(stTable, != NULL);
  ___DParam(type, != Acttype_none);
  ___DParam(fullnames, != NULL);
  ___DParam(oper, != NULL);

  const st_act_def* p_act_def = &StActionDefinitions[type];

  st_action* stAct = act_constr_(p_act_def, oper, ign_err);
    HandleErr(HANDLE_ERR);

  if (prev_file_word != NULL) {
    sarr_apps(&stAct->fnames, &prev_file_word->oper[0]);
      HandleErr(HANDLE_ERR);
  } else {
    sarr_appA(&stAct->fnames, fullnames);
      HandleErr(HANDLE_ERR);
  }
  arrptrs_app(&stTable->arrActKinds[type], stAct);
    HandleErr(HANDLE_ERR);

  return;

 HANDLE_ERR:
  if (stAct != NULL) { act_destr_(&stAct); }
  return;
}

static
void
assemble_action_to_table_prior (
  st_actions_table* stTable,  //! [out] file system actions
  acttype_en type,
  const st_word_ch* prev_file_word,
  const strarr_t* fullnames,
  const str_t* oper,
  bool ign_err
){
  ___DParam(stTable, != NULL);
  ___DParam(type, != Acttype_none);
  ___DParam(fullnames, != NULL);
  ___DParam(oper, != NULL);

  const st_act_def* p_act_def = &StActionDefinitions[type];

  st_action* stAct = act_constr_(p_act_def, oper, ign_err);
    HandleErr(HANDLE_ERR);

  if (prev_file_word != NULL) {
    sarr_apps(&stAct->fnames, &prev_file_word->oper[0]);
      HandleErr(HANDLE_ERR);
  } else {
    sarr_appA(&stAct->fnames, fullnames);
      HandleErr(HANDLE_ERR);
  }
  arrptrs_app(&stTable->arrActKindsPrior, stAct);
    HandleErr(HANDLE_ERR);

  return;

 HANDLE_ERR:
  if (stAct != NULL) { act_destr_(&stAct); }
  return;
}




#endif /*- TFMAN_SRCPARS_DATATOACTS_AUX_H_INC -*/
