/* This file is part of Tfman, program distributed under GNU General Public
   License. For more information see README and LICENSE files. */

#ifndef TFMAN_SRCPARS_DATATOACTS_INTER_H_INC
#define TFMAN_SRCPARS_DATATOACTS_INTER_H_INC


/*!F::Transformator data-actions - internal interface file.
 *
 *  ::Provides functions for basic manipulations of this module data structures.
 */




/* employs includes from data_toactions.c */




static inline
st_action*
act_constr_ (
   const st_act_def* p_act_def,      //! type of action
   const str_t* operand,
   bool ign_err
){
   st_action* act_new = mem_alloc(sizeof *act_new);
      HandleErr(handle_err);

   act_new->p_act_def = p_act_def;
   act_new->fnames = STRARR_T_Z;
  if (operand->buf != NULL) {
    //: only if operand was supplied for action
    str_inis(&act_new->oper, operand);
  } else {
    str_ini(&act_new->oper, "");
  }
   act_new->ign_err = ign_err;

   return act_new;

  handle_err:
   return NULL;
}

static inline
void
act_destr_ (
   st_action** stAct
){
   st_action* act = *stAct;
   sarr_fin(&act->fnames);
   str_fin(&act->oper);
   mem_free(*stAct);
   return;
}

static inline
void
act_dbg_print_ (
   FILE* fs,
   st_action* stAct,
   const char* pref,
   const char* pref_el,
   const char* post_el,
   const char* post
){
   if (pref_el == NULL) { pref_el = ""; }
   if (post_el == NULL) { post_el = ""; }

   if (pref != NULL) { fprintf(fs, pref); }

   fprintf(fs, "%s%d%s%s%s%s", pref_el, stAct->p_act_def->type, post_el, pref_el, stAct->oper.buf, post_el);
   fprintf(fs, "%s[", pref_el);
   uint i_cs = 0;
   const char* el_cs = NULL;
   while ((el_cs = sarr_it(&stAct->fnames, &i_cs)) != NULL) {
      fprintf(fs, "%s,", el_cs);
   }
   fprintf(fs, "]%s", post_el);

   if (post != NULL) { fprintf(fs, post); }
}

static inline
void
actstab_dbg_print_ (
   FILE* fs,
   st_actions_table* stTable
){
   fprintf(fs, "Priority actions:\n");
   for (uint i = 0; i < stTable->arrActKindsPrior.ns; ++i) {
      st_action* el_act = arrptrs_el_ret(&stTable->arrActKindsPrior, i);
      act_dbg_print_(fs, el_act, "\t", "", ",", "\n");
   }

   fprintf(fs, "Normal actions:\n");
   for (size_t i = 0; i < acttype_count; ++i)
   {
      for (uint j = 0; j < stTable->arrActKinds[i].ns; ++j) {
         st_action* el_act = arrptrs_el_ret(&stTable->arrActKinds[i], j);
         act_dbg_print_(fs, el_act, "\t", "", ",", "\n");
      }
   }
}




#endif /*- TFMAN_SRCPARS_DATATOACTS_INTER_H_INC -*/
