/* This file is part of Tfman, program distributed under GNU General Public
   License. For more information see README and LICENSE files. */

/* current project */
#include "common_tfman.h"
#include "parse_img.h"

/* mh library */
#include "core.h"
#include "errors.h"
#include "debug.h"
#include "faults.h"
#include "memory.h"
#include "cstrings.h"
#include "strings_arr.h"
#include "string_operations.h"
#include "dstream_info.h"

/* libc */

/* current module */
#include "parse_acts.h"
#include "parse_acts_internal.h.inc"
#include "parse_acts_aux.h.inc"

#include "parse_img_internal.h.inc"



static
bool
parsacts_usage_pattern (void)
{
   printf("\t:: TEST: starting %s\n", __func__);

   bool succ = true;
   uint test_num = 1;
   uint test_variant = 1;

  tests_start:
   if (test_num == 1)   // pref_fun1()
   {
      if (test_variant == 1) { // usage
      /* Prepare tags: */
      st_parser_main istParserMain = ST_PARSER_MAIN_Z;
      istParserMain.ind_width = 2;

      /* Read parsed file */
      st_txt_img istTxtRep = ST_TXT_IMG_Z;

      const char* filename = "/home/Magazyn/M/t";
      FILE* fs = fopen (filename, "r");
         ErrorTrigIf(fs == NULL, finish_scope);

      pars_prepare_txtrep(&istTxtRep, fs, filename, &istParserMain);
         HandleErr(handle_abort);

      fclose(fs);

      //: Print faults encountered during parsing
      if (istParserMain.ParsFaults.caths_num[Faultcath_continue] > 0) {
         fprintf(stderr, "\nCONTINUE FAULTS DISCOVERED\n");
         faultsarr_print(0, stderr, &istParserMain.ParsFaults, NULL, "", "\n", NULL);
         fprintf(stderr, "\n");
      }
      if (istParserMain.ParsFaults.caths_num[Faultcath_exit] > 0) {
         fprintf(stderr, "\nSHOULD EXIT - EXIT FAULTS DISCOVERED\n");
         faultsarr_print(0, stderr, &istParserMain.ParsFaults, NULL, "", "\n", NULL);
         fprintf(stderr, "\n");
      }

      //char* cs_sig = cs_cref("\n\nTXT REPRESENTATION from: %s\n", filename);
      //txtrep_dbg_print_(stdout, &istTxtRep, cs_sig, "", "", "\n\n");
      //cs_des(&cs_sig);



      /* Text rep to actions */
      st_actions_table istTable = ST_ACTIONS_TABLE_Z;

      pars_txtrepdata_to_actions(&istTable, &istTxtRep, &istParserMain, 0, 0);
         HandleErr(handle_abort);

      actstab_dbg_print_ (stdout, &istTable);


   finish_scope:
      pars_cleanup_actstable(&istTable);
      pars_cleanup_txtrep(&istTxtRep);
      pars_cleanup_parser_main(&istParserMain);
      CatchError(E_ALL_TYPES, EPR_DEBUG_INFO) {
         puts("ERROR");
      }
      goto common_checks;
      }

      if (test_variant == 2) { //"new2"
      goto common_checks;
      }
   }

   if (test_num == 2)   // pref_fun2()
   {
      if (test_variant == 1) { // "el2"
      goto common_checks;
      }

      if (test_variant == 2) { // ""
      goto common_checks;
      }

      if (test_variant == 3) { // "1"
      goto common_checks;
      }
   }

   if (test_num == 3) { // END
      goto ret;
   }

   ++test_num;
   test_variant = 1;
   goto tests_start;

  common_checks:
   if (succ == true) printf("\t:: TEST: number %d variant %d done\n", test_num, test_variant);
   ++test_variant;
   goto tests_start;

  ret:
   return succ;

  handle_abort:
   CatchError(E_ALL_TYPES, EPR_DEBUG_INFO) {
      abort();
   }
   return false;
}

bool
parsacts_all_tests_run (void)
{
   bool succ = true;
   printf("starting Tests in %s\n", __FILE__);

   if (!parsacts_usage_pattern()) { succ = false; }

   if (succ == true) {
      printf("TESTS:: %s: Tests PASSED\n\n", __FILE__); }
   else {
      printf("TESTS:: %s: ! Tests FAILED\n\n", __FILE__); }

   return succ;
}
