/* This file is part of Tfman, program distributed under GNU General Public
   License. For more information see README and LICENSE files. */

#ifndef TFMAN_SRCPARS_TXTREPTODATA_AUX_H_INC
#define TFMAN_SRCPARS_TXTREPTODATA_AUX_H_INC


/*!F::Parser of file system text representation - auxillary file.
 *
 *  ::Provides helper functions for main implementation file.
 */




/* employs includes from txtrep_todata.c */




static void readin_parsed_file (strarr_t* p_file_content, FILE* fs, const char* filename)
               __attr_nonnull__(1,2,3)
               ;

static void assemble_groups_of_entries_from_txtrep (st_txt_img* stTxtRep,
                        strarr_t* p_file_content, st_parser_main* stParsMain)
               __attr_nonnull__(1,2,3)
               ;

static st_line_ch* factorize_line (str_t* cur_line, uint file_lnum, bool ign_err,
                          st_faults_arr* Faults)
                     __attr_nonnull__(1,4)
                     __attr_wures__
                     ;

static void find_ltype_tag_in_string (size_t* idx, char** p_tag, char** p_after,
                     char* cstr)
               __attr_nonnull__(1,2,3,4)
               ;

static void find_lmod_tag_in_string (size_t* idx, char** p_tag, char** p_after,
                     char* cstr)
               __attr_nonnull__(1,2,3,4)
               ;

static void find_act_tag_in_string (size_t* idx, const char** p_tag, const char** p_after,
                     const char* cstr)
               __attr_nonnull__(1,2,3,4)
               ;




/* Read contents of given file stream to array of strings.
   File stream must be opened for reading.
   ENOMEM
   fs_rline()     -- this function errors
 */
static
void
readin_parsed_file (
   strarr_t* p_file_content,  //! [out] content of parsed file
   FILE* fs,                  //! file stream opened for reading
   const char* filename       //! file name
){
   ___Dassert(p_file_content, "'p_file_content' parameter invalid");
   ___Dassert(fs, "'fs' parameter invalid");
   ___Dassert(filename, "'filename' parameter invalid");

   char* cs_line = NULL;
   size_t nb_read = 0;  //: bytes read from fs (not including \0 added internally by function)

   /*: Record each line: */
   while ((cs_line = fs_rline(fs, &nb_read)))
   {
      //: replace ending \n with \0
      if (nb_read > 1) {
         if (cs_line[nb_read-1] == '\n') { cs_line[nb_read-1] = '\0'; --nb_read; }
      }

      //: append line to array of strings
      sarr_app(p_file_content, cs_line);
         HandleErr(HANDLE_ERR);

      /*cu*/ cs_des(&cs_line);
   }
      HandleErrm(HANDLE_ERR, "error reading parsed file '%s'", filename);

   /*: Finish: */
   return;

  HANDLE_ERR:
   if (cs_line != NULL) { cs_des(&cs_line); }
   return;

}

#define AG_FAULT_CONT(Cond, ...) if ((Cond)) { \
   parsmain_add_fault(stParsMain, Faultcath_continue, __VA_ARGS__); }

#define AG_FAULT_EXIT(Cond, ...) if ((Cond)) { \
   parsmain_add_fault(stParsMain, Faultcath_exit, __VA_ARGS__); }


/* Auxillary struct for assemble_groups_of_entries_from_txtrep(). */
typedef struct {
  uint file_lnum;     //! parsed line number in file
  str_t* cur_line;    //! parsed line
  char* orig_line;    //! parsed line in original state (parsing may change its content)
  st_line_ch* lnch;   //! ptr to line characteristics of current line
  bool ign_err;       //! ignore errors for subsequent actions
} st_parser_line;


/* Transform content of parsed file to structure of groups of entries. This
   structur can be used later to deduce file system actions.
   ENOMEM
 */
static
void
assemble_groups_of_entries_from_txtrep (
   st_txt_img* stTxtRep,      //! [out] textual representation
   strarr_t* p_file_content,  //! content of parsed file
   st_parser_main* stParsMain //! main parser structure
){
   ___Dassert(stTxtRep, "'stTxtRep' parameter invalid");
   ___Dassert(p_file_content, "'p_file_content' parameter invalid");
   ___Dassert(stParsMain, "'stParsMain' parameter invalid");


   /*: Prepare */
   /* ....................................................................... */

   //: create first (default) top group
   stTxtRep->stGrpCh_Top = grpch_constr_(&StLineDefinitions[Linetype_grpd_beg],
                &StLinemodsDefinitions[Linemod_none]);   //: construct top group
      HandleErr(HANDLE_ERR);

   //: prepare main parsing struct
   parsmain_set_cur_grp_(stParsMain, stTxtRep->stGrpCh_Top);   //: set top as current group
   parsmain_set_top_dir_(stParsMain, NULL);
   parsmain_set_par_dir_(stParsMain, NULL);

   //: prepare line parsing struct
   st_parser_line PL = {0};


   /*: Assemble groups from lines, prepare words with dir actions: */
   /* ....................................................................... */

   //: Get each line in content:
   while ((PL.cur_line = sarr_its(p_file_content, &PL.file_lnum)) != NULL)
   {

      /*: Prepare line */
      /* .................................................................... */

      //: ignore empty lines
      if (PL.cur_line->buf[0] == '\0' || PL.cur_line->buf[0] == '\n') { continue; }

      PL.orig_line = cs_cre(PL.cur_line->buf);
         HandleErr(HANDLE_ERR);

      /*: Take line apart into factors: */
      PL.lnch = factorize_line(PL.cur_line, PL.file_lnum, PL.ign_err, &stParsMain->ParsFaults);
         HandleErr(HANDLE_ERR);

      if (PL.lnch == NULL) { goto loop_end; }

      PL.lnch->lnum_in_file = PL.file_lnum;      //: line number in file    //PL.file_lnum is already incremented


    /*: Work out line dependant on type: */
    /* .................................................................... */

    linetype_en ltype = PL.lnch->p_line_def->type;
    switch (ltype)
    {
      /*: '#' dir */
      case Linetype_dir:
        if ((PL.lnch->ind % stParsMain->ind_width) != 0) {
          AG_FAULT_EXIT(true, "line %d: indentation step wrong:\n  %s", PL.file_lnum, PL.orig_line);
        }
        AG_FAULT_EXIT(PL.lnch->ename.nc == 0, "line %d: name required:\n  %s", PL.file_lnum, PL.orig_line);

        if (PL.lnch->ind == 0)
          /*: --- Not indented: --- */
        {
          //: attach line to current group
          grpch_attach_line_(stParsMain->p_cur_group, PL.lnch);
            HandleErr(HANDLE_ERR);

          //: set parent directory
          PL.lnch->p_dir = NULL;    //there is no parent dir

          //: place line in table of lines keeping same order as in parsed file
          txtrep_app_entry_(stTxtRep, PL.lnch);

          //: set parsing state
          parsmain_set_par_dir_(stParsMain, PL.lnch);
          parsmain_set_top_dir_(stParsMain, PL.lnch);
        }
        else
          /*: --- Indented: --- */
        {
          st_line_ch* stPrevDir = stParsMain->p_cur_par_dir; //helper var
          //ind left
          if (stPrevDir->ind > PL.lnch->ind) {
            if (stParsMain->p_cur_top_dir->ind > PL.lnch->ind) {
               AG_FAULT_EXIT(true, "line %d: indentation wrong:\n  %s", PL.file_lnum, PL.orig_line);
            }
            size_t ind_diff = stPrevDir->ind - PL.lnch->ind;
            size_t ind_lvl = ind_diff / stParsMain->ind_width;
            lnch_set_parent_dir_(PL.lnch, 1+ind_lvl, stParsMain->p_cur_par_dir);

          //ind same
          } else if (stPrevDir->ind == PL.lnch->ind) {
            lnch_set_parent_dir_(PL.lnch, 1, stParsMain->p_cur_par_dir);

          //ind right
          } else if (stPrevDir->ind < PL.lnch->ind) {
            size_t ind_diff = PL.lnch->ind - stPrevDir->ind;
            if (ind_diff > stParsMain->ind_width) {
               AG_FAULT_EXIT(true, "line %d: indentation wrong:\n  %s", PL.file_lnum, PL.orig_line);
            }
            lnch_set_parent_dir_(PL.lnch, 0, stParsMain->p_cur_par_dir);
          }
          //: attach line to current group
          grpch_attach_line_(stParsMain->p_cur_group, PL.lnch);
             HandleErr(HANDLE_ERR);

          //: place line in table of lines keeping same order as in parsed file
          txtrep_app_entry_(stTxtRep, PL.lnch);

          //: set parsing state
          parsmain_set_par_dir_(stParsMain, PL.lnch);
        }

        break;

         /*: '|' or '$' entry */
         case Linetype_entry:
         case Linetype_prior:
            //if ((PL.lnch->ind % stParsMain->ind_width) != 0) {
               //AG_FAULT_EXIT(true, "line %d: indentation step wrong:\n  %s", PL.file_lnum, PL.orig_line);
            //}
            AG_FAULT_EXIT(PL.lnch->ename.nc == 0, "line %d: name required:\n  %s", PL.file_lnum, PL.orig_line);

            //point ptr inside line struct to parent dir
            PL.lnch->p_dir = stParsMain->p_cur_par_dir;

            //: attach line to current group
            grpch_attach_line_(stParsMain->p_cur_group, PL.lnch);
               HandleErr(HANDLE_ERR);

            //: place line in table of lines keeping same order as in parsed file
            txtrep_app_entry_(stTxtRep, PL.lnch);

            break;

         /*: '{' '{{' group begin */
         case Linetype_grp_beg:
         case Linetype_grpd_beg:
            { /*- -*/
            AG_FAULT_CONT(PL.lnch->arrWordsCh.ns > 0, "WARN: line %d: action tags without effect for this line type", PL.file_lnum);

            st_group_ch* grp_curr = stParsMain->p_cur_group;  //current group - helper var
            st_group_ch* grp_new = grpch_constr_(PL.lnch->p_line_def, PL.lnch->p_lmod_def); //: construct new group
               HandleErr(err_scope1);

            grpch_attach_group_(grp_curr, grp_new);  //: add new group to current group
               HandleErr(err_scope1);
            //grpch_set_grpparent_(grp_new, grp_curr); //: set current as parent for new
            parsmain_set_cur_grp_(stParsMain, grp_new);   //: set new as current group
            lnch_destr_(&PL.lnch);

            break;

           err_scope1:
            if (grp_new != NULL) { grpch_destr_(&grp_new); }
            goto HANDLE_ERR;
            } /*- -*/

            break;

         /*: '}' '}}' group end */
         case Linetype_grp_end:
         case Linetype_grpd_end:
            if (stParsMain->p_cur_group->p_grp_parent == NULL) {
               AG_FAULT_EXIT(true, "line %d: redundant group closing tag", PL.file_lnum);
               //: line with redundant closing tag is not needed
               lnch_destr_(&PL.lnch);
               goto loop_end;
            }
            linetype_en grp_type = stParsMain->p_cur_group->p_line_def->type;
            AG_FAULT_EXIT((grp_type == Linetype_grp_beg && ltype != Linetype_grp_end)
                      || (grp_type == Linetype_grpd_beg && ltype != Linetype_grpd_end),
                      "line %d: group closing wrong", PL.file_lnum);
            AG_FAULT_CONT(PL.lnch->p_lmod_def->mod != Linemod_none, "WARN: line %d: line modification tag without effect for this line type", PL.file_lnum);

            //: move group words from line to current group
            if (PL.lnch->arrWordsCh.ns > 0) {
               grpch_move_words_(stParsMain->p_cur_group, &PL.lnch->arrWordsCh);
                  HandleErr(HANDLE_ERR);
            }
            lnch_destr_(&PL.lnch);

            //: switch current group to current's parent
            parsmain_switch_current_to_parent_group(stParsMain);

            break;

          /*: Ignore errors */
          case Linetype_ign_errors:
            AG_FAULT_CONT(PL.lnch->p_lmod_def->mod != Linemod_none, "WARN: line %d: line modification tag without effect for this line type", PL.file_lnum);
            AG_FAULT_CONT(PL.lnch->arrWordsCh.ns > 0, "WARN: line %d: action tags without effect for this line type", PL.file_lnum);
            PL.ign_err = !PL.ign_err;
            lnch_destr_(&PL.lnch);
            /*cu*/ cs_des(&PL.orig_line);
            break;

         /*: End parsing */
         case Linetype_parsing_end:
            lnch_destr_(&PL.lnch);
            /*cu*/ cs_des(&PL.orig_line);
            goto HANDLE_SUCC;

         /*: No effective tag */
         case Linetype_none:
            //ignore line
            lnch_destr_(&PL.lnch);
            goto loop_end;

         /*: Wrong types - should never be */
         case linetype_count:
         default:
            ___Sentinel("Wrong effective tag type"); abort();
      }

     loop_end:
      /*cu*/ cs_des(&PL.orig_line);
   }

   if (stParsMain->p_cur_group != stTxtRep->stGrpCh_Top) {
      AG_FAULT_EXIT(true, "some groups are not closed !!!");
   }

  HANDLE_SUCC:
   return;

  HANDLE_ERR:
   if (PL.lnch != NULL) { lnch_destr_(&PL.lnch); }
   if (PL.orig_line != NULL) { cs_des(&PL.orig_line); }
   return;
}

/* Compute following line characteristics: indentation, type of line, line
   modifiers, entry name (closed in [] or first non-action word).
   ENOMEM
   */
static
st_line_ch*
factorize_line (
  str_t* cur_line,        //! parsed line
  uint file_lnum,         //! number of line in parsed file
  bool ign_err,           //! ignore errors for actions in this line
  st_faults_arr* Faults   //! 0-none, 1-missing [ or ], 2-action tag found in the middle of name
){
  ___DParam(cur_line, != NULL);
  ___DParam(file_lnum, != 0);
  ___DParam(Faults, != NULL);

  char* pline = cur_line->buf;   //: pointer to character in line

  strarr_t sarr_words = {0};
  st_line_ch* stLineCh_new = NULL;
  st_word_ch* stWch_new = NULL;

  stLineCh_new = lnch_constr_();
    HandleErr(HANDLE_ERR);


  /* ....................................................................... */
  /*: Compute indentation level and bypass spaces: */
  stLineCh_new->ind = 0;      //: indentation of line
  BypassSpacesC(pline, stLineCh_new->ind);


  /* ....................................................................... */
  /*: Compute type of line and modifier: */

  {/*--*/

  linetype_en ltype = Linetype_none;
  linemod_en lmod = Linemod_none;

  size_t tag_idx = 0;   //: index of found tag in array of tags, 0=NONE=not_found
  char* p_tag = NULL;   //: pointer to found tag, NULL=not_found
  char* p_after = NULL; //: pointer to char right after found tag, NULL-not_found
  bool space_after = false;   //: true if space present after tags

  //: type of line
  find_ltype_tag_in_string(&tag_idx, &p_tag, &p_after, pline);

  if (tag_idx == 0 || p_tag > pline) {
    //: not found OR found in the middle
    if (stLineCh_new->ind > 0) {
      //: line is indented what makes it normal entry
      ltype = Linetype_entry;
    }
  } else if (p_tag == pline) {
    //: found at begining
    ltype = (linetype_en)tag_idx;
    pline = p_after;
  }

  //: line modifiers
  find_lmod_tag_in_string(&tag_idx, &p_tag, &p_after, pline);

  if (p_tag == pline) {
    //: found at begining
    lmod = (linemod_en)tag_idx;
    pline = p_after;
  }

  //: space after tags?
  if (*pline == ' ' || *pline == '\0') {
    space_after = true;
    ++pline;
  }

  //: deduce if line is active
  if (ltype == Linetype_none && lmod == Linemod_none) {
    //: no type, no mod, line inactive
    goto IGNORE_LINE;
  } else if (lmod == Linemod_none) {
    //: type only
    ;
  } else if (ltype == Linetype_none) {
    //: mod only
    if (space_after == true) {
      ltype = Linetype_entry;
      faultsarr_add_fault(Faults, Faultcath_continue, "WARN: line %d: To use line modification tag as name close it in []", file_lnum);
    } else {
      faultsarr_add_fault(Faults, Faultcath_exit, "ERROR: line %d: Entry name begins with line modification tag, separate it or close in []", file_lnum);
      goto IGNORE_LINE;
    }
  } else {
    //: both found
    if (space_after == true) {
      ;
    } else {
      faultsarr_add_fault(Faults, Faultcath_exit, "ERROR: line %d: Entry name begins with line modification tag, separate it or close in []", file_lnum);
      goto IGNORE_LINE;
    }
  }

  stLineCh_new->p_line_def = &StLineDefinitions[ltype];
  stLineCh_new->p_lmod_def = &StLinemodsDefinitions[lmod];

  }/*--*/


  /* ....................................................................... */
  /*: Find name in [] */

  bool got_name = false;              //: true if name isalready found in this line

  {/*--*/

  BypassSpaces(pline);
  const char* const tag_beg = StActionDefinitions[Actspec_name_beg].tag;
  const char* const tag_end = StActionDefinitions[Actspec_name_end].tag;
  const size_t tag_beg_len = StActionDefinitions[Actspec_name_beg].tag_len;
  const size_t tag_end_len = StActionDefinitions[Actspec_name_end].tag_len;

  char* pbeg = NULL;
  char* pend = NULL;
  char* name_beg = NULL;

  //: Find begining of name:
  pbeg = strstr(pline, tag_beg);

  //: Find end of name:
  if (pbeg != NULL) {
    name_beg = pbeg + tag_beg_len;

    char* res = name_beg - 1;
    while (true) {
      res = strstr(res+1, tag_end);
      if (res == NULL) { break; }
      pend = res;
    }
  }

  if (pend != NULL) {
    //: name in [] found
    //: create name
    BypassSpaces(name_beg);
    str_ini2(&stLineCh_new->ename, name_beg, pend - name_beg);
      HandleErr(HANDLE_ERR);
    //: set parsing data and erase name
    while (pbeg <= pend) { *pbeg = ' '; ++pbeg; }
    got_name = true;
  }

  }/*--*/


  /* ....................................................................... */
  /*: Factorize rest of words */

  {/*--*/
    BypassSpaces(pline);

    //: Split to words
    sarr_split_line2(&sarr_words, pline, " ");
      HandleErr(HANDLE_ERR);

    //: For each word
    const char* word = NULL;
    for (size_t i = 0; i < sarr_words.ns; ++i)
    {
      word = sarr_el_ret(&sarr_words, i);

      size_t tag_idx = 0;    //: index of found tag in array of tags, 0=not_found
      const char* p_tag1 = NULL;    //: pointer to found tag, NULL=not_found
      const char* p_after = NULL;  //: pointer to char right after found tag, NULL=not_found

      find_act_tag_in_string(&tag_idx, &p_tag1, &p_after, word);
      const size_t len = StActionDefinitions[tag_idx].tag_len;

      if (tag_idx == 0)
      {
        //: not found
        if (got_name != true) {
          str_ini(&stLineCh_new->ename, word);
            HandleErr(HANDLE_ERR);
          //: set parsing data
          got_name = true;
        }
        goto IGNORE_WORD;
      }
      else
      {
        stWch_new = wch_constr_();
          HandleErr(HANDLE_ERR);
        stWch_new->p_act_def = &StActionDefinitions[tag_idx];
        stWch_new->ign_err = ign_err;

        //: found at begining OR in the middle
        size_t tag2_idx = 0;
        const char* p_tag2 = NULL;
        find_act_tag_in_string(&tag2_idx, &p_tag2, &p_after, p_after);

        switch (stWch_new->p_act_def->kind) {
          case Actkind_nooper:
            if (tag2_idx != 0) {
              faultsarr_add_fault(Faults, Faultcath_exit, "ERROR: line %d: Too many action tags in word: '%s'", file_lnum, word);
              goto IGNORE_WORD;
            }
            if (p_tag1 != word) {
              if (*(p_tag1+len) != '\0') {
                //: cos~cos
                faultsarr_add_fault(Faults, Faultcath_exit, "ERROR: line %d: Tag '%s' in the middle of word", file_lnum, stWch_new->p_act_def->tag);
                goto IGNORE_WORD;
              } else {
                //: cos~
                if (got_name != true) {
                  str_ini2(&stLineCh_new->ename, word, p_tag1 - word);
                    HandleErr(HANDLE_ERR);
                  //: set parsing data
                  got_name = true;
                }
              }
            } else {
              if (*(p_tag1+len) != '\0') {
                //: ~cos
                if (got_name != true) {
                  str_ini(&stLineCh_new->ename, p_tag1 + len);
                    HandleErr(HANDLE_ERR);
                  //: set parsing data
                  got_name = true;
                }
              } else {
                //: ~
              }
            }
            break;
          case Actkind_oper:
            if (tag2_idx != 0) {
              faultsarr_add_fault(Faults, Faultcath_exit, "ERROR: line %d: Too many action tags in word: '%s'", file_lnum, word);
              goto IGNORE_WORD;
            }
            if (p_tag1 != word) {
              if (*(p_tag1+len) != '\0') {
                //: cos->cos
                if (got_name != true) {
                  str_ini2(&stLineCh_new->ename, word, p_tag1 - word);
                    HandleErr(HANDLE_ERR);
                  //: set parsing data
                  got_name = true;
                } else {
                  faultsarr_add_fault(Faults, Faultcath_exit, "ERROR: line %d: more than one name given", file_lnum);
                }
                str_ini(&stWch_new->oper[0], p_tag1 + len);
                  HandleErr(HANDLE_ERR);
              } else {
                //: cos->
                faultsarr_add_fault(Faults, Faultcath_exit, "ERROR: line %d: Tag '%s' requires operand", file_lnum, stWch_new->p_act_def->tag);
                goto IGNORE_WORD;
              }
            } else {
              if (*(p_tag1+len) != '\0') {
                //: ->cos
                str_ini(&stWch_new->oper[0], p_tag1 + len);
                  HandleErr(HANDLE_ERR);
              } else {
                //: ->
                faultsarr_add_fault(Faults, Faultcath_exit, "ERROR: line %d: Tag '%s' requires operand", file_lnum, stWch_new->p_act_def->tag);
                goto IGNORE_WORD;
              }
            }
            break;
          case Actkind_multioper:
            if (tag2_idx != 0) {
              if (stWch_new->p_act_def != &StActionDefinitions[tag2_idx]) {
                faultsarr_add_fault(Faults, Faultcath_exit, "ERROR: line %d: Too many action tags in word: '%s'", file_lnum, word);
                goto IGNORE_WORD;
              }
              size_t tag3_idx = 0;
              const char* p_tag3 = NULL;
              find_act_tag_in_string(&tag3_idx, &p_tag3, &p_after, p_after);
              if (tag3_idx != 0) {
                faultsarr_add_fault(Faults, Faultcath_exit, "ERROR: line %d: Too many action tags in word: '%s'", file_lnum, word);
                goto IGNORE_WORD;
              }
            }

            if (tag2_idx != 0) {
              if (isdigit(word[0])) { goto IGNORE_WORD; }
              //: 2 tags
              const size_t len2 = StActionDefinitions[tag2_idx].tag_len;
              if (p_tag1 == word || (p_tag2+len2) == '\0') {
                //: :any:any, any:any:, :any:
                faultsarr_add_fault(Faults, Faultcath_exit, "ERROR: line %d: Format of tag '%s' is user:group:perms or excerpt", file_lnum, stWch_new->p_act_def->tag);
                goto IGNORE_WORD;
              } else {
                //: any:any:any
                // set 3 opers
                const char* beg = word;
                ptrdiff_t oper_len = p_tag1 - word;
                str_ini2(&stWch_new->oper[0], beg, oper_len);
                beg = p_tag1 + stWch_new->p_act_def->tag_len;
                oper_len = p_tag2 - beg;
                str_ini2(&stWch_new->oper[1], beg, oper_len);
                beg = p_tag2 + stWch_new->p_act_def->tag_len;
                str_ini(&stWch_new->oper[2], beg);
                  HandleErr(HANDLE_ERR);
              }
            } else {
              //: 1 tag
              if (p_tag1 != word) {
                if (*(p_tag1+len) != '\0') {
                  //: any:any
                  // set 1st and 2nd oper
                  const char* beg = word;
                  ptrdiff_t oper_len = p_tag1 - word;
                  str_ini2(&stWch_new->oper[0], beg, oper_len);
                  beg = p_tag1 + stWch_new->p_act_def->tag_len;
                  str_ini(&stWch_new->oper[1], beg);
                } else {
                  //: any:
                  // set 1st oper
                  const char* beg = word;
                  ptrdiff_t oper_len = p_tag1 - word;
                  str_ini2(&stWch_new->oper[0], beg, oper_len);
                }
              } else {
                if (*(p_tag1+len) != '\0') {
                  //: :any
                  // set 2nd oper
                  const char* beg = p_tag1 + stWch_new->p_act_def->tag_len;
                  if (isdigit(beg[0])) { goto IGNORE_WORD; }
                  str_ini(&stWch_new->oper[1], beg);
                } else {
                  //: :
                  faultsarr_add_fault(Faults, Faultcath_exit, "ERROR: line %d: Format of tag '%s' is user:group:perms or excerpt", file_lnum, stWch_new->p_act_def->tag);
                  goto IGNORE_WORD;
                }
              }
            }

            break;
          case Actkind_clear:
            if (p_tag1 != word || *(p_tag1+len) != '\0') {
              faultsarr_add_fault(Faults, Faultcath_exit, "ERROR: line %d: Tag '%s' must be separated from other words", file_lnum, stWch_new->p_act_def->tag);
              goto IGNORE_WORD;
            }
            break;

          case Actkind_dummy:
              goto IGNORE_WORD;

          case Actkind_inactive:
          case Actkind_none:
          case Actkind_name:
          case actkind_count:
          default:
            ___Sentinel("wrong action kind (%d) passed in %s", stWch_new->p_act_def->kind, __func__); abort();
        }
      }
      arrptrs_app(&stLineCh_new->arrWordsCh, stWch_new);
        HandleErr(HANDLE_ERR);
      stWch_new = NULL;   // to avoid double destruction
      continue;

     IGNORE_WORD:
      if (stWch_new != NULL) { wch_destr_(&stWch_new); }
    }

    sarr_fin(&sarr_words);

  }/*--*/


  /* ....................................................................... */
  /*: Finish: */
  return stLineCh_new;

 IGNORE_LINE:
  if (stLineCh_new != NULL) { lnch_destr_(&stLineCh_new); }
  return stLineCh_new;

 HANDLE_ERR:
  if (stLineCh_new != NULL) { lnch_destr_(&stLineCh_new); }
  if (stWch_new != NULL) { wch_destr_(&stWch_new); }
  sarr_fin(&sarr_words);
  return NULL;
}

/* Look for tags from given array of strings in given string. Tag must be found
   at begining of string. First hit is recognized as success. */
void
find_ltype_tag_in_string (
  size_t* idx,          //! [out] index of tag in array (0 - not found)
  char** p_tag,   //! [out] ptr to tag in string
  char** p_after, //! [out] one char after found tag in string
  char* cstr      //! we look here
){
  ___DParam(idx, != NULL);
  ___DParam(p_tag, != NULL);
  ___DParam(p_after, != NULL);
  ___DParam(cstr, != NULL);

  char* cur_res = cstr;
  uint cur_idx = 0;
  size_t cur_len = 0;

  for (uint i = 1; i < linetype_count; ++i) {
    const char* tag = StLineDefinitions[i].tag;
    size_t tag_len = StLineDefinitions[i].tag_len;

    if (tag[0] == '\0') { continue; }
    char* res = strstr(cstr, tag);

    if (res != NULL) {
      /*: found */
      if (res < cur_res || (res == cur_res && tag_len > cur_len)) {
        cur_res = res;
        cur_idx = i;
        cur_len = tag_len;
      }
    }
  }

  *idx = cur_idx;
  *p_tag = cur_res;
  *p_after = *p_tag + cur_len;

   return;
}

void
find_lmod_tag_in_string (
  size_t* idx,          //! [out] index of tag in array (0 - not found)
  char** p_tag,   //! [out] ptr to tag in string
  char** p_after, //! [out] one char after found tag in string
  char* cstr      //! we look here
){
  ___DParam(idx, != NULL);
  ___DParam(p_tag, != NULL);
  ___DParam(p_after, != NULL);
  ___DParam(cstr, != NULL);

  char* cur_res = cstr;
  uint cur_idx = 0;
  size_t cur_len = 0;

  for (uint i = 1; i < linemod_count; ++i)
  {
    const char* tag = StLinemodsDefinitions[i].tag;
    size_t tag_len = StLinemodsDefinitions[i].tag_len;

    if (tag[0] == '\0') { continue; }
    char* res = strstr(cstr, tag);

    if (res != NULL) {
      /*: found */
      if (res < cur_res || (res == cur_res && tag_len > cur_len)) {
        cur_res = res;
        cur_idx = i;
        cur_len = tag_len;
      }
    }
  }

  *idx = cur_idx;
  *p_tag = cur_res;
  *p_after = *p_tag + cur_len;

   return;
}

void
find_act_tag_in_string (
  size_t* idx,          //! [out] index of tag in array (0 - not found)
  const char** p_tag,   //! [out] ptr to tag in string
  const char** p_after, //! [out] one char after found tag in string
  const char* cstr      //! we look here
){
  ___DParam(idx, != NULL);
  ___DParam(p_tag, != NULL);
  ___DParam(p_after, != NULL);
  ___DParam(cstr, != NULL);

  const char* cur_res = NULL;
  uint cur_idx = 0;
  size_t cur_len = 0;


  for (uint i = 1; i < acttype_count; ++i) {
    const char* tag = StActionDefinitions[i].tag;
    size_t tag_len = StActionDefinitions[i].tag_len;
    actkind_en act_kind  = StActionDefinitions[i].kind;

    if (tag[0] == '\0' || act_kind == Actkind_inactive) { continue; }
    char* res = strstr(cstr, tag);

    if (res != NULL) {
      /*: found */
      if (cur_res == NULL || res < cur_res || (res == cur_res && tag_len > cur_len)) {
        cur_res = res;
        cur_idx = i;
        cur_len = tag_len;
      }
    }
  }

  *idx = cur_idx;
  *p_tag = cur_res;
  *p_after = *p_tag + cur_len;

   return;
}




#endif /*- TFMAN_SRCPARS_TXTREPTODATA_AUX_H_INC -*/
