/* This file is part of Tfman, program distributed under GNU General Public
   License. For more information see README and LICENSE files. */

#ifndef TFMAN_SRCPARS_TXTREPTODATA_INTER_H_INC
#define TFMAN_SRCPARS_TXTREPTODATA_INTER_H_INC


/*!F::Parser of file system text representation - internal interface file.
 *
 *  ::Provides functions for basic manipulations of this module data structures.
 */




/* employs includes from txtrep_todata.c */




static inline st_word_ch* wch_constr_ (void)
                           __attr_wures__
                           ;

static inline void wch_destr_ (st_word_ch** pstWordCh)
                     __attr_nonnull__(1)
                     ;

static inline void wch_dbg_print_ (FILE* fs, st_word_ch* stWordCh,
   const char* pref, const char* post)
                     __attr_nonnull__(1,2)
                     ;




static inline st_line_ch* lnch_constr_ (void)
                           __attr_wures__
                           ;

static inline void lnch_destr_ (st_line_ch** pstLineCh)
                     __attr_nonnull__(1)
                     ;

static inline void lnch_set_parent_dir_ (st_line_ch* stLineCh,
                        uint lvls_back, st_line_ch* stParDir)
                     __attr_nonnull__(1,3)
                     ;
/*
static inline void lnch_set_parent_dir_name_ (st_line_ch* stLineCh, str_t* pdir_name)
                     __attr_nonnull__(1,2)
                     ; */

static inline void lnch_dbg_print_ (FILE* fs, st_line_ch* stLineCh,
   const char* pref, const char* pref_el, const char* post_el, const char* post)
                     __attr_nonnull__(1,2)
                     ;




static inline st_group_ch* grpch_constr_ (const st_line_def* ltype, const st_lmod_def* lmod)
                              __attr_wures__
                              ;

static inline void grpch_destr_ (st_group_ch** pstGrpCh)
                     __attr_nonnull__(1)
                     ;
/*
static inline void grpch_set_grpparent_ (st_group_ch* stGrpCh, st_group_ch* stGrpCh_set)
                     __attr_nonnull__(1,2)
                     ; */

static inline void grpch_attach_line_ (st_group_ch* stGrpCh, st_line_ch* stLineCh)
                     __attr_nonnull__(1,2)
                     ;

static inline void grpch_move_words_ (st_group_ch* stGrpCh, arr_ptrs_t* arrWordsCh)
                     __attr_nonnull__(1,2)
                     ;

static inline void grpch_attach_group_ (st_group_ch* stGrpCh, st_group_ch* stGrpCh_app)
                     __attr_nonnull__(1,2)
                     ;

static inline void grpch_dbg_print_ (FILE* fs, st_group_ch* stGrpCh,
   const char* pref, const char* pref_el, const char* post_el, const char* post)
                     __attr_nonnull__(1,2)
                     ;

static inline void txtrep_clean_ (st_txt_img* stTxtRep)
                     __attr_nonnull__(1)
                     ;

static inline void txtrep_app_entry_ (st_txt_img* stTxtRep, st_line_ch* stLineCh)
                     __attr_nonnull__(1,2)
                     ;

static inline void txtrep_dbg_print_ (FILE* fs, st_txt_img* stTxtRep,
   const char* pref, const char* pref_el, const char* post_el, const char* post)
                     __attr_nonnull__(1,2)
                     ;




static inline void parsmain_switch_current_to_parent_group (st_parser_main* stParsMain)
                     __attr_nonnull__(1)
                     ;

static inline void parsmain_clean_ (st_parser_main* stParsMain)
                     __attr_nonnull__(1)
                     ;

static inline void parsmain_add_fault (st_parser_main* stParsMain,
                        faultcath_en cath, const char* fmt, ...)
                     __attr_nonnull__(1,3)
                     __attr_format__(__printf__,3,4)
                     ;

static inline void parsmain_set_top_dir_ (st_parser_main* stParsMain, st_line_ch* stTopDir)
                     __attr_nonnull__(1)
                     ;

static inline void parsmain_set_par_dir_ (st_parser_main* stParsMain,
                              st_line_ch* stPrevDir)
                     __attr_nonnull__(1)
                     ;

static inline void parsmain_set_cur_grp_ (st_parser_main* stParsMain, st_group_ch* stGrpCh)
                     __attr_nonnull__(1,2)
                     ;

static inline void parsmain_dbg_print_ (FILE* fs, st_parser_main* stParsMain)
                     __attr_nonnull__(1,2)
                     ;




/* Construct word characteristics. */
static inline
st_word_ch*
wch_constr_ (void)
{
   st_word_ch* wordch_new = mem_alloc(sizeof *wordch_new);
      HandleErr(handle_err);
   wordch_new->p_act_def = NULL;
   for (size_t i = 0; i < STWORDCH_OPER_NUM; ++i) {
     wordch_new->oper[i] = STR_T_Z;
   }
   wordch_new->prev_file_word = NULL;
   wordch_new->oper_prepared = false;
   wordch_new->ign_err = false;

   return wordch_new;

  handle_err:
   return NULL;
}

/* Destruct word characteristics. */
static inline
void
wch_destr_ (
  st_word_ch** pstWordCh  //! pointer to word characteristics
){
  ___DParam(pstWordCh, != NULL);
  ___DParam(*pstWordCh, != NULL);
  st_word_ch* wordch = *pstWordCh;  //helper var
  for (size_t i = 0; i < STWORDCH_OPER_NUM; ++i) {
    str_fin(&wordch->oper[i]);
  }
  mem_free(*pstWordCh);
  return;
}

/* Print debug info about word characteristics. */
static inline
void
wch_dbg_print_ (
  FILE* fs,            //! file stream
  st_word_ch* stWordCh,//! word characteristics
  const char* pref,    //! prefix for whole print
  const char* post     //! postfix for whole print
){
  ___Dassert(fs, "'fs' parameter invalid");
  ___Dassert(stWordCh, "'stWordCh' parameter invalid");

  if (pref == NULL) { pref = ""; }
  if (post == NULL) { post = ""; }

  fprintf(fs, pref);

  fprintf(fs, "name=%s,tag=%s,type=%d; ops=", stWordCh->p_act_def->name,
      stWordCh->p_act_def->tag, stWordCh->p_act_def->type);
  for (size_t i = 0; i < STWORDCH_OPER_NUM; ++i) {
    fprintf(fs, "%s,", stWordCh->oper[i].buf);
  }

  fprintf(fs, post);

  return;
}




/* Construct line characteristics. */
static inline
st_line_ch*
lnch_constr_ (void
){
   st_line_ch* stLineCh = mem_alloc(sizeof *stLineCh);
      HandleErr(handle_err);

   stLineCh->lnum_in_file = 0;
   stLineCh->ind = 0;
   stLineCh->p_line_def = NULL;
   stLineCh->p_lmod_def = NULL;
   stLineCh->ename = STR_T_Z;
   stLineCh->fullnames = STRARR_T_Z;
   stLineCh->arrWordsCh = ARR_PTRS_T_Z;
   stLineCh->p_grp = NULL;
   stLineCh->p_dir = NULL;

   arrptrs_ini(&stLineCh->arrWordsCh, NULL);
      HandleErr(handle_err);

   return stLineCh;

  handle_err:
   if (stLineCh != NULL) { lnch_destr_(&stLineCh); }
   return NULL;
}

/* Destruct line characteristics. */
static inline
void
lnch_destr_ (
   st_line_ch** pstLineCh
){
   ___Dassert(pstLineCh, "'pstLineCh' parameter invalid");
   ___Dassert(*pstLineCh, "'*pstLineCh' parameter invalid");

   st_line_ch* line_ch = *pstLineCh;

   //str_fin(&line_ch->pdir_name);
   str_fin(&line_ch->ename);
   sarr_fin(&line_ch->fullnames);
   for (size_t i = 0; i < line_ch->arrWordsCh.ns; ++i) {
      st_word_ch* stWordCh = (st_word_ch*) arrptrs_el_ret(&line_ch->arrWordsCh, i);
      if (stWordCh == NULL) { puts("N"); continue; }
      wch_destr_(&stWordCh);
   }
   arrptrs_fin(&line_ch->arrWordsCh, NULL);
   mem_free(*pstLineCh);
   return;
}

static inline
void
lnch_set_parent_dir_ (
   st_line_ch* stLineCh,
   uint lvls_back,
   st_line_ch* stParDir
){
   ___Dassert(stLineCh, "'stLineCh' parameter invalid");

   size_t lb = 0;  //: levels back
   st_line_ch* res = stParDir;

   for (lb = 0; lb < lvls_back; ++lb) {
      res = res->p_dir;
   }
   stLineCh->p_dir = res;

   return;
}

/* Print debug info about line characteristics. */
static inline
void
lnch_dbg_print_ (
   FILE* fs,            //! file stream
   st_line_ch* stLineCh,//! struct
   const char* pref,    //! prefix for whole print
   const char* pref_el, //! prefix for each printed element
   const char* post_el, //! postfix for each printed element
   const char* post     //! postfix for whole print
){
   ___Dassert(fs, "'fs' parameter invalid");
   ___Dassert(stLineCh, "'stLineCh' parameter invalid");

   if (pref_el == NULL) { pref_el = ""; }
   if (post_el == NULL) { post_el = ""; }

   if (pref != NULL) { fprintf(fs, pref); }

   fprintf(fs, "%s%d_%d.%d [%s]%s", pref_el, stLineCh->ind, stLineCh->p_line_def->type,
                              stLineCh->p_lmod_def->mod, stLineCh->ename.buf, post_el);
   st_word_ch* el_wch = NULL;
   fprintf(fs, "%s{", pref_el);
   for (uint i = 0; (el_wch = arrptrs_it(&stLineCh->arrWordsCh, &i)) != NULL ; ) {
      wch_dbg_print_(fs, el_wch, ":", "");
   }
   const char* pdir_name = "";
   if (stLineCh->p_dir != NULL) { pdir_name = stLineCh->p_dir->ename.buf; }
   fprintf(fs, "}%s(pd=%p,%s : pg=%p)%s",
         pref_el, (void*)stLineCh->p_dir, pdir_name, (void*)stLineCh->p_grp, post_el);

   if (post != NULL) { fprintf(fs, post); }
   return;
}




/* Construct group characteristics. */
static inline
st_group_ch*            //! created struct
grpch_constr_ (
   const st_line_def* ltype,  //! normal group or group including dirs
   const st_lmod_def* lmod    //! kind modifiers
){
   st_group_ch* grpch_new = mem_alloc(sizeof *grpch_new);
      HandleErr(handle_err);

   grpch_new->p_line_def = ltype;
   grpch_new->p_lmod_def = lmod;
   grpch_new->arrGroupsCh = ARR_PTRS_T_Z;
   grpch_new->arrLinesCh = ARR_PTRS_T_Z;
   grpch_new->arrWordsCh = ARR_PTRS_T_Z;
   grpch_new->p_grp_parent = NULL;

   return grpch_new;

  handle_err:
   if (grpch_new != NULL) { grpch_destr_(&grpch_new); }
   return NULL;
}

/* Destruct group characteristics. */
static inline
void
grpch_destr_ (
   st_group_ch** pstGrpCh
){
   ___Dassert(pstGrpCh, "'pstGrpCh' parameter invalid");
   ___Dassert(*pstGrpCh, "'*pstGrpCh' parameter invalid");

   st_group_ch* grp = *pstGrpCh;   //helper var

   for (size_t i = 0; i < grp->arrGroupsCh.ns; ++i) {
      st_group_ch* grp_nest = (st_group_ch*) arrptrs_el_ret(&grp->arrGroupsCh, i);
      grpch_destr_(&grp_nest);
   }
   for (size_t i = 0; i < grp->arrLinesCh.ns; ++i) {
      st_line_ch* ln = (st_line_ch*) arrptrs_el_ret(&grp->arrLinesCh, i);
      lnch_destr_(&ln);
   }
   for (size_t i = 0; i < grp->arrWordsCh.ns; ++i) {
      st_word_ch* w = (st_word_ch*) arrptrs_el_ret(&grp->arrWordsCh, i);
      wch_destr_(&w);
   }

   arrptrs_fin(&grp->arrGroupsCh, NULL);
   arrptrs_fin(&grp->arrLinesCh, NULL);
   arrptrs_fin(&grp->arrWordsCh, NULL);
   mem_free(*pstGrpCh);
   return;
}

/* Attach line ch. to group. */
static inline
void
grpch_attach_line_ (
   st_group_ch* stGrpCh,
   st_line_ch* stLineCh
){
   ___Dassert(stGrpCh, "'stGrpCh' parameter invalid");
   ___Dassert(stLineCh, "'stLineCh' parameter invalid");
   stLineCh->p_grp = stGrpCh;
   arrptrs_app(&stGrpCh->arrLinesCh, (void*)stLineCh);
      HandleErr(handle_err);
   return;
  handle_err:
   return;
}

/* Move all elements of words ch. to group. Used for moving actions from line
   to group. */
static inline
void
grpch_move_words_ (
   st_group_ch* stGrpCh,
   arr_ptrs_t* arrWordsCh
){
   ___Dassert(stGrpCh, "'stGrpCh' parameter invalid");
   ___Dassert(arrWordsCh, "'arrWordsCh' parameter invalid");

   st_word_ch* el_wordch = NULL;
   uint i = 0;
   while ((el_wordch = arrptrs_it(arrWordsCh, &i)) != NULL) {
      arrptrs_app(&stGrpCh->arrWordsCh, el_wordch);
         HandleErr(handle_err);
   }
   arrptrs_fin(arrWordsCh, NULL);
   return;

  handle_err:
   arrptrs_fin(&stGrpCh->arrWordsCh, NULL);
   return;
}

/* Attach given (new) group to other (parent) group. */
static inline
void
grpch_attach_group_ (
   st_group_ch* stGrpCh,
   st_group_ch* stGrpCh_app
){
   ___Dassert(stGrpCh, "'stGrpCh' parameter invalid");
   ___Dassert(stGrpCh_app, "'stGrpCh_app' parameter invalid");
   arrptrs_app(&stGrpCh->arrGroupsCh, (void*)stGrpCh_app);
      HandleErr(handle_err);
   stGrpCh_app->p_grp_parent = stGrpCh;
   return;
  handle_err:
   return;
}

/* Print debug info about group characteristics. */
static inline
void
grpch_dbg_print_ (
   FILE* fs,            //! filestream
   st_group_ch* stGrpCh,//! group characteristics
   const char* pref,  //! preceding line
   const char* pref_el,    //! prefix for each printed element
   const char* post_el,    //! postfix for each printed element
   const char* post    //! succeding line
){
   ___Dassert(fs, "'fs' parameter invalid");
   ___Dassert(stGrpCh, "'stGrpCh' parameter invalid");

   if (pref_el == NULL) { pref_el = ""; }
   if (post_el == NULL) { post_el = ""; }

   if (pref != NULL) { fprintf(fs, pref); }

   fprintf(fs, "grp[%p] kind %d:\t\tpar->%p", (void*)stGrpCh, stGrpCh->p_line_def->type, (void*)stGrpCh->p_grp_parent);
   fprintf(fs, "\nGRP_WORDS:");
   for (size_t i = 0; i < stGrpCh->arrWordsCh.ns; ++i) {
      if (stGrpCh->arrWordsCh.buf[i] == NULL) { continue; }
      wch_dbg_print_(fs, stGrpCh->arrWordsCh.buf[i], "", NULL);
   }
   fprintf(fs, "\nLINES:\n");
   for (size_t i = 0; i < stGrpCh->arrLinesCh.ns; ++i) {
      if (stGrpCh->arrLinesCh.buf[i] == NULL) { continue; }
      lnch_dbg_print_(fs, stGrpCh->arrLinesCh.buf[i], "  ", "", " ", "\n");
   }
   fprintf(fs, "\nGROUPS:\n");
   for (size_t i = 0; i < stGrpCh->arrGroupsCh.ns; ++i) {
      if (stGrpCh->arrGroupsCh.buf[i] == NULL) { continue; }
      grpch_dbg_print_(fs, stGrpCh->arrGroupsCh.buf[i], "\n", pref_el, post_el, NULL);
   }

   if (post != NULL) { fprintf(fs, post); }
}




/* Clean text rep. */
static inline
void
txtrep_clean_ (
   st_txt_img* stTxtRep    //! textual representation
){
   ___Dassert(stTxtRep, "'stTxtRep' parameter invalid");
   arrptrs_fin(&stTxtRep->arrEntries, NULL);
   if (stTxtRep->stGrpCh_Top != NULL) { grpch_destr_(&stTxtRep->stGrpCh_Top); }
   return;
}

/* Append line to table of entries keeping order encountered in parsed line. */
static inline
void
txtrep_app_entry_ (
   st_txt_img* stTxtRep,   //! textual representation
   st_line_ch* stLineCh    //! entry line characteristics
){
   ___Dassert(stTxtRep, "'stTxtRep' parameter invalid");
   ___Dassert(stLineCh, "'stLineCh' parameter invalid");
   arrptrs_app(&stTxtRep->arrEntries, (void*)stLineCh);
}

/* Print debug info about text rep. */
static inline
void
txtrep_dbg_print_ (
   FILE* fs,            //! filestream
   st_txt_img* stTxtRep,//! textual representation
   const char* pref,    //! preceding line
   const char* pref_el, //! prefix for each printed element
   const char* post_el, //! postfix for each printed element
   const char* post     //! succeding line
){
   ___Dassert(fs, "'fs' parameter invalid");
   ___Dassert(stTxtRep, "'stTxtRep' parameter invalid");

   if (pref_el == NULL) { pref_el = ""; }
   if (post_el == NULL) { post_el = ""; }

   if (pref != NULL) { fprintf(fs, pref); }
#if 0
   st_line_ch* el_linech = NULL;
   uint i = 0;
   while ((el_linech = arrptrs_it(&stTxtRep->arrDirActions, &i)) != NULL) {
      fprintf(fs, "dirwords [%p]: ", (void*)el_linech);
      st_word_ch* el_wch = NULL;
      uint j = 0;
      while ((el_wch = arrptrs_it(&el_linech->arrWordsCh, &j)) != NULL) {
         wch_dbg_print_(fs, el_wch, "", " ", "", NULL);
      }
      fprintf(fs, "\n");
   }
#endif

   grpch_dbg_print_(fs, stTxtRep->stGrpCh_Top, "\n", pref_el, post_el, NULL);

   if (post != NULL) { fprintf(fs, post); }

   return;
}




/* Set current group to current's parent group (level up). */
static inline
void
parsmain_switch_current_to_parent_group (
   st_parser_main* stParsMain    //! main parser structure
){
   ___Dassert(stParsMain, "'stParsMain' parameter invalid");
   stParsMain->p_cur_group = stParsMain->p_cur_group->p_grp_parent;
   return;
}

/* Clean main parser partialy. */
static inline
void
parsmain_clean_for_next_parse_ (
   st_parser_main* stParsMain    //! main parser structure
){
   ___Dassert(stParsMain, "'stParsMain' parameter invalid");
   //str_fin(&stParsMain->cur_parent_dir);
   stParsMain->p_cur_top_dir = NULL;
   stParsMain->p_cur_par_dir = NULL;
   stParsMain->p_cur_group = NULL;
   sarr_fin(&stParsMain->pars_record);
   faultsarr_finalize(&stParsMain->ParsFaults);
   return;
}

/* Clean main parser. */
static inline
void
parsmain_clean_ (
   st_parser_main* stParsMain    //! main parser structure
){
   ___Dassert(stParsMain, "'stParsMain' parameter invalid");
   parsmain_clean_for_next_parse_(stParsMain);
   stParsMain->ind_width = 0;
   return;
}

/* Add fault to faults array. */
static inline
void
parsmain_add_fault (
   st_parser_main* stParsMain,//! main parser structure
   faultcath_en cath,         //! fault cathegory
   const char* fmt,           //! printf-like format
   ...                        //! format data
){
   ___Dassert(stParsMain, "'stParsMain' parameter invalid");
   ___Dassert(fmt, "'fmt' parameter invalid");

   va_list ap;
   va_start(ap, fmt);
   faultsarr_vadd_fault(&stParsMain->ParsFaults, cath, fmt, ap);
      HandleErr(handle_err);
   va_end(ap);
   return;

  handle_err:
   return;
}

/* Set member - current top directory. */
static inline
void
parsmain_set_top_dir_ (
   st_parser_main* stParsMain,//! main parser structure
   st_line_ch* stTopDir       //! top directory line
){
   ___Dassert(stParsMain, "'stParsMain' parameter invalid");
   stParsMain->p_cur_top_dir = stTopDir;
   return;
}

/* Set member - current parent directory. */
static inline
void
parsmain_set_par_dir_ (
   st_parser_main* stParsMain,//! main parser structure
   st_line_ch* stDir          //! top directory line
){
   ___Dassert(stParsMain, "'stParsMain' parameter invalid");
   stParsMain->p_cur_par_dir = stDir;
   return;
}

/* Set member - pointer to current group. */
static inline
void
parsmain_set_cur_grp_ (
   st_parser_main* stParsMain,   //! main parser structure
   st_group_ch* stGrpCh
){
   ___Dassert(stParsMain, "'stParsMain' parameter invalid");
   ___Dassert(stGrpCh, "'stGrpCh' parameter invalid");
   stParsMain->p_cur_group = stGrpCh;
   return;
}

/* Print debug info about main parser. */
static inline
void
parsmain_dbg_print_ (
   FILE* fs,                     //! file stream
   st_parser_main* stParsMain    //! main parser structure
){
   ___Dassert(stParsMain, "'stParsMain' parameter invalid");
   fprintf(fs, "====== Parser Main struct ======\n");
   fprintf(fs, "p_cur_top_dir=%p (%s)\n", (void*)stParsMain->p_cur_top_dir, stParsMain->p_cur_top_dir->ename.buf);
   fprintf(fs, "p_cur_par_dir=%p (%s)\n", (void*)stParsMain->p_cur_par_dir, stParsMain->p_cur_par_dir->ename.buf);
   fprintf(fs, "p_grp_cur=%p\n", (void*)stParsMain->p_cur_group);
   fprintf(fs, "=============================\n");
   return;
}




#endif /*- TFMAN_SRCPARS_TXTREPTODATA_INTER_H_INC -*/
