#ifndef TFMAN_SRCREPIMG_REP_IMAGE_H
#define TFMAN_SRCREPIMG_REP_IMAGE_H


/*!F::File system data image - external interface file.
 *
 *  ::Provides functions for other program modules.
 */




/* current project */
#include "common_tfman.h"
#include "opts.h"

/* mh library */
#include "core.h"
#include "errors.h"
#include "dstream_info.h"




/* Data. */
/* -------------------------------------------------------------------------- */


/* ...struct.st_imgent........................................*/

typedef struct st_imgent_aux st_imgent_aux;
/*!:: Auxillary struct for 'st_imgent'.
 *    Keeps lengths of longest values of entry fields.
 */
struct st_imgent_aux
{
   uint name;    //! longest name
   uint user;    //! longest user name
   uint group;   //! longest group name
   uint lntar;   //! longest symlink target
   uint entrs;   //! longest number of entries
   uint props;   //! longest properties
};

typedef struct st_imgent st_imgent;
/*!:: Single entry.
 *    'tfman' internal representation of entry.
 */
struct st_imgent
{
   st_ent* stEnt;    //! entry data from 'dstre_info' module
   //!:: Modifiable
   str_t entryname;  //! entryname (=basename.ext)
   str_t filename;   //! absolute/relative filename
   char type;        //! type: "-dlbcsf"  0-indicates that entry is errorneous
   char perms[10];   //! permissions: "rwxrwxrwx", last char must be \0
   str_t owner[2];   //! user and group
   //!:: Unmodifiable
   str_t times[3];   //! access, content modification, entry change times
   str_t sizes[2];   //! sizes in bytes: normal size, block size (sparse bytes accounted)
   uint n_entrs;     //! number of subentries in dir, 0-couldn't open dir (probably perm denied, maybe too many fds in ps/OS open
   st_imgent* stTar; //! symlink target data
   str_t lntar;      //! symlink target name
   str_t props;      //! properties (bitrate, duration time, colors number etc.)
   //!:: Administrative
   st_imgent* _subentry_first;//! for directories only: first entry inside dir
   st_imgent* _next;          //! for each entry: next entry in containig dir
   st_imgent_aux _longest;    //! auxillary data for entry - longest field values
};


/* ...struct.st_fsimg.........................................*/

typedef struct st_fsimg st_fsimg;
/*!:: File system image.
 *    Structure of entries representing directory or whole tree of directories.
 */
struct st_fsimg
{
   st_tfman_opts* stOpts;        //! options used with image creation
   str_t filename;               //! top directory absolute/realtive filename
   uint n_entrs;                 //! number of entries in top directory
   //!:: Administrative
   st_imgent* _subentry_first;   //! first entry in top dir
   st_imgent_aux _longest;       //! auxillary data for entry - longest field values
};




/* Functions. */
/* -------------------------------------------------------------------------- */


/*!1:: Scan directory tree to create file system image. */
extern void repimg_scan_tree
   (st_fsimg* fs_image, const char* topdirname, st_tfman_opts* options)
               __attr_nonnull__(1,2,3)
               ;

/*!1:: Release resources occupied by entries in fs image. */
extern void repimg_cleanup_entries (st_fsimg* fs_image)
               __attr_nonnull__(1)
               ;

/*!1:: Release all resources occupied by fs image. */
extern void repimg_cleanup (st_fsimg* fs_image)
               __attr_nonnull__(1)
               ;




#ifdef DEBUG
/* Used only by tests in debug mode. */
extern bool image_all_tests_run (void)
      __attr_wures__;
#endif




#endif /*- TFMAN_SRCREPIMG_REP_IMAGE_H -*/
