/* This file is part of Tfman, program distributed under GNU General Public
   License. For more information see README and LICENSE files. */

#ifndef TFMAN_SRCREPIMG_REP_IMAGE_AUX_H_INC
#define TFMAN_SRCREPIMG_REP_IMAGE_AUX_H_INC

/*!F::File system data image - auxillary file.
 *
 *  ::Provides helper functions for main implementation file.
 */


/* employs includes from rep_image.c */




/*!1:: Scan directory. */
static inline void fsimg_scan_dir_ (uint* num_entries, str_t* dirname,
         st_imgent** first_entry, uint recurs_lvl, const st_ent_sort* stSort, st_imgent_aux* stLon)
                     __attr_nonnull__(1,2,3,5)
                     ;

/*!1:: Create entry from 'st_ent' structure. */
static inline st_imgent* fsimg_ent_create_ (st_ent* entry)
                           __attr_nonnull__(1)
                           __attr_wures__
                           ;

/*!1:: Insert entry into entries list sorted. */
static inline st_imgent* fsimg_ins_ent_sorted_
      (st_imgent* first_entry, st_imgent* insert_entry, const st_ent_sort* stSort)
                           __attr_nonnull__(1,2,3)
                           ;

/*!1-17
 * 1:: Compare functions. */
static inline trool comp_entries_none_ (const st_imgent* entryA, const st_imgent* entryB)
                     __attr_nonnull__(1,2)
                     __attr_wures__
                     __attr_pure__
                     ;

static inline trool comp_entries_name_ (const st_imgent* entryA, const st_imgent* entryB)
                     __attr_nonnull__(1,2)
                     __attr_wures__
                     __attr_pure__
                     ;

static inline trool comp_entries_nameb_ (const st_imgent* entryA, const st_imgent* entryB)
                     __attr_nonnull__(1,2)
                     __attr_wures__
                     __attr_pure__
                     ;

static inline trool comp_entries_nameidot_ (const st_imgent* entryA, const st_imgent* entryB)
                     __attr_nonnull__(1,2)
                     __attr_wures__
                     __attr_pure__
                     ;

static inline trool comp_entries_nameidotb_ (const st_imgent* entryA, const st_imgent* entryB)
                     __attr_nonnull__(1,2)
                     __attr_wures__
                     __attr_pure__
                     ;

static inline trool comp_entries_type_ (const st_imgent* entryA, const st_imgent* entryB)
                     __attr_nonnull__(1,2)
                     __attr_wures__
                     __attr_pure__
                     ;

static inline trool comp_entries_typeb_ (const st_imgent* entryA, const st_imgent* entryB)
                     __attr_nonnull__(1,2)
                     __attr_wures__
                     __attr_pure__
                     ;

static inline trool comp_entries_dirfst_ (const st_imgent* entryA, const st_imgent* entryB)
                     __attr_nonnull__(1,2)
                     __attr_wures__
                     __attr_pure__
                     ;

static inline trool comp_entries_dirfstb_ (const st_imgent* entryA, const st_imgent* entryB)
                     __attr_nonnull__(1,2)
                     __attr_wures__
                     __attr_pure__
                     ;

static inline trool comp_entries_size_ (const st_imgent* entryA, const st_imgent* entryB)
                     __attr_nonnull__(1,2)
                     __attr_wures__
                     __attr_pure__
                     ;

static inline trool comp_entries_sizeb_ (const st_imgent* entryA, const st_imgent* entryB)
                     __attr_nonnull__(1,2)
                     __attr_wures__
                     __attr_pure__
                     ;

static inline trool comp_entries_atime_ (const st_imgent* entryA, const st_imgent* entryB)
                     __attr_nonnull__(1,2)
                     __attr_wures__
                     __attr_pure__
                     ;

static inline trool comp_entries_atimeb_ (const st_imgent* entryA, const st_imgent* entryB)
                     __attr_nonnull__(1,2)
                     __attr_wures__
                     __attr_pure__
                     ;

static inline trool comp_entries_mtime_ (const st_imgent* entryA, const st_imgent* entryB)
                     __attr_nonnull__(1,2)
                     __attr_wures__
                     __attr_pure__
                     ;

static inline trool comp_entries_mtimeb_ (const st_imgent* entryA, const st_imgent* entryB)
                     __attr_nonnull__(1,2)
                     __attr_wures__
                     __attr_pure__
                     ;

static inline trool comp_entries_ctime_ (const st_imgent* entryA, const st_imgent* entryB)
                     __attr_nonnull__(1,2)
                     __attr_wures__
                     __attr_pure__
                     ;

static inline trool comp_entries_ctimeb_ (const st_imgent* entryA, const st_imgent* entryB)
                     __attr_nonnull__(1,2)
                     __attr_wures__
                     __attr_pure__
                     ;

static inline trool comp_entries_user_ (const st_imgent* entryA, const st_imgent* entryB)
                     __attr_nonnull__(1,2)
                     __attr_wures__
                     __attr_pure__
                     ;

static inline trool comp_entries_userb_ (const st_imgent* entryA, const st_imgent* entryB)
                     __attr_nonnull__(1,2)
                     __attr_wures__
                     __attr_pure__
                     ;

static inline trool comp_entries_group_ (const st_imgent* entryA, const st_imgent* entryB)
                     __attr_nonnull__(1,2)
                     __attr_wures__
                     __attr_pure__
                     ;

static inline trool comp_entries_groupb_ (const st_imgent* entryA, const st_imgent* entryB)
                     __attr_nonnull__(1,2)
                     __attr_wures__
                     __attr_pure__
                     ;




/*!:: Read all entries and attributes. Call itself for subdirectories
 *    accordingly to recursion level. Get spcial statistics for certain
 *    file types like target for symlinks.
 *    Place entries in entries tree in `first_entry`
 *
 * ::Preconditions:
 *  - all other parameters cannot be NULL, except `recurs_lvl` which can be 0
 *
 * ::Errors:
 *    . all ds_dir_create() errors
 *    . all ds_read_directory_stats() errors
 *    . all fsimg_ent_create_() errors
 *    . all ds_count_dir_entries() errors
 *    . all stat_readlink() errors
 */
static inline
void
fsimg_scan_dir_ (
   uint* nentrs,              //! [out] number of entries read
   str_t* dname,        //! directory name to scan
   st_imgent** pstEntFirst,   //! [out] place for first entry in directory
   uint rlvl,                 //! recursion level - how deep descend into subdirs
   const st_ent_sort* stSort, //! sort options
   st_imgent_aux* stLon       //! longest strings for fields
){
   ___Dassert(nentrs, "'nentrs' parameter invalid");
   ___Dassert(dname, "'dname' parameter invalid");
   ___Dassert(pstEntFirst, "'pstEntFirst' parameter invalid");
   ___Dassert(stSort, "'stSort' parameter invalid");

   /* ................................................................ */
   /*:   Read directory with 'ds_*' routines: */
   st_dir* stDir = ds_dir_create(dname->buf);

   char* cs_msg = NULL;
   if (errno == EACCES) { cs_msg = cs_cref("%s: !!! Access denied !!!", dname->buf); }
   else if (errno == ENOENT) { cs_msg = cs_cref("%s: !!! Does not exist !!!", dname->buf); }
   else if (errno == ENOTDIR) { cs_msg = cs_cref("%s: !!! Not directory !!!", dname->buf); }

   /*: Handle error: */
   if (cs_msg != NULL) {
      errno = 0;
      st_imgent* stEntry_err = mem_alloc(sizeof(*stEntry_err));
         HandleErr(handle_err);
      imgent_set_wrong_(stEntry_err, cs_msg);
      *nentrs = 1;            //: set entries read in current dir
      *pstEntFirst = stEntry_err; //: set first entry in dir
      cs_des(&cs_msg);
      goto handle_succ;
   }
      HandleErr(handle_err);

   /*: Get entries and their attributes: */
   ds_read_directory_stats(stDir);
      HandleErr(handle_err);

   /* ................................................................ */
   /*:   Create representation image:
    *    For each entry in directory:
    *    . create img entry
    *    . insert img entry into list
    *    . if entry is dir:
    *       . recursion level allows: scan it - new entries list spawned from
    *                                           current entry
    *       . otherwise: count entries
    *    . if entry is symlink read its target
    */
   size_t nels = stDir->nels;
   st_imgent* stEnt_first = NULL;

   for (ssize_t i = nels-1; i >= 0; --i) {

      st_imgent* stEnt_new = fsimg_ent_create_(stDir->Dentrs[i]);
         HandleErr(handle_err);

      //: insert as first entry or if it's already set insert in sort order
      if (stEnt_first == NULL) {
         stEnt_first = stEnt_new;
      } else {
         stEnt_first = fsimg_ins_ent_sorted_(stEnt_first, stEnt_new, stSort);
      }
      --stDir->nels; //: current entry transfered to image, prevent freeing it by ds_dir_destroy()

      if (stEnt_new->type == 'd')
      {
         //: check if we reached maximum level of recursion
         if ((rlvl > 1 || rlvl == 0)
            //: avoid scanning current directory and its parent directory
            && strcmp(stEnt_new->entryname.buf, "..") != 0
            && strcmp(stEnt_new->entryname.buf, ".") != 0)
         {
            fsimg_scan_dir_(&stEnt_new->n_entrs, &stEnt_new->stEnt->filename,
               &stEnt_new->_subentry_first, rlvl>0?rlvl-1:0, stSort, &stEnt_new->_longest);
               HandleErr(handle_err);
            //if (stEnt_new->_longest.name < stLongest.name) { stEnt_new->_longest.name = stLongest.name; }
            //if (stEnt_new->_longest.user < stLongest.user) { stEnt_new->_longest.user = stLongest.user; }
            //if (stEnt_new->_longest.group < stLongest.group) { stEnt_new->_longest.group = stLongest.group; }
            //if (stEnt_new->_longest.lntar < stLongest.lntar) { stEnt_new->_longest.lntar = stLongest.lntar; }
            //if (stEnt_new->_longest.entrs < stLongest.entrs) { stEnt_new->_longest.entrs = stLongest.entrs; }
            //if (stEnt_new->_longest.props < stLongest.props) { stEnt_new->_longest.props = stLongest.props; }
         }
         else
         {
            stEnt_new->n_entrs = ds_count_dir_entries(stEnt_new->stEnt->filename.buf);
               if (GetErr() == EACCES) {
                  ClrErr(); //: on error ret 0, n_entrs==0 means that num of entires couldn't be read
               } else {
                  HandleErr(handle_err);
               }
         }
      }
      else if (stEnt_new->type == 'l')
      {
         size_t nc = stEnt_new->stEnt->attrs.st_size;
         char* cs_lntar = NULL;
         stat_readlink(&cs_lntar, nc, stEnt_new->stEnt->filename.buf); //: on error assign NULL, means that link couldn't be read
            HandleErr(handle_err);
         if (cs_lntar != NULL) {
            str_fill(&stEnt_new->lntar, cs_lntar);
         }
      }

      /* Compute longest cstrings in current dir */
      /* Compute longest cstring */
      setlocale(LC_ALL,"");
      mbstate_t state = {0};

      size_t len = stEnt_new->entryname.nc;
      const char* cstr = stEnt_new->entryname.buf;
      size_t mlen = mbsrtowcs(NULL, &cstr, len, &state);
      if (mlen > stLon->name) { stLon->name = mlen; }
      stEnt_new->entryname.mnc = mlen;

      len = stEnt_new->owner[0].nc;
      cstr = (const char*)stEnt_new->owner[0].buf;
      mlen = mbsrtowcs(NULL, &cstr, len, &state);
      if (mlen > stLon->user) { stLon->user = mlen; }
      stEnt_new->owner[0].mnc = mlen;

      len = stEnt_new->owner[1].nc;
      cstr = (const char*)stEnt_new->owner[1].buf;
      mlen = mbsrtowcs(NULL, &cstr, len, &state);
      if (mlen > stLon->group) { stLon->group = mlen; }
      stEnt_new->owner[1].mnc = mlen;

      if (stEnt_new->type == 'l') {
         len = stEnt_new->lntar.nc;
         cstr = stEnt_new->lntar.buf;
         mlen = mbsrtowcs(NULL, &cstr, len, &state);
         if (mlen > stLon->lntar) { stLon->lntar = mlen; }
         stEnt_new->lntar.mnc = mlen;
      }

      uint ents = stEnt_new->n_entrs;
      uint digits = 0;
      while (ents) { ents /= 10; ++digits; }
      if (digits > stLon->entrs) { stLon->entrs = digits; }

      if (stEnt_new->props.buf != NULL) {
         len = stEnt_new->props.nc;
         cstr = stEnt_new->props.buf;
         mlen = mbsrtowcs(NULL, &cstr, len, &state);
         if (mlen > stLon->props) { stLon->props = mlen; }
         stEnt_new->props.mnc = mlen;
      }

   }

   ds_dir_destroy(&stDir);
      HandleErr(handle_err);

   *nentrs = nels;            //: set entries read in current dir
   *pstEntFirst = stEnt_first; //: set first entry in dir

    /*: Finish: */

  handle_succ:
   return;

  handle_err:
   if (stDir != NULL) { ds_dir_destroy(&stDir); }

   return;
}

/*!:: Create image entry using entry structure obtained by one of functions from
 *    'dstre_info' module.
 * ::Errors:
 *    ENOMEM      -- not enough memory available
 *    . all ent_init_() errors
 */
static inline
st_imgent*              //! pointer to created entry or NULL on error
fsimg_ent_create_ (
   st_ent* stEnt        //! entry structure
){
   ___Dassert(stEnt, "'stEnt' parameter invalid");

   st_imgent* stEntry_new = NULL;

   //: alloc memory for image entry
   stEntry_new = mem_alloc(sizeof(*stEntry_new));
      HandleErr(handle_err);

   //: init image entry based on ds_info entry values
   if (stEnt->attrs.st_ino == 0) { //was error in reading entry stats
      imgent_set_wrong_attrs_(stEntry_new, stEnt);
   } else {
      imgent_init_(stEntry_new, stEnt);
         HandleErr(handle_err);
   }

    /*: Finish: */
   return stEntry_new;

  handle_err:
   if (stEntry_new != NULL) { imgent_finish_(stEntry_new); }
   mem_free(stEntry_new);
   return NULL;
}

/*!:: Iterate entries list and insert entry accordingly to sorting order.
 * ::Errors:
 *    . all ent_init_() errors
 */
static inline
st_imgent*                    //! first entry in list after inserting
fsimg_ins_ent_sorted_ (
   st_imgent* stEnt_first,    //! current first entry in list
   st_imgent* stEnt,          //! entry to insert
   const st_ent_sort* stSort  //! sort options
){
   //: prepare sort functions array
   typedef trool (*ptr_comp_funs)(const st_imgent*, const st_imgent*);
   ptr_comp_funs comp_funs[] = {
      comp_entries_none_,
      comp_entries_name_,
      comp_entries_nameb_,
      comp_entries_nameidot_,
      comp_entries_nameidotb_,
      comp_entries_type_,
      comp_entries_typeb_,
      comp_entries_dirfst_,
      comp_entries_dirfstb_,
      comp_entries_size_,
      comp_entries_sizeb_,
      comp_entries_atime_,
      comp_entries_atimeb_,
      comp_entries_mtime_,
      comp_entries_mtimeb_,
      comp_entries_ctime_,
      comp_entries_ctimeb_,
      comp_entries_user_,
      comp_entries_userb_,
      comp_entries_group_,
      comp_entries_groupb_
   };

   st_imgent* list_cur = stEnt_first;  // current entry in list
   st_imgent* list_prev = NULL;        // prev entry in list

   /*:   Insert entry:
    *    . iterate elements (entries) in list
    *    . compare entry to element
    *    . insert according to sort options
    */
   while (list_cur != NULL)      //: iterate until last+1 el reached
   {
      //: for each sort kind required compare elements
      //:   until all kinds used or non-equality found
      for (size_t i = 0; i < stSort->els; ++i)
      {
         trool res = 0;
         //: . and .. insert always on two first places
         if (strcmp(stEnt->entryname.buf, ".") == 0) {            //entry to insert
            res = tr_great;
         } else if (strcmp(list_cur->entryname.buf, ".") == 0) {  //entry on list
            res = tr_less;
         } else if (strcmp(stEnt->entryname.buf, "..") == 0) {    //entry to insert
            res = tr_great;
         } else if (strcmp(list_cur->entryname.buf, "..") == 0) { //entry on list
            res = tr_less;
         } else {
            //: use compare function of current sort kind
            res = comp_funs[stSort->ESort[i]](list_cur, stEnt);
         }

         if (res == tr_less) { //: entry on list less then entry to insert
            break;
         }
         else if (res == tr_great) {   //: entry on list greater then entry to insert
            //: insert entry before element on list
            if (list_prev != NULL) { list_prev->_next = stEnt; }
            else { stEnt_first = stEnt; } //: no prev el - insert as first
            stEnt->_next = list_cur;
            return stEnt_first;
         }
         else if (res == tr_equal) {   //: entries equal
            continue;      //: try next kind of sort
         }
         else {                        //: something critically wrong
            ___Sentinel("Invalid trool value detected\n");
            abort();
         }
      }
      //: shift to next element on list
      list_prev = list_cur;
      list_cur = list_cur->_next;
   }
   //: all elements equal to entry to insert in all kinds of sort
   list_prev->_next = stEnt;  //: insert as last
   return stEnt_first;
}

/*!1-17
 * :: Compare 'tfman' entry structs based on various entry attributes.
 */
static inline
trool
comp_entries_none_ (
   const st_imgent* stA __attr_unused__,
   const st_imgent* stB __attr_unused__
){
   return tr_great;
}

static inline
trool
comp_entries_name_ (
   const st_imgent* stA,
   const st_imgent* stB
){
   int res = strcasecmp(stA->entryname.buf, stB->entryname.buf);
   if (res == 0) {
      return tr_equal;
   } else if (res > 0) {
      return tr_great;
   } else {
      return tr_less;
   }
}

static inline
trool
comp_entries_nameb_ (
   const st_imgent* stA,
   const st_imgent* stB
){
   int res = strcasecmp(stA->entryname.buf, stB->entryname.buf);
   if (res == 0) {
      return tr_equal;
   } else if (res > 0) {
      return tr_less;
   } else {
      return tr_great;
   }
}

static inline
trool
comp_entries_nameidot_ (
   const st_imgent* stA,
   const st_imgent* stB
){
   size_t idxA = 0;
   size_t idxB = 0;
   if (stA->entryname.buf[0] == '.') { idxA = 1; }
   if (stB->entryname.buf[0] == '.') { idxB = 1; }

   int res = strcasecmp(&stA->entryname.buf[idxA], &stB->entryname.buf[idxB]);
   if (res == 0) {
      return tr_equal;
   } else if (res > 0) {
      return tr_great;
   } else {
      return tr_less;
   }
}

static inline
trool
comp_entries_nameidotb_ (
   const st_imgent* stA,
   const st_imgent* stB
){
   size_t idxA = 0;
   size_t idxB = 0;
   if (stA->entryname.buf[0] == '.') { idxA = 1; }
   if (stB->entryname.buf[0] == '.') { idxB = 1; }

   int res = strcasecmp(&stA->entryname.buf[idxA], &stB->entryname.buf[idxB]);
   if (res == 0) {
      return tr_equal;
   } else if (res > 0) {
      return tr_less;
   } else {
      return tr_great;
   }
}

static inline
trool
comp_entries_type_ (
   const st_imgent* stA,
   const st_imgent* stB
){
   const st_imgent* cur = stA;
   uint val[2] = { 0, 0 };
   for (uint i = 0; i < 2; ++i) {
   switch (cur->type)
   {
      case '-': val[i] = 2; break;
      case 'd': val[i] = 1; break;     //dirs first
      case 'l': val[i] = 3; break;
      case 'b': val[i] = 4; break;
      case 'c': val[i] = 4; break;
      case 's': val[i] = 4; break;
      case 'f': val[i] = 4; break;
      default: val[i] = 10; ___Sentinel("wrong type value"); break;
   }
   cur = stB;
   }

   if (val[0] == val[1]) { return tr_equal; }
   else if (val[0] < val[1]) { return tr_less; }
   else { return tr_great; }
}

static inline
trool
comp_entries_typeb_ (
   const st_imgent* stA,
   const st_imgent* stB
){
   const st_imgent* cur = stA;
   uint val[2] = { 0, 0 };
   for (uint i = 0; i < 2; ++i) {
   switch (cur->type)
   {
      case '-': val[i] = 2; break;
      case 'd': val[i] = 1; break;
      case 'l': val[i] = 2; break;
      case 'b': val[i] = 3; break;
      case 'c': val[i] = 3; break;
      case 's': val[i] = 3; break;
      case 'f': val[i] = 3; break;
      default: val[i] = 10; ___Sentinel("unpossible type value"); break;
   }
   cur = stB;
   }

   if (val[0] == val[1]) { return tr_equal; }
   else if (val[0] < val[1]) { return tr_great; }
   else { return tr_less; }
}

static inline
trool
comp_entries_dirfst_ (
   const st_imgent* stA,
   const st_imgent* stB
){
   const st_imgent* cur = stA;
   uint val[2] = { 0, 0 };
   for (uint i = 0; i < 2; ++i) {
   switch (cur->type)
   {
      case '-': val[i] = 2; break;
      case 'd': val[i] = 1; break;     //dirs first
      case 'l': val[i] = 2; break;
      case 'b': val[i] = 2; break;
      case 'c': val[i] = 2; break;
      case 's': val[i] = 2; break;
      case 'f': val[i] = 2; break;
      default: val[i] = 10; ___Sentinel("wrong type value"); break;
   }
   cur = stB;
   }

   if (val[0] == val[1]) { return tr_equal; }
   else if (val[0] < val[1]) { return tr_less; }
   else { return tr_great; }
}

static inline
trool
comp_entries_dirfstb_ (
   const st_imgent* stA,
   const st_imgent* stB
){
   const st_imgent* cur = stA;
   uint val[2] = { 0, 0 };
   for (uint i = 0; i < 2; ++i) {
   switch (cur->type)
   {
      case '-': val[i] = 1; break;
      case 'd': val[i] = 2; break;
      case 'l': val[i] = 1; break;
      case 'b': val[i] = 1; break;
      case 'c': val[i] = 1; break;
      case 's': val[i] = 1; break;
      case 'f': val[i] = 1; break;
      default: val[i] = 10; ___Sentinel("unpossible type value"); break;
   }
   cur = stB;
   }

   if (val[0] == val[1]) { return tr_equal; }
   else if (val[0] < val[1]) { return tr_great; }
   else { return tr_less; }
}

static inline
trool
comp_entries_size_ (
   const st_imgent* stA,
   const st_imgent* stB
){
   if (stA->stEnt->attrs.st_size == stB->stEnt->attrs.st_size) { return tr_equal; }
   else if (stA->stEnt->attrs.st_size < stB->stEnt->attrs.st_size) { return tr_great; }
   else { return tr_less; }
}

static inline
trool
comp_entries_sizeb_ (
   const st_imgent* stA,
   const st_imgent* stB
){
   if (stA->stEnt->attrs.st_size == stB->stEnt->attrs.st_size) { return tr_equal; }
   else if (stA->stEnt->attrs.st_size < stB->stEnt->attrs.st_size) { return tr_less; }
   else { return tr_great; }
}

//ctime - atters change (inode) AND data change
//mtime - data change ONLY
static inline
trool
comp_entries_atime_ (
   const st_imgent* stA,
   const st_imgent* stB
){
   if (stA->stEnt->attrs.st_atime == stB->stEnt->attrs.st_atime) { return tr_equal; }
   else if (stA->stEnt->attrs.st_atime < stB->stEnt->attrs.st_atime) { return tr_great; }
   else { return tr_less; }
}

static inline
trool
comp_entries_atimeb_ (
   const st_imgent* stA,
   const st_imgent* stB
){
   if (stA->stEnt->attrs.st_atime == stB->stEnt->attrs.st_atime) { return tr_equal; }
   else if (stA->stEnt->attrs.st_atime < stB->stEnt->attrs.st_atime) { return tr_less; }
   else { return tr_great; }
}

static inline
trool
comp_entries_mtime_ (
   const st_imgent* stA,
   const st_imgent* stB
){
   if (stA->stEnt->attrs.st_mtime == stB->stEnt->attrs.st_mtime) { return tr_equal; }
   else if (stA->stEnt->attrs.st_mtime < stB->stEnt->attrs.st_mtime) { return tr_great; }
   else { return tr_less; }
}

static inline
trool
comp_entries_mtimeb_ (
   const st_imgent* stA,
   const st_imgent* stB
){
   if (stA->stEnt->attrs.st_mtime == stB->stEnt->attrs.st_mtime) { return tr_equal; }
   else if (stA->stEnt->attrs.st_mtime < stB->stEnt->attrs.st_mtime) { return tr_less; }
   else { return tr_great; }
}

static inline
trool
comp_entries_ctime_ (
   const st_imgent* stA,
   const st_imgent* stB
){
   if (stA->stEnt->attrs.st_ctime == stB->stEnt->attrs.st_ctime) { return tr_equal; }
   else if (stA->stEnt->attrs.st_ctime < stB->stEnt->attrs.st_ctime) { return tr_great; }
   else { return tr_less; }
}

static inline
trool
comp_entries_ctimeb_ (
   const st_imgent* stA,
   const st_imgent* stB
){
   if (stA->stEnt->attrs.st_ctime == stB->stEnt->attrs.st_ctime) { return tr_equal; }
   else if (stA->stEnt->attrs.st_ctime < stB->stEnt->attrs.st_ctime) { return tr_less; }
   else { return tr_great; }
}

static inline
trool
comp_entries_user_ (
   const st_imgent* stA,
   const st_imgent* stB
){
   int res = strcmp(stA->owner[0].buf, stB->owner[0].buf);
   if (res == 0) {
      return tr_equal;
   } else if (res > 0) {
      return tr_great;
   } else {
      return tr_less;
   }
}

static inline
trool
comp_entries_userb_ (
   const st_imgent* stA,
   const st_imgent* stB
){
   int res = strcmp(stA->owner[0].buf, stB->owner[0].buf);
   if (res == 0) {
      return tr_equal;
   } else if (res > 0) {
      return tr_less;
   } else {
      return tr_great;
   }
}

static inline
trool
comp_entries_group_ (
   const st_imgent* stA,
   const st_imgent* stB
){
   int res = strcmp(stA->owner[1].buf, stB->owner[1].buf);
   if (res == 0) {
      return tr_equal;
   } else if (res > 0) {
      return tr_great;
   } else {
      return tr_less;
   }
}

static inline
trool
comp_entries_groupb_ (
   const st_imgent* stA,
   const st_imgent* stB
){
   int res = strcmp(stA->owner[1].buf, stB->owner[1].buf);
   if (res == 0) {
      return tr_equal;
   } else if (res > 0) {
      return tr_less;
   } else {
      return tr_great;
   }
}




#endif /*- TFMAN_SRCREPIMG_REP_IMAGE_AUX_H_INC -*/
