/* This file is part of Tfman, program distributed under GNU General Public
   License. For more information see README and LICENSE files. */

#ifndef TFMAN_SRCREPIMG_REP_IMAGE_INTER_H_INC
#define TFMAN_SRCREPIMG_REP_IMAGE_INTER_H_INC


/*!F::File system data image - internal interface file.
 *
 *  ::Provides functions for basic manipulations of module data structures.
 */




/* employs includes from rep_image.c */




static inline void imgent_zero_ (st_imgent* stEntry)
                     __attr_nonnull__(1)
                     ;

static inline void imgent_init_ (st_imgent* stEntry, st_ent* stEnt_src)
                     __attr_nonnull__(1,2)
                     ;

static inline void imgent_finish_ (st_imgent* stEntry)
                     __attr_nonnull__(1)
                     ;

static inline void imgent_destr_ (st_imgent** ptr_stEntry)
                     __attr_nonnull__(1)
                     ;




static inline void imgent_set_wrong_(st_imgent* stEntry, const char* content)
                     __attr_nonnull__(1)
                     ;

static inline void imgent_set_wrong_attrs_ (st_imgent* stEntry, st_ent* stEnt_src)
                     __attr_nonnull__(1)
                     ;




static inline void imgent_print_ (const st_imgent* stEntry, char prefix_char, uint prefix_count)
                     __attr_nonnull__(1)
                     ;

static inline void imgent_print_name_ (const st_imgent* stEntry, char prefix_char, uint prefix_count)
                     __attr_nonnull__(1)
                     ;

static inline void imgent_print_times_ (const st_imgent* stEntry, char prefix_char, uint prefix_count)
                     __attr_nonnull__(1)
                     ;

static inline void imgent_tree_print_ (const st_imgent* stEntry, char prefix_char, uint prefix_count)
                     __attr_nonnull__(1)
                     ;




static inline void fsimg_zero_ (st_fsimg* stFsImage)
                     __attr_nonnull__(1)
                     ;




/*!:: Zero tfman entry.
 * :: Set all members to empty values.
 */
static inline
void
imgent_zero_ (
   st_imgent* stEntry
){
   ___Dassert(stEntry, "'stEntry' parameter invalid");
   stEntry->stEnt = NULL;
   stEntry->entryname = (str_t){0};
   stEntry->filename = (str_t){0};
   stEntry->type = '\0';
   memset(&stEntry->perms[0], 0, sizeof stEntry->perms);
   stEntry->owner[0] = (str_t){0};
   stEntry->owner[1] = (str_t){0};
   for (size_t i = 0; i < sizeof(stEntry->times)/sizeof(stEntry->times[0]); ++i)
      { stEntry->times[i] = (str_t){0}; }
   for (size_t i = 0; i < sizeof(stEntry->sizes)/sizeof(stEntry->sizes[0]); ++i)
      { stEntry->sizes[i] = (str_t){0}; }
   stEntry->n_entrs = 0;
   stEntry->stTar = NULL;
   stEntry->props = (str_t){0};
   stEntry->lntar = (str_t){0};
   stEntry->_subentry_first = NULL;
   stEntry->_next = NULL;
   stEntry->_longest = (st_imgent_aux){0};
   return;
}

// this function has no declaration, it is used only by ent_init_
/* Compress size in B (multiple, demultiple):
      place number in `number` and postfix in `postfix`. */
static inline
void
compact_size_ (
   size_t* number,      //! [out] number
   char** postfix,      //! [out] postfix
   off_t size           //! size in Bytes
){
   const char* pf[] = TFM_SIZE_POSTFIXES; //: postfixes defined somewhere else
   /*: compress number, count compression */
   uint pf_i = 0; //: compression rate
   while (size > 9999)
   {
      size /= 1024;
      ++pf_i;
   }
   /*: Create number and postfix according to compression rate: */
   *number = size;
   char* cstr = *postfix;
   cstr[0] = pf[pf_i][0];
   cstr[1] = pf[pf_i][1];
   cstr[2] = '\0';
   return;
}

/*!:: Initialize tfman entry.
 * :: Initialize all members. Deducing most from entry struct obained by
 *    functions in module 'ds_info'.
 */
static inline
void
imgent_init_ (
   st_imgent* stEntry,
   st_ent* stEnt_src
){
   ___Dassert(stEntry, "'stEntry' parameter invalid");
   ___Dassert(stEnt_src, "'stEnt_src' parameter invalid");

   /*: Prerequisities: */
   imgent_zero_(stEntry);
   stEntry->stEnt = stEnt_src;

   /*: Names: */
   str_ini(&stEntry->entryname, stEntry->stEnt->dent->d_name);
   str_inis(&stEntry->filename, &stEntry->stEnt->filename);

   /*: Type, perms, user, group: */
   stat_extract_ftype(&stEntry->type, stEnt_src->attrs.st_mode);
   stat_extract_perms(&stEntry->perms, stEnt_src->attrs.st_mode);
      HandleErr(handle_err);  // common error check

   char* owner = NULL;
   stat_look_user(&owner, stEnt_src->attrs.st_uid);
      HandleErr(handle_err);
   str_fill(&stEntry->owner[0], owner);
   stat_look_group(&owner, stEnt_src->attrs.st_gid);
      HandleErr(handle_err);
   str_fill(&stEntry->owner[1], owner);

   /*: Times: */
   char* tim = NULL;
   tm_stringize_time(&tim, stEnt_src->attrs.st_atime, "%F %T");
      HandleErr(handle_err);
   str_fill(&stEntry->times[0], tim);
   tm_stringize_time(&tim, stEnt_src->attrs.st_ctime, "%F %T");
      HandleErr(handle_err);
   str_fill(&stEntry->times[1], tim);
   tm_stringize_time(&tim, stEnt_src->attrs.st_mtime, "%F %T");
      HandleErr(handle_err);
   str_fill(&stEntry->times[2], tim);

   /*: Times: */
   size_t size = 0;
   char* postf = mem_alloc(3);
   compact_size_(&size, &postf, stEnt_src->attrs.st_size);
   char* cstr = cs_cref("%zu%s", size, postf);
   str_fill(&stEntry->sizes[0], cstr);
   compact_size_(&size, &postf, stEnt_src->attrs.st_blocks*512);
   cstr = cs_cref("%zu%s", size, postf);
   str_fill(&stEntry->sizes[1], cstr);
      HandleErr(handle_err);  // common error check
   mem_free(postf);

    /*: Finish: */
   return;

  handle_err:
   str_fin(&stEntry->owner[0]);
   str_fin(&stEntry->owner[1]);
   str_fin(&stEntry->times[0]);
   str_fin(&stEntry->times[1]);
   str_fin(&stEntry->times[2]);
   str_fin(&stEntry->sizes[0]);
   str_fin(&stEntry->sizes[1]);
   return;
}


/*!:: Finish tfman entry.
 * :: Release all resources occupied inside struct and set all members to
 *    empty values.
 */
static inline
void
imgent_finish_ (
   st_imgent* stEntry
){
   ___Dassert(stEntry, "'stEntry' parameter invalid");

   if (stEntry->stEnt != NULL) { ds_ent_free(&stEntry->stEnt); }
   str_fin(&stEntry->entryname);
   str_fin(&stEntry->filename);
   str_fin(&stEntry->owner[0]);
   str_fin(&stEntry->owner[1]);
   str_fin(&stEntry->times[0]);
   str_fin(&stEntry->times[1]);
   str_fin(&stEntry->times[2]);
   str_fin(&stEntry->sizes[0]);
   str_fin(&stEntry->sizes[1]);
   if (stEntry->stTar != NULL) { imgent_finish_(stEntry->stTar); }
   str_fin(&stEntry->props);
   str_fin(&stEntry->lntar);

   imgent_zero_(stEntry);

   return;
}

/*!:: Destruct tfman entry.
 * :: Release all resources occupied inside struct and set all members to
 *    empty values. Destruct structs being subentries and next entries.
 *    Free struct itself.
 */
static inline
void
imgent_destr_ (
   st_imgent** pstEntry
){
   ___Dassert(pstEntry, "'pstEntry' parameter invalid");

   st_imgent* stEntry = *pstEntry;

   if (stEntry->_subentry_first != NULL) {
      imgent_destr_(&stEntry->_subentry_first);
   }
   if (stEntry->_next != NULL) {
      imgent_destr_(&stEntry->_next);
   }

   imgent_finish_(stEntry);
   mem_free(*pstEntry);
   *pstEntry = NULL;

   return;
}

/*!:: Set tfman entry content to indicate error.
 * :: Initialize all memebers to reflect some kind of error occured when
 *    reading entry. Type '0', entryname to `content`.
 */
static inline
void
imgent_set_wrong_ (
   st_imgent* stEntry,
   const char* content
){
   ___Dassert(stEntry, "'stEntry' parameter invalid");
   ___Dassert(content, "'content' parameter invalid");

   imgent_zero_(stEntry);

   stEntry->type = '0';
   str_ini(&stEntry->entryname, content);
      HandleErr(handle_err);

    /*: Finish: */
   return;

  handle_err:
   return;
}

/*!:: Set tfman entry content to indicate error in attributes.
 * :: Initialize all memebers to reflect some kind of error occured when
 *    reading attributes of entry.
 *    Set names to normal values.
 *    Set attribute strings to '?' characters.
 */
static inline
void
imgent_set_wrong_attrs_ (
   st_imgent* stEntry,
   st_ent* stEnt_src
){
   ___Dassert(stEntry, "'stEntry' parameter invalid");
   ___Dassert(stEnt_src, "'stEnt_src' parameter invalid");

   imgent_zero_(stEntry);
   stEntry->stEnt = stEnt_src;
   str_ini(&stEntry->entryname, stEntry->stEnt->dent->d_name);
   str_ini(&stEntry->filename, stEntry->stEnt->filename.buf);

   stEntry->type = '?';
   strcpy(stEntry->perms, "?????????");
   str_ini(&stEntry->owner[0], "?");
   str_ini(&stEntry->owner[1], "?");
   str_ini(&stEntry->times[0], "?");
   str_ini(&stEntry->times[1], "?");
   str_ini(&stEntry->times[2], "?");
   str_ini(&stEntry->sizes[0], "?");
   str_ini(&stEntry->sizes[1], "?");
      HandleErr(handle_err);

    /*: Finish: */
   return;

  handle_err:
   str_fin(&stEntry->owner[0]);
   str_fin(&stEntry->owner[1]);
   str_fin(&stEntry->times[0]);
   str_fin(&stEntry->times[1]);
   str_fin(&stEntry->times[2]);
   str_fin(&stEntry->sizes[0]);
   str_fin(&stEntry->sizes[1]);
   return;
}

/*!:: Print tfman entry.
 * :: Print all member values in readable manner.
 */
static inline
void
imgent_print_ (
   const st_imgent* stEntry,
   char pref_c,
   uint pref_n
){
   ___Dassert(stEntry, "'stEntry' parameter invalid");
   for (size_t i = 0; i < pref_n; ++i) { putc(pref_c, stdout); }
   printf("%p: %s %s %c %s ",
         (void*)stEntry->stEnt,
         stEntry->entryname.buf, stEntry->filename.buf, stEntry->type, stEntry->perms);

   printf("%s:%s ", stEntry->owner[0].buf, stEntry->owner[1].buf);
   printf("%s/%s ", stEntry->sizes[0].buf, stEntry->sizes[1].buf);
   printf("%d ->%p (%s)\n", stEntry->n_entrs, (void*)stEntry->stTar, stEntry->props.buf);
   return;
}

/*!:: Print name of tfman entry.
 * :: Print name member.
 */
static inline
void
imgent_print_name_ (
   const st_imgent* stEntry,
   char pref_c,
   uint pref_n
){
   ___Dassert(stEntry, "'stEntry' parameter invalid");
   for (size_t i = 0; i < pref_n; ++i) { putc(pref_c, stdout); }
   printf("%s\n", stEntry->entryname.buf);
   return;
}

/*!:: Print times of tfman entry.
 * :: Print atime, mtime nad ctime members.
 */
static inline
void
imgent_print_times_ (
   const st_imgent* stEntry,
   char pref_c,
   uint pref_n
){
   ___Dassert(stEntry, "'stEntry' parameter invalid");
   for (size_t i = 0; i < pref_n; ++i) { putc(pref_c, stdout); }
   printf("%s - %s - %s\n", stEntry->times[0].buf, stEntry->times[1].buf, stEntry->times[2].buf);
   return;
}

/*!:: Print owners of tfman entry.
 * :: Print user and group.
 */
static inline
void
imgent_print_owners_ (
   const st_imgent* stEntry,
   char pref_c,
   uint pref_n
){
   ___Dassert(stEntry, "'stEntry' parameter invalid");
   for (size_t i = 0; i < pref_n; ++i) { putc(pref_c, stdout); }
   printf("%s %s:%s\n", stEntry->entryname.buf, stEntry->owner[0].buf, stEntry->owner[1].buf);
   return;
}

/*!:: Print tfman entris tree.
 * :: Print all entries in image.
 */
static inline
void
imgent_tree_print_ (
   const st_imgent* stEntry_first,
   char pref_c,
   uint pref_n
){
   ___Dassert(stEntry_first, "'stEntry_first' parameter invalid");
   const st_imgent* cur = stEntry_first;
   while (cur != NULL)
   {
      imgent_print_(cur, pref_c, pref_n);
      if (cur->type == 'd' && cur->_subentry_first != NULL)
      {
         pref_n += 2;
         imgent_tree_print_(cur->_subentry_first, pref_c, pref_n);
         pref_n -= 2;
      }
      cur = cur->_next;
   }
   return;
}

/*!:: Print tfman entries tree, times only.
 * :: Print all entries in image, times only.
 */
static inline
void
ent_tree_print_times_ (
   const st_imgent* stEntry_first,
   char pref_c,
   uint pref_n
){
   ___Dassert(stEntry_first, "'stEntry_first' parameter invalid");
   const st_imgent* cur = stEntry_first;
   while (cur != NULL)
   {
      imgent_print_times_(cur, pref_c, pref_n);
      if (cur->type == 'd' && cur->_subentry_first != NULL)
      {
         pref_n += 2;
         ent_tree_print_times_(cur->_subentry_first, pref_c, pref_n);
      }
      cur = cur->_next;
   }
   return;
}

/*!:: Print tfman entries tree, owners only.
 * :: Print all entries in image, owners only.
 */
static inline
void
imgent_tree_print_owners_ (
   const st_imgent* stEntry_first,
   char pref_c,
   uint pref_n
){
   ___Dassert(stEntry_first, "'stEntry_first' parameter invalid");
   const st_imgent* cur = stEntry_first;
   while (cur != NULL)
   {
      imgent_print_owners_(cur, pref_c, pref_n);
      if (cur->type == 'd' && cur->_subentry_first != NULL)
      {
         pref_n += 2;
         imgent_tree_print_owners_(cur->_subentry_first, pref_c, pref_n);
         pref_n -= 2;
      }
      cur = cur->_next;
   }
   return;
}




/*!:: Zero fs image.
 * :: Set all members to empty values.
 */
static inline
void
fsimg_zero_ (
   st_fsimg* stImg
){
   ___Dassert(stImg, "'stImg' parameter invalid");
   stImg->stOpts = NULL;
   stImg->filename = (str_t){0};
   stImg->_subentry_first = NULL;
   stImg->n_entrs = 0;
   stImg->_longest = (st_imgent_aux){0};
   return;
}




#endif /*- TFMAN_SRCREPIMG_REP_IMAGE_INTER_H_INC -*/
