/* This file is part of Tfman, program distributed under GNU General Public
   License. For more information see README and LICENSE files. */

/* strcasecmp():  _DEFAULT_SOURCE */
#define _DEFAULT_SOURCE

/* mh library */
#include "core.h"
#include "errors.h"
#include "memory.h"
#include "cstrings.h"
#include "strings.h"
#include "times.h"         // tm_stringize_time()
#include "dstream_info.h"       // ds_count_dir_entries()
#include "stat_info.h"     // stat_readlink()

/* current project */
#include "common_tfman.h"
#include "opts.h"

/* libc */
#include <errno.h>
#include <string.h>     // memset(), strcmp()
#include <stdio.h>      // printf(), putc()
#include <locale.h>     // setlocale()
#include <wchar.h>      // mbsrtowcs()

/* current module */
#include "scan_img.h"
#include "scan_img_internal.h.inc"
#include "scan_img_aux.h.inc"




static
bool
fsimg_usage_pattern (void)
{
   printf("\t:: TEST: starting %s\n", __func__);

   bool succ = true;
   uint test_num = 1;
   uint test_variant = 1;

   sort_en Sort_order[] = { Sort_groupb, Sort_user, Sort_sizeb };
   format_en Format[] = { Format_name, Format_perms, Format_user, Format_props, Format_group, Format_type };
   st_tfman_opts stSfmanOpts;
   tfmopts_initialize (
      &stSfmanOpts,
   /* general */
      Mode_parse,
      0,
      false,               //usage
      false,               //help
      2,
   /* scan */
      false,               //append to file in scan mode
      0,
      Sort_order, 3,
      Format, 6,
      Lay_block,
      NULL,                //output file name
      false,               //clone entry names
      false,               //names only
   /* parse */
      false,               //show actions only and exit
      true,        //ask user before perfroming actions
      0,                   //start retrieving actions at this line
      0                    //end retrieving actions at this line
   );

  tests_start:
   if (test_num == 1)   // pref_fun1()
   {
      if (test_variant == 1) { // ""
      st_fsimg stFsImage;
      repimg_scan_tree(&stFsImage, "/m/test", &stSfmanOpts);
         HandleErrm(handle_abort, "reported error in %s (%s)\n", err_ret_msg(), err_descr(errno));

      // --- program actions
      //printf("entries:");
      //ent_tree_print_(stFsImage._subentry_first, ' ', 2);

      repimg_cleanup(&stFsImage);
      goto common_checks;
      }

      if (test_variant == 2) { //"new2"
      goto common_checks;
      }
   }

   if (test_num == 2)   // pref_fun2()
   {
      if (test_variant == 1) { // "el2"
      goto common_checks;
      }

      if (test_variant == 2) { // ""
      goto common_checks;
      }

      if (test_variant == 3) { // "1"
      goto common_checks;
      }
   }

   if (test_num == 3) { // END
      goto ret;
   }

   ++test_num;
   test_variant = 1;
   goto tests_start;

  common_checks:
   if (succ == true) printf("\t:: TEST: number %d variant %d done\n", test_num, test_variant);
   ++test_variant;
   goto tests_start;

  ret:
   return succ;

  handle_abort:
   abort();
}

/*
static
bool
internal_interface_test (void)
{
   printf("\t:: TEST: starting %s\n", __func__);

   bool succ = true;
   uint test_num = 1;
   uint test_variant = 1;

  tests_start:
   if (test_num == 1)   // ent_*_()
   {
      if (test_variant == 1) { // ent_zero_
      st_imgent stEntry;
      memset(&stEntry, 9, sizeof stEntry);
      ent_zero_ (&stEntry);
      if (stEntry.dirent != NULL) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stEntry.stat != NULL) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (!str_is_zero(&stEntry.name)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stEntry.type != 0) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
         for (uint i = 0; i < 9; ++i) {
      if (stEntry.perms[i] != '\0') { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
         }
      if (!strarr_is_zero(&stEntry.owner)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (!strarr_is_zero(&stEntry.times)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (!strarr_is_zero(&stEntry.sizes)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stEntry.n_entrs != 0) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (!str_is_zero(&stEntry.props)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stEntry._subentry_first != NULL) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stEntry._next != NULL) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      goto common_checks;
      }

      if (test_variant == 2) { // ent_init_/finish_
      st_imgent stEntry = {0};
      memset(&stEntry, 9, sizeof stEntry);
      st_imgent stEntry2 = {0};
      memset(&stEntry2, 9, sizeof stEntry2);
      st_imgent stEntry3 = {0};
      memset(&stEntry3, 9, sizeof stEntry3);
      st_imgent stEntry4 = {0};
      memset(&stEntry4, 9, sizeof stEntry4);
      st_imgent stEntry5 = {0};
      memset(&stEntry5, 9, sizeof stEntry5);

      const char* times[3] = { "atime", "mtime", "ctime" };
      const char* sizes[2] = { "256kB", "512mB" };

      ent_init_ (&stC, &stEntry5, (struct dirent*)7, (struct stat*)7, "ole", 't', "aaaaaaaaa",
            "user", "group", times, sizes, 12, "", NULL, NULL);
      ent_init_ (&stC, &stEntry3, (struct dirent*)7, (struct stat*)7, "ole", 't', "aaaaaaaaa",
            "user", "group", times, sizes, 12, "", NULL, &stEntry5);
      ent_init_ (&stC, &stEntry4, (struct dirent*)7, (struct stat*)7, "ole", 't', "aaaaaaaaa",
            "user", "group", times, sizes, 12, "", NULL, &stEntry3);
      if (stEntry4._next == NULL) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      ent_init_ (&stC, &stEntry, (struct dirent*)7, (struct stat*)7, "ole", 't', "aaaaaaaaa",
            "user", "group", times, sizes, 12, "", &stEntry4, NULL);
      ent_init_ (&stC, &stEntry2, (struct dirent*)7, (struct stat*)7, "ole", 't', "aaaaaaaaa",
            "user", "group", times, sizes, 12, "props", &stEntry, NULL);
         HandleErrm(handle_abort, "'ent_init_' error (%s)\n", errt_Labels[stC.stErr.eType]);
      if (stEntry2.dirent != (struct dirent*)7) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stEntry2.stat != (struct stat*)7) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (strcmp(stEntry2.name.cstr, "ole") != 0) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stEntry2.type != 't') { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
         for (uint i = 0; i < 9; ++i) {
      if (stEntry2.perms[i] != 'a') { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
         }

      if (strcmp(stEntry2.owner.Strings[0]->cstr, "user") != 0) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (strcmp(stEntry2.owner.Strings[1]->cstr, "group") != 0) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (strcmp(stEntry2.times.Strings[0]->cstr, "atime") != 0) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (strcmp(stEntry2.times.Strings[1]->cstr, "mtime") != 0) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (strcmp(stEntry2.times.Strings[2]->cstr, "ctime") != 0) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (strcmp(stEntry2.sizes.Strings[0]->cstr, "256kB") != 0) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (strcmp(stEntry2.sizes.Strings[1]->cstr, "512mB") != 0) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stEntry2.n_entrs != 12) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (strcmp(stEntry2.props.cstr, "props") != 0) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stEntry2._subentry_first == NULL) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stEntry2._next != NULL) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      ent_finish_ (&stEntry2);
      if (stEntry2.dirent != NULL) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stEntry2.stat != NULL) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (!str_is_zero(&stEntry2.name)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stEntry2.type != 0) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
         for (uint i = 0; i < 9; ++i) {
      if (stEntry2.perms[i] != '\0') { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
         }
      if (!strarr_is_zero(&stEntry2.owner)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (!strarr_is_zero(&stEntry2.times)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (!strarr_is_zero(&stEntry2.sizes)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stEntry2.n_entrs != 0) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (!str_is_zero(&stEntry2.props)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stEntry2._subentry_first != NULL) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stEntry2._next != NULL) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      goto common_checks;
      }
   }

   if (test_num == 2)   // img_*_()
   {
      if (test_variant == 1) { // img_zero_()
      st_fsimg stFsImage;
      memset(&stFsImage, 9, sizeof stFsImage);
      img_zero_ (&stFsImage);
      if (!str_is_zero(&stFsImage.dirname)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stFsImage.stOpts != NULL) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stFsImage.stEntryTop.dirent != NULL) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stFsImage.stEntryTop.stat != NULL) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (!str_is_zero(&stFsImage.stEntryTop.name)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stFsImage.stEntryTop.type != 0) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
         for (uint i = 0; i < 9; ++i) {
      if (stFsImage.stEntryTop.perms[i] != '\0') { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
         }
      if (!strarr_is_zero(&stFsImage.stEntryTop.owner)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (!strarr_is_zero(&stFsImage.stEntryTop.times)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (!strarr_is_zero(&stFsImage.stEntryTop.sizes)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stFsImage.stEntryTop.n_entrs != 0) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (!str_is_zero(&stFsImage.stEntryTop.props)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stFsImage.stEntryTop._subentry_first != NULL) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stFsImage.stEntryTop._next != NULL) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      goto common_checks;
      }

      if (test_variant == 2) { // img_init_/finish_()
      st_fsimg stFsImage;
      memset(&stFsImage, 9, sizeof stFsImage);
      img_init_(&stC, &stFsImage, "dirname", (st_tfman_opts*)777);
         HandleErrm(handle_abort, "'img_init_' error (%s)\n", errt_Labels[stC.stErr.eType]);
      if (strcmp(stFsImage.dirname.cstr, "dirname") != 0) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stFsImage.stOpts != (st_tfman_opts*)777) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stFsImage.stEntryTop.dirent != NULL) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stFsImage.stEntryTop.stat != NULL) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (!str_is_zero(&stFsImage.stEntryTop.name)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stFsImage.stEntryTop.type != 0) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
         for (uint i = 0; i < 9; ++i) {
      if (stFsImage.stEntryTop.perms[i] != '\0') { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
         }
      if (!strarr_is_zero(&stFsImage.stEntryTop.owner)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (!strarr_is_zero(&stFsImage.stEntryTop.times)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (!strarr_is_zero(&stFsImage.stEntryTop.sizes)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stFsImage.stEntryTop.n_entrs != 0) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (!str_is_zero(&stFsImage.stEntryTop.props)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stFsImage.stEntryTop._subentry_first != NULL) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stFsImage.stEntryTop._next != NULL) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }

      const char* times[3] = { "atime", "mtime", "ctime" };
      const char* sizes[2] = { "256kB", "512mB" };
      ent_init_ (&stC, &stFsImage.stEntryTop, (struct dirent*)7, (struct stat*)7, "ole", 't', "aaaaaaaaa",
            "user", "group", times, sizes, 12, "", NULL, NULL);
      img_finish_(&stFsImage);
      if (!str_is_zero(&stFsImage.dirname)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stFsImage.stOpts != NULL) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stFsImage.stEntryTop.dirent != NULL) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stFsImage.stEntryTop.stat != NULL) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (!str_is_zero(&stFsImage.stEntryTop.name)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stFsImage.stEntryTop.type != 0) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
         for (uint i = 0; i < 9; ++i) {
      if (stFsImage.stEntryTop.perms[i] != '\0') { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
         }
      if (!strarr_is_zero(&stFsImage.stEntryTop.owner)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (!strarr_is_zero(&stFsImage.stEntryTop.times)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (!strarr_is_zero(&stFsImage.stEntryTop.sizes)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stFsImage.stEntryTop.n_entrs != 0) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (!str_is_zero(&stFsImage.stEntryTop.props)) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stFsImage.stEntryTop._subentry_first != NULL) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      if (stFsImage.stEntryTop._next != NULL) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
      goto common_checks;
      }

      if (test_variant == 3) { // "1"
      goto common_checks;
      }
   }

   if (test_num == 3) { // END
      goto ret;
   }

   ++test_num;
   test_variant = 1;
   goto tests_start;

  common_checks:
   if (stC.stErr.eGroup != Eg_NONE) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
   if (stC.stErr.eType != Et_NONE) { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
   if (stC.stErr.acstr[0] != '\0') { LogTest("test num %d failed at line %d", test_num, __LINE__); succ = false; }
   if (succ == true) printf("\t:: TEST: number %d variant %d done\n", test_num, test_variant);
   ++test_variant;
   goto tests_start;

  ret:
   return succ;

  lab_abort:
   abort();
}
*/




bool
image_all_tests_run (void)
{
   bool succ = true;
   printf("starting Tests in %s\n", __FILE__);

   if (!fsimg_usage_pattern()) { succ = false; }
 //  if (!internal_interface_test()) { succ = false; }

   if (succ == true) {
      printf("TESTS:: %s: Tests PASSED\n\n", __FILE__); }
   else {
      printf("TESTS:: %s: ! Tests FAILED\n\n", __FILE__); }

   return succ;
}
