#ifndef TFMAN_SRCREPTXT_REP_TEXT_H
#define TFMAN_SRCREPTXT_REP_TEXT_H

/*!F::File system text representing - external interface file.
 *
 *  ::Provides functions for other program modules.
 */




/* mh library */
#include "core.h"
#include "errors.h"
#include "strings_arr.h"

/* current project */
#include "common_tfman.h"
#include "scan_img.h"




typedef struct st_fstxt st_fstxt;
/*!:: Text representation of single directory.
 *    One line for each entry plus other lines and other data like prefix
 *    strings.
 */
struct st_fstxt
{
   char* l_top_dname;
   char** lines;        //! array of lines (c-strings)
   size_t nlns;         //! num of lines in array allocated
   size_t nptrs;        //! num of pointers in array allocated
   st_fsimg* stFsImg;   //! file system data image - some its members needed during forging lines
   strarr_t prefstrs;   //! tags for 0-dir blk 1-dir tree 2-active entry
   strarr_t otherstrs;  //! action tags
};




/*!1:: Set prefix strings. */
extern void fstxt_set_pref_strings (st_fstxt* stFsText,
      const char* cs_dir_blk,
      const char* cs_ent)
               __attr_nonnull__(1,2,3)
               ;

/*!1:: Set field strings. */
extern void fsrxt_set_field_strings (st_fstxt* stFsText,
      const char* cs_name_beg, const char* cs_name_end,
      const char* link_target, const char* link_target_nex,
      const char* cs_separ, const char* cs_space,
      const char* cs_atime, const char* cs_mtime, const char* cs_ctime)
               __attr_nonnull__(1,2,3,4,5,6,7,8,9)
               ;

/*!1:: Create text representation of directory. */
extern void fstxt_image_to_text (st_fstxt* fs_text_repr, st_fsimg* fs_image)
               __attr_nonnull__(1,2)
               ;

/*!1:: Cleanup text representation struct. */
extern void fstxt_cleanup (st_fstxt* fs_text_repr)
               __attr_nonnull__(1)
               ;

/*!1:: Cleanup c-strings with text in text representation struct. */
extern void fstxt_cleanup_text (st_fstxt* fs_text_repr)
               __attr_nonnull__(1)
               ;

/*!1:: Print text representation to given file stream. */
extern void fstxt_print (FILE* fs, st_fstxt* stFsTxt, uint verbosity,
                  st_tfman_opts* stOpts)
               __attr_nonnull__(1,2,4)
               ;




#ifdef DEBUG
/* Used only by tests in debug mode. */
extern bool fstxt_all_tests_run (void)
      __attr_wures__;
#endif




#endif /*- TFMAN_SRCREPTXT_REP_TEXT_H -*/
