/* This file is part of Tfman, program distributed under GNU General Public
   License. For more information see README and LICENSE files. */

#ifndef TFMAN_SRCREPTXT_REP_TEXT_INTER_H_INC
#define TFMAN_SRCREPTXT_REP_TEXT_INTER_H_INC


/*!F::File system text representation - internal interface file.
 *
 *  ::Provides functions for basic manipulations of module data structures.
 */




/* employs includes from image.c */




/*!:: Internal constants: */
#define PSTR_DIRBLK 0
#define PSTR_ENTRY_ACTIVE 1

#define OSTR_NAME_BEG 0
#define OSTR_NAME_END 1
#define OSTR_TARGET 2
#define OSTR_TARGET_NEX 3
#define OSTR_SEPAR 4
#define OSTR_SPACE 5
#define OSTR_ATIME 6
#define OSTR_MTIME 7
#define OSTR_CTIME 8




static inline void fstxt_zero_ (st_fstxt* fs_text_repr)
                     __attr_nonnull__(1)
                     ;

static inline void fstxt_init_ (st_fstxt* stFsTxt, st_fsimg* stFsImg)
                     __attr_nonnull__(1,2)
                     ;

static inline void fstxt_txt_finish_ (st_fstxt* fs_text_repr)
                     __attr_nonnull__(1)
                     ;

static inline void fstxt_finish_ (st_fstxt* fs_text_repr)
                     __attr_nonnull__(1)
                     ;

static inline void fstxt_expand_ (st_fstxt* fs_text_repr, size_t num_ptrs_add)
                     __attr_nonnull__(1)
                     ;




/* Manipulate text representation structure.
 *    Zero - set all members to empty values.
 *    Initialize - init members, do not free previous data
 *    Text finish - release resources taken by entry lines
 *    Finish - release all resources
 *    Expand - expand array of c-strings holding entry lines
 */
static inline
void
fstxt_zero_ (
   st_fstxt* stFsTxt
){
   ___Dassert(stFsTxt, "'stFsTxt parameter invalid");
   stFsTxt->l_top_dname = NULL;
   stFsTxt->lines = NULL;
   stFsTxt->nlns = 0;
   stFsTxt->nptrs = 0;
   return;
}

static inline
void
fstxt_init_ (
   st_fstxt* stFsTxt,
   st_fsimg* stFsImg
){
   ___Dassert(stFsTxt, "'stFsTxt parameter invalid");
   fstxt_zero_(stFsTxt);
   stFsTxt->stFsImg = stFsImg;
   return;
}

static inline
void
fstxt_txt_finish_ (
   st_fstxt* stFsTxt
){
   ___Dassert(stFsTxt, "'stFsTxt parameter invalid");
   /*: Release lines in array. */
   cs_des(&stFsTxt->l_top_dname);
   for (size_t i = 0; i < stFsTxt->nlns; ++i) {
      cs_des(&stFsTxt->lines[i]);
   }
   /*: Release array itself. */
   mem_free(stFsTxt->lines);
   stFsTxt->nlns = 0;
   stFsTxt->nptrs = 0;
   return;
}

static inline
void
fstxt_finish_ (
   st_fstxt* stFsTxt
){
   ___Dassert(stFsTxt, "'stFsTxt parameter invalid");
   if (stFsTxt->l_top_dname != NULL) { fstxt_txt_finish_(stFsTxt); }
   //repimg_cleanup(stFsTxt->stFsImg);  -- cleared outside
   sarr_fin(&stFsTxt->prefstrs);
   sarr_fin(&stFsTxt->otherstrs);
   return;
}

static inline
void
fstxt_expand_ (
   st_fstxt* stFsTxt,
   size_t nptrs_add
){
   ___Dassert(stFsTxt, "'stFsTxt' parameter invalid");
   ___Dassert(nptrs_add > 0, "'nptrs_add' parameter invalid");

   char** lines_new = NULL;

      OverflowIntIf(stFsTxt->nptrs + nptrs_add < nptrs_add, handle_err);
   size_t n_new = stFsTxt->nptrs + nptrs_add;
      OverflowIntIf(n_new > SIZE_MAX/sizeof(*lines_new), handle_err);
   size_t nb = n_new * sizeof(*lines_new);

   if (stFsTxt->lines == NULL) {
      lines_new = (char**) mem_alloc(nb);
         HandleErr(handle_err);
   } else {
      lines_new = (char**) mem_realloc(stFsTxt->lines, nb);
         HandleErr(handle_err);
   }

   stFsTxt->lines = lines_new;
   stFsTxt->nptrs = n_new;

    /*: Finish: */
   return;

  handle_err:
   mem_free(lines_new);
   return;
}





#endif /*- TFMAN_SRCREPTXT_REP_TEXT_INTER_H_INC -*/
