/* This file is part of Tfman, program distributed under GNU General Public
   License. For more information see README and LICENSE files. */

/* mh library */
#include "core.h"
#include "errors.h"
#include "memory.h"
#include "cstrings.h"
//#include "times.h"         // tm_stringize_time()
//#include "dstream_info.h"       // ds_count_dir_entries()
#include "stat_info.h"

/* current project */
#include "common_tfman.h"
#include "opts.h"
#include "scan_img.h"

/* libc */
#include <string.h>     // memset(), strcmp()
#include <stdio.h>      // printf(), putc()

/* current module */
#include "scan_txt.h"
#include "scan_txt_internal.h.inc"
#include "scan_txt_aux.h.inc"




static
bool
fstxt_usage_pattern (void)
{
   printf("\t:: TEST: starting %s\n", __func__);

   bool succ = true;
   uint test_num = 1;
   uint test_variant = 1;

   sort_en Sort_order[] = { Sort_groupb, Sort_user, Sort_sizeb };
   format_en Format[] = { Format_name, Format_lntarget, Format_perms, Format_name, Format_user, Format_group, Format_type, Format_separ, Format_props };
   int format_els = sizeof(Format) / sizeof(Format[0]);
   st_tfman_opts stSfmanOpts;
   tfmopts_initialize (
      &stSfmanOpts,
   /* general */
      Mode_parse,
      0,
      false,               //usage
      false,               //help
      2,
   /* scan */
      false,               //append to file in scan mode
      0,
      Sort_order, 3,
      Format, format_els,
      Lay_block,
      NULL,                //output file name
      false,               //clone entry names
      false,               //names only
   /* parse */
      false,               //show actions only and exit
      true,        //ask user before perfroming actions
      0,                   //start retrieving actions at this line
      0                    //end retrieving actions at this line
   );

  tests_start:
   if (test_num == 1)   // pref_fun1()
   {
      if (test_variant == 1) { // usage
      st_fsimg stFsImage;
      repimg_scan_tree(&stFsImage, "/m/test", &stSfmanOpts);
         HandleErrm(handle_abort, "reported error (%s)\n", err_descr(errno));

      st_fstxt stFsText;
      fstxt_set_pref_strings(&stFsText, "#", "|");
      fsrxt_set_field_strings(&stFsText, "[", "]", "_>", "!!>", " -- ", "  ", "a.", "m.", "c.");
      fstxt_image_to_text(&stFsText, &stFsImage);
         HandleErrm(handle_abort, "reported error (%s)\n", err_descr(errno));
      // --- write to file
      fstxt_print(stdout, &stFsText, true);

      fstxt_cleanup(&stFsText);
      goto common_checks;
      }

      if (test_variant == 2) { //"new2"

      goto common_checks;
      }
   }

   if (test_num == 2)   // pref_fun2()
   {
      if (test_variant == 1) { // "el2"
      goto common_checks;
      }

      if (test_variant == 2) { // ""
      goto common_checks;
      }

      if (test_variant == 3) { // "1"
      goto common_checks;
      }
   }

   if (test_num == 3) { // END
      goto ret;
   }

   ++test_num;
   test_variant = 1;
   goto tests_start;

  common_checks:
   if (succ == true) printf("\t:: TEST: number %d variant %d done\n", test_num, test_variant);
   ++test_variant;
   goto tests_start;

  ret:
   return succ;

  handle_abort:
   abort();
}




bool
fstxt_all_tests_run (void)
{
   bool succ = true;
   printf("starting Tests in %s\n", __FILE__);

   if (!fstxt_usage_pattern()) { succ = false; }
 //  if (!internal_interface_test()) { succ = false; }

   if (succ == true) {
      printf("TESTS:: %s: Tests PASSED\n\n", __FILE__); }
   else {
      printf("TESTS:: %s: ! Tests FAILED\n\n", __FILE__); }

   return succ;
}
